/* **********************************************************************
    Copyright 2003 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.uwcal.icalendar;

import edu.rpi.cct.uwcal.calsvci.CalSvcI;
import edu.rpi.cct.uwcal.common.URIgen;
import edu.rpi.cct.uwcal.resources.Resources;
import edu.washington.cac.calfacade.shared.CalFacadeException;
import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.GuidVO;
import edu.washington.cac.calfacade.shared.IntArrayList;
import edu.washington.cac.calfacade.shared.KeywordVO;
import edu.washington.cac.calfacade.shared.LocationVO;
import edu.washington.cac.calfacade.shared.MyCalendarVO;
import edu.washington.cac.calfacade.shared.RecurVO;
import edu.washington.cac.calfacade.shared.RecurRuleVO;
import edu.washington.cac.calfacade.shared.SponsorVO;
import edu.washington.cac.calfacade.shared.UserVO;

import net.fortuna.ical4j.data.CalendarBuilder;
import net.fortuna.ical4j.data.CalendarOutputter;
import net.fortuna.ical4j.data.CalendarParserImpl;
import net.fortuna.ical4j.model.Calendar;
import net.fortuna.ical4j.model.CategoryList;
import net.fortuna.ical4j.model.component.VEvent;
import net.fortuna.ical4j.model.ComponentList;
import net.fortuna.ical4j.model.DayList;
import net.fortuna.ical4j.model.NumberList;
import net.fortuna.ical4j.model.parameter.AltRep;
import net.fortuna.ical4j.model.parameter.Value;
import net.fortuna.ical4j.model.Property;
import net.fortuna.ical4j.model.Recur;
import net.fortuna.ical4j.model.property.Categories;
import net.fortuna.ical4j.model.property.Clazz;
import net.fortuna.ical4j.model.property.Contact;
import net.fortuna.ical4j.model.property.Created;
import net.fortuna.ical4j.model.property.Description;
import net.fortuna.ical4j.model.property.DtEnd;
import net.fortuna.ical4j.model.property.DtStamp;
import net.fortuna.ical4j.model.property.DtStart;
//import net.fortuna.ical4j.model.property.ExDate;
import net.fortuna.ical4j.model.property.ExRule;
import net.fortuna.ical4j.model.property.LastModified;
import net.fortuna.ical4j.model.property.Location;
import net.fortuna.ical4j.model.property.Organizer;
import net.fortuna.ical4j.model.property.ProdId;
//import net.fortuna.ical4j.model.property.RDate;
import net.fortuna.ical4j.model.property.RRule;
import net.fortuna.ical4j.model.property.Sequence;
import net.fortuna.ical4j.model.property.Summary;
import net.fortuna.ical4j.model.property.Uid;
import net.fortuna.ical4j.model.property.Url;
import net.fortuna.ical4j.model.property.Version;
import net.fortuna.ical4j.model.PropertyList;

import java.io.StringReader;
import java.io.StringWriter;
import java.io.Serializable;
import java.net.URI;
import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import org.apache.log4j.Logger;

/** Object to provide translation between an EventVO and an Icalendar format.
 *
 * @author Mike Douglass   douglm@rpi.edu
 */
public class IcalTranslator implements Serializable {
  private boolean debug;

  /** We'll use this to parameterize some of the behaviour
   */
  public static class Pars {
    /** Include VALUE=DATE_TIME property where appropriate
     */
    public boolean includeDateTimeProperty;

    /** Support simple location only. Many iCalendar-aware 
     * applications only support a simple string valued location. 
     *
     * <p>If this value is true we only pass the address part of the location
     * and provide an altrep which will allow (some) clients to retrieve the
     * full form of a location.
     */
    public boolean simpleLocation = true;

    /** Support simple contacts only.
     */
    public boolean simpleContact = true;
  }

  /* This needs to come from a property updated with each release.
   */
  private static final String prodId = "UW calendar V2.0";

  private Logger log;

  private CalSvcI svci;

  private URIgen uriGen;

  private Pars pars = new Pars();

  /** Constructor:
   *
   * @param svci     CalSvcI object for retrieval of entities
   * @param uriGen   URIgen object used to provide ALTREP values.
   */
  public IcalTranslator(CalSvcI svci, URIgen uriGen) {
    this(svci, uriGen, false);
  }

  /** Constructor:
   *
   * @param svci     CalSvcI object for retrieval of entities
   * @param uriGen   URIgen object used to provide ALTREP values.
   */
  public IcalTranslator(CalSvcI svci, URIgen uriGen, boolean debug) {
    this.svci = svci;
    this.uriGen = uriGen;
    this.debug = debug;
  }

  /** ===================================================================
   *                     Translation methods
   *  =================================================================== */

  public Calendar toIcal(EventVO val) throws CalFacadeException {
    if (val == null) {
      return null;
    }

    try {
      Calendar cal = new Calendar();

      cal.getProperties().add(new ProdId(prodId));
      cal.getProperties().add(new Version());

      /* Add it to the calendar */
      cal.getComponents().add(toIcalEvent(val));

      return cal;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }
  }

  /** Make a VEvent object from an EventVO.
   */
  public VEvent toIcalEvent(EventVO val) throws CalFacadeException {
    if (val == null) {
      return null;
    }

    try {
      VEvent vev = new VEvent();
      boolean pubEvent = val.getPublic();

      PropertyList pl = vev.getProperties();
      Property prop;

      /* ------------------- Categories -------------------- */
      /* This needs to be changed if we add non-category keys */

      Collection ks = val.getKeywords();

      if (ks != null) {
        /* This event has a category */

        Iterator it = ks.iterator();
        prop = new Categories();
        CategoryList cl = ((Categories)prop).getCategories();

        while (it.hasNext()) {
          KeywordVO k = (KeywordVO)it.next();

          cl.add(k.getWord());
        }

        pl.add(prop);
      } else {
        /* Most personal events don't have a category */
      }

      /* ------------------- Class -------------------- */

      if (pubEvent) {
        pl.add(new Clazz("PUBLIC"));
      } else {
        pl.add(new Clazz("PRIVATE"));
      }

      /* ------------------- Contact -------------------- */

      SponsorVO sp = val.getSponsor();
      if (sp != null) {
        prop = new Contact(makeContactString(sp));
        prop.getParameters().add(new AltRep(uriGen.getSponsorURI(sp)));
        pl.add(prop);
      }

      /* ------------------- Created -------------------- */

      prop = new Created(val.getCreated());
      if (pars.includeDateTimeProperty) {
        prop.getParameters().add(new Value(Value.DATE_TIME));
      }
      pl.add(prop);

      /* ------------------- Description -------------------- */

      pl.add(new Description(val.getLongdesc()));

      /* ------------------- DtEnd -------------------- */


      prop = new DtEnd(makeDateTime(val.getEnddate(), val.getEndtime()));
      if (pars.includeDateTimeProperty) {
        prop.getParameters().add(new Value(Value.DATE_TIME));
      }
      pl.add(prop);

      /* ------------------- DtStamp -------------------- */

      /* ------------------- DtStart -------------------- */

      prop = new DtStart(makeDateTime(val.getStartdate(), val.getStarttime()));
      if (pars.includeDateTimeProperty) {
        prop.getParameters().add(new Value(Value.DATE_TIME));
      }
      pl.add(prop);


      /* ------------------- LastModified -------------------- */

      prop = new LastModified(val.getLastmod());
      if (pars.includeDateTimeProperty) {
        prop.getParameters().add(new Value(Value.DATE_TIME));
      }
      pl.add(prop);

      /* ------------------- Location -------------------- */

      LocationVO loc = val.getLocation();
      if (loc != null) {
        prop = new Location(makeLocationString(loc));
        prop.getParameters().add(new AltRep(uriGen.getLocationURI(loc)));
        pl.add(prop);
      }

      /* ------------------- Organizer -------------------- */

      Organizer org = new Organizer();
      org.setValue(val.getCreator().getAccount());
      pl.add(org);

      /* ------------------- Sequence -------------------- */

      pl.add(new Sequence(val.getSeq()));

      /* ------------------- Summary -------------------- */

      pl.add(new Summary(val.getShortdesc()));

      /* ------------------- Uid -------------------- */

      pl.add(new Uid(val.getGuid()));

      /* ------------------- Url -------------------- */

      String uri = val.getLink();

      if (uri != null) {
        pl.add(new Url(new URI(uri)));
      }

      return vev;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }
  }

  /** Convert the EventVO object to a String representation
   */
  public String toStringIcal(EventVO val) throws CalFacadeException {
    Calendar ical = toIcal(val);

    CalendarOutputter calOut = new CalendarOutputter(true);

    StringWriter sw = new java.io.StringWriter();

    try {
      calOut.output(ical, sw);

      return sw.toString();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }
  }

  /** Convert the given string representation of an Icalendar object to an EventVO
   *
   * <p>Because an icalendar object can contain 0 or more VEvents we return
   * a collection of events which may be empty.
   *
   */
  public Collection toEvent(String val) throws CalFacadeException {
    try {
      CalendarBuilder bldr = new CalendarBuilder(new CalendarParserImpl());

      return toEvent(bldr.build(new StringReader(val)));
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }
  }

  /** Convert the given string representation of an Icalendar object to a
   * collection of VEvent
   *
   * <p>Because an icalendar object can contain 0 or more VEvents we return
   * a collection of vevents which may be empty.
   *
   */
  public Collection toVEvent(String val) throws CalFacadeException {
    try {
      CalendarBuilder bldr = new CalendarBuilder(new CalendarParserImpl());

      Calendar cal = bldr.build(new StringReader(val));
      Vector evs = new Vector();

      if (cal == null) {
        return evs;
      }

      ComponentList clist = cal.getComponents();

      Iterator it = clist.iterator();

      while (it.hasNext()) {
        Object o = it.next();

        if (o instanceof VEvent) {
         evs.add(o);
       }
      }

      return evs;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }
  }

  public Collection toEvent(Calendar val) throws CalFacadeException {
    Vector evs = new Vector();

    if (val == null) {
      return evs;
    }

    ComponentList clist = val.getComponents();

    Iterator it = clist.iterator();

    while (it.hasNext()) {
      Object o = it.next();

      if (o instanceof VEvent) {
        EventVO ev = toEvent((VEvent)o, null);

        if (ev != null) {
          evs.add(ev);
        }
      }
    }

    return evs;
  }

  /** We are going to try to construct an EventVO object from a VEvent. This
   * may represent a new event or an update to a pre-existing event. In any
   * case, the VEvent probably has insufficient information to completely
   * reconstitute the event object so we'll get the uid first and retrieve
   * the event if it exists.
   *
   * <p>If it doesn't exist, we'll first fill in the appropriate fields,
   * (non-public, creator, created etc) then for both cases update the
   * remaining fields from the VEvent.
   *
   * @param val         VEvent object
   * @param guidStr     String guid value
   * @return EventVO object representing new entry or updated entry
   */
  public EventVO toEvent(VEvent val, String guidStr) throws CalFacadeException {
    if (val == null) {
      return null;
    }

    try {
      PropertyList pl = val.getProperties();

      if (pl == null) {
        // Empty VEvent
        return null;
      }

      /** We need to know if this is a public event.
       */
      Property prop = pl.getProperty(Property.CLASS);
      boolean pubEvent = (prop != null) && ("PUBLIC".equals(prop.getValue()));

      if (guidStr == null) {
        // Try for one in the VEvent

        prop = pl.getProperty(Property.UID);
        if (prop != null) {
          guidStr = prop.getValue();
        }
      }

      GuidVO guid = null;

      if (guidStr != null) {
        guid = GuidVO.valueOf(guidStr);
      }

      EventVO ev = null;

      if ((guid != null) && (guid.getExternalGuid() == null)) {
        // It's one of ours
        ev = svci.getEvent(guid.getId());
      }

      if (ev != null) {
        // Check the validity of some of the fields?
      } else {
        ev = new EventVO();
        ev.setCreator(svci.getUserVO());
        ev.setPublic(pubEvent);
      }

      Iterator it = pl.iterator();

      while (it.hasNext()) {
        prop = (Property)it.next();

        String pval = prop.getValue();
        if ((pval != null) && (pval.length() == 0)) {
          pval = null;
        }

        if (prop instanceof Categories) {
          /* ------------------- Categories -------------------- */
          /* This needs to be changed if we add non-category keys */

          CategoryList cl = ((Categories)prop).getCategories();

          if (cl != null) {
            /* Got some categories */

            Iterator cit = cl.iterator();

            while (cit.hasNext()) {
              String wd = (String)cit.next();

              KeywordVO k = new KeywordVO();
              k.setWord(wd);

              k = svci.findKeyword(k);

              if (k == null) {
                System.out.println("************************************");
                System.out.println("************************************");
                System.out.println("Fix missing keyword in icalendar->event");
                System.out.println("************************************");
                System.out.println("************************************");
              } else {
                ev.addKeyword(k);
              }
            }
          } else {
            /* Most personal events don't have a category */
          }
        } else if (prop instanceof Clazz) {
          /* ------------------- Class -------------------- */

          /* We did this above */
        } else if (prop instanceof Contact) {
          /* ------------------- Contact -------------------- */

          AltRep alt = getAltRep(prop);

          /* The related db sponsor entry */
          SponsorVO dbsp = null;

          boolean same = false;

          if (ev != null ) {
            dbsp = ev.getSponsor();
          } else if (alt != null) {
            /* Try to use the altrep to retrieve the contact
             */
            dbsp = uriGen.getSponsor(alt.getUri());

            if (dbsp != null) {
              dbsp = svci.getSponsor(dbsp.getId());
            }
          }

          if (dbsp != null) {
            /* Build what we think we sent them
             */
            String dbspVal = makeContactString(dbsp);

            if ((pval == null) && (dbspVal == null)) {
              same = true;
            } else if ((pval != null) && pval.equals(dbspVal)) {
              same = true;
            }
          }

          if (!same) {
            /** Value seems to have changed - any way to parse new value?*/
            ev.setSponsor(makeContact(pval));
          } else {
            ev.setSponsor(dbsp);
          }
        } else if (prop instanceof Created) {
          /* ------------------- Created -------------------- */

          ev.setCreated(makeSqlTimestamp(((Created)prop).getDateTime()));
        } else if (prop instanceof Description) {
          /* ------------------- Description -------------------- */

          ev.setLongdesc(pval);
        } else if (prop instanceof DtEnd) {
          /* ------------------- DtEnd -------------------- */

          DtEnd dtProp = (DtEnd)prop;
          java.util.Date dt = dtProp.getTime();

          ev.setEnddate(makeSqlDate(dt));
          ev.setEndtime(makeSqlTime(dt));
        } else if (prop instanceof DtStamp) {
          /* ------------------- DtStamp -------------------- */
        } else if (prop instanceof DtStart) {
          /* ------------------- DtStart -------------------- */

          DtStart dtProp = (DtStart)prop;
          java.util.Date dt = dtProp.getTime();

          ev.setStartdate(makeSqlDate(dt));
          ev.setStarttime(makeSqlTime(dt));
        } else if (prop instanceof LastModified) {
          /* ------------------- LastModified -------------------- */

          ev.setLastmod(makeSqlTimestamp(((LastModified)prop).getDateTime()));
        } else if (prop instanceof Location) {
          /* ------------------- Location -------------------- */

          AltRep alt = getAltRep(prop);

          /* The related db LocationVO entry */
          LocationVO dbloc = null;

          boolean same = false;

          if (ev != null ) {
            dbloc = ev.getLocation();
          } else if (alt != null) {
            /* Try to use the altrep to retrieve the location
             */
            dbloc = uriGen.getLocation(alt.getUri());

            if (dbloc != null) {
              dbloc = svci.getLocation(dbloc.getId());
            }
          }

          if (dbloc != null) {
            /* Build what we think we sent them
             */
            String dblocVal = makeLocationString(dbloc);

            if ((pval == null) && (dblocVal == null)) {
              same = true;
            } else if ((pval != null) && pval.equals(dblocVal)) {
              same = true;
            }
          }

          if (!same) {
            /** Value seems to have changed - any way to parse new value?*/
            ev.setLocation(makeLocation(pval));
          } else {
            ev.setLocation(dbloc);
          }
        } else if (prop instanceof Organizer) {
          /* ------------------- Organizer -------------------- */

          ev.setCreator(new UserVO(pval));
        } else if ((prop instanceof ExRule) ||
                   (prop instanceof RRule)) {
          /* ------------------- RRule -------------------- */

          Recur r;
          RecurRuleVO rr = new RecurRuleVO();

          RecurVO rvo = ev.getRecur();

          if (rvo == null) {
            rvo = new RecurVO();
            ev.setRecur(rvo);
          }

          if (prop instanceof ExRule) {
            r = ((ExRule)prop).getRecur();
            rr.setRecurrence(false);
            rvo.addExclusionRule(rr);
          } else {
            r = ((RRule)prop).getRecur();
            rr.setRecurrence(true);
            rvo.addRecurrenceRule(rr);
          }

          rr.setRecur(rvo);
          rr.setId(rvo.getRecurrenceRules().size());

          rr.setUntil(r.getUntil());
          rr.setCount(r.getCount());

          rr.setInterval(r.getInterval());

          rr.setBySecond(makeIAL(r.getSecondList()));
          rr.setByMinute(makeIAL(r.getMinuteList()));
          rr.setByHour(makeIAL(r.getHourList()));

          DayList dl = r.getDayList();
          if ((dl != null) && (dl.size() > 0)) {
            Iterator dlit = dl.iterator();
            IntArrayList ial = new IntArrayList();

            while (dlit.hasNext()) {
              String s = (String)dlit.next();

              ial.add(RecurRuleVO.toIntWday(s));
            }

            rr.setByDay(ial);
          }

          rr.setByMonthday(makeIAL(r.getMonthDayList()));
          rr.setByYearday(makeIAL(r.getYearDayList()));
          rr.setByWeekno(makeIAL(r.getWeekNoList()));
          rr.setByMonth(makeIAL(r.getMonthList()));
          rr.setBySetpos(makeIAL(r.getSetPosList()));

          rr.setWkst(RecurRuleVO.toIntWday(r.getWeekStartDay()));
        } else if (prop instanceof Sequence) {
          /* ------------------- Sequence -------------------- */

          ev.setSeq(((Sequence)prop).getSequenceNo());
        } else if (prop instanceof Summary) {
          /* ------------------- Summary -------------------- */

          ev.setShortdesc(pval);
        } else if (prop instanceof Uid) {
          /* ------------------- Uid -------------------- */

          /* We did this above */
        } else if (prop instanceof Url) {
          /* ------------------- Url -------------------- */

          ev.setLink(pval);
        } else {
          if (debug) {
            getLog().debug("Unsupported property with class " + prop.getClass() +
                           " and value " + pval);
          }
        }
      }

      return ev;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }
  }

  /** ===================================================================
                      Private methods
      =================================================================== */

  private IntArrayList makeIAL(NumberList val) {
    if ((val == null) || (val.size() == 0)) {
      return null;
    }

    IntArrayList res = new IntArrayList();
    Iterator it = val.iterator();

    while (it.hasNext()) {
      res.add(((Integer)it.next()).intValue());
    }

    return res;
  }

  private static final String defaultDelim = ", ";
  private static final String urlDelim = " - ";

  private static final String nameValSep = ": ";

  private void addNonNull(String delim, String val, StringBuffer sb) {
    if ((val != null) && (val.length() > 0)) {
      sb.append(delim);
      sb.append(val);
    }
  }

  private void addNonNull(String delim, String resourceKey,
                          String val, StringBuffer sb) {
    if ((val != null) && (val.length() > 0)) {
      sb.append(delim);
      sb.append(svci.getResources().getString(resourceKey));
      sb.append(nameValSep);
      sb.append(val);
    }
  }

  private String makeContactString(SponsorVO sp) {
    if (pars.simpleContact) {
      return sp.getName();
    }

    StringBuffer sb = new StringBuffer(sp.getName());
    addNonNull(defaultDelim, Resources.PHONENBR, sp.getPhone(), sb);
    addNonNull(defaultDelim, Resources.EMAIL, sp.getEmail(), sb);
    addNonNull(urlDelim, Resources.URL, sp.getLink(), sb);

    if (sb.length() == 0) {
      return null;
    }

    return sb.toString();
  }

  /** Build a location string value from the location.
   *
   * <p>We try to build something we can parse later.
   */
  private String makeLocationString(LocationVO loc) {
    if (pars.simpleLocation) {
      return loc.getAddress();
    }

    StringBuffer sb = new StringBuffer(loc.getAddress());
    addNonNull(defaultDelim, Resources.SUBADDRESS, loc.getSubaddress(), sb);
    addNonNull(urlDelim, Resources.URL, loc.getLink(), sb);

    if (sb.length() == 0) {
      return null;
    }

    return sb.toString();
  }

  private String makeFieldStart(String delim, String resourceKey) {
    StringBuffer sb = new StringBuffer(delim);
    sb.append(svci.getResources().getString(resourceKey));
    sb.append(nameValSep);

    return sb.toString();
  }

  /** Try to build a location from the supplied string. This only works if
   * the delimiters and keywords remain intact.
   */
  private LocationVO makeLocation(String val) {
    if (val == null) {
      return null;
    }

    LocationVO loc = new LocationVO();

    if (pars.simpleLocation) {
      loc.setAddress(val);
      return loc;
    }

    /** The following is untested, unfinished and does not work.
     */
    int pos = 0;
    int fldStart = 0;

    int subAddressPos = val.indexOf(makeFieldStart(defaultDelim, Resources.SUBADDRESS));
    int urlPos = val.indexOf(makeFieldStart(urlDelim, Resources.URL));

    if ((subAddressPos < 0) && (urlPos < 0)) {
      loc.setAddress(val);
      return loc;
    }

    return loc;
  }

  /** Try to build a contact from the supplied string. This only works if
   * the delimiters and keywords remain intact.
   */
  private SponsorVO makeContact(String val) {
    if (val == null) {
      return null;
    }

    SponsorVO sp = new SponsorVO();

    if (pars.simpleContact) {
      sp.setName(val);
      return sp;
    }

    /** The following is untested, unfinished and does not work.
     */

    sp.setName(val);
    return sp;
  }

  /** Return the AltRep parameter if it exists
   */
  private AltRep getAltRep(Property prop) {
    return (AltRep)prop.getParameters().getParameter("ALTREP");
  }

  /** Make a date object that includes the date and time
   */
  private java.util.Date makeDateTime(Date dt, Time tm) {
    return new MyCalendarVO(dt, tm).getTime();
  }

  /** Get a sql timestamp object from a date-time */
  private Timestamp makeSqlTimestamp(java.util.Date dtTm) {
    return new Timestamp(dtTm.getTime());
  }

  /** Get a sql date object from a date-time */
  private Date makeSqlDate(java.util.Date dtTm) {
    return new Date(dtTm.getTime());
  }

  /** Get a sql time object from a date-time */
  private Time makeSqlTime(java.util.Date dtTm) {
    return new Time(dtTm.getTime());
  }

  private Logger getLog() {
    if (log != null) {
      return log;
    }

    log = Logger.getLogger(this.getClass());
    return log;
  }
}

