/* **********************************************************************
    Copyright 2003 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.uwcal.webcommon;

import edu.rpi.cct.uwcal.common.URIgen;

import edu.washington.cac.calfacade.shared.CalFacadeException;
import edu.washington.cac.calfacade.shared.LocationVO;
import edu.washington.cac.calfacade.shared.SponsorVO;

import java.net.URI;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/** This class defines methods used to return uris which will provide access
 * to entities located somewhere in the implementing calendar system.
 *
 * <p>For example, a call to getLocationURI(loc) might return something like<br/>
 *     "http://cal.myplace.edu/ucal/locations.do?id=1234"
 *
 * <p>Implementing classes will be used by services like the synch process
 * which needs to embed usable urls in the generated Icalendar objects.
 *
 * The generated
 *
 * @author Mike Douglass   douglm@rpi.edu
 */
public class UWCalWebURIgen implements URIgen {
  private String urlPrefix;

  private Pattern locidPattern = Pattern.compile("(.*locationId=)(\\d*)");
  private Pattern spidPattern = Pattern.compile("(.*sponsorId=)(\\d*)");

  /** Constructor - the prefix must point to a server application which
   * implements the requests built below. This application should run
   * unauthenticated or authenticated but in unauthenticated mode it will
   * only return public entities.
   *
   * @param urlPrefix   String value which will prefix all the urls.
   *                    Does NOT have trailing "/"
   */
  public UWCalWebURIgen(String urlPrefix) {
    this.urlPrefix = urlPrefix;
  }

  public URI getLocationURI(LocationVO val) throws CalFacadeException {
    try {
      return new URI(urlPrefix + "/location.do?locationId=" + val.getId());
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }
  }

  /** Attempt to create a dummy object representing the given URI.
   * Throw an exception if this is not a URI representing a location.
   *
   * <p>No web or network interactions need take place. - some implementations
   * may choose to return a real object.
   *
   * @param val     URI referencing a single location
   * @return LocationVO object with id filled in
   */
  public LocationVO getLocation(URI val) throws CalFacadeException {
    try {
      String query = val.getQuery();

      if (query == null) {
        throw new CalFacadeException("Not a location URI");
      }

      Matcher m = locidPattern.matcher(query);

      if (!m.matches()) {
        throw new CalFacadeException("Not a location URI");
      }

      Integer id = Integer.valueOf(m.group(2));

      LocationVO loc = new LocationVO();

      loc.setId(id.intValue());

      return loc;
    } catch (CalFacadeException cfe) {
      throw cfe;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }
  }

  public URI getSponsorURI(SponsorVO val) throws CalFacadeException {
    try {
      return new URI(urlPrefix + "/contact.do?sponsorId=" + val.getId());
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }
  }

  public SponsorVO getSponsor(URI val) throws CalFacadeException {
    try {
      String query = val.getQuery();

      if (query == null) {
        throw new CalFacadeException("Not a sponsor URI");
      }

      Matcher m = spidPattern.matcher(query);

      if (!m.matches()) {
        throw new CalFacadeException("Not a sponsor URI");
      }

      Integer id = Integer.valueOf(m.group(2));

      SponsorVO sp = new SponsorVO();

      sp.setId(id.intValue());

      return sp;
    } catch (CalFacadeException cfe) {
      throw cfe;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }
  }
}

