// java

package edu.washington.cac.calendar.filter;

import org.apache.regexp.RE;
import org.apache.regexp.RESyntaxException;

import edu.washington.cac.calendar.data.Event;
import edu.washington.cac.calendar.db.Field;
import edu.washington.cac.calendar.db.Fields;

/**
  A filter that selects events that match a regular expression in
  a particular field

  @author Greg Barnes
  @version 1.0
 */
public abstract class RegexpFilter extends Filter
{
  /** The Regular expression to match */
  private RE re;

  /** Factory method to create a new Regexp filter 
    @param re Regular expression to match
    @param field to match on
    @return new RegexpFilter object
    @exception IllegalArgumentException if the field is not supported yet
   */
  public static RegexpFilter create(RE re, Field field)
  {
    if (field.equals(Fields.LADDRESS)) {
      return new RegexpLAddressFilter(re);
    } else {
      throw new IllegalArgumentException("Field " + field + 
                                         " is not supported.");
    }
  }

  /** Factory method to create a new Regexp filter 
    @param pattern Pattern for the regular expression 
    @param field to match on
    @return new RegexpFilter object
    @exception IllegalArgumentException if the field is not supported yet
    @exception RESyntaxException if there's a problem with the pattern
   */
  public static RegexpFilter create(String pattern, Field field) 
      throws RESyntaxException
  {
    return create(new RE(pattern), field);
  }

  /**
    Create a new instance
    @param re Regular expression to match
   */
  protected RegexpFilter(RE re)
  {
    this.re = re;
  }

  /**
    Get the regular expression
    @return the regular expresion
   */
  protected RE getRe() 
  {
    return this.re;
  }

  /** 
    Does the filter selects the event?
    @param e Event to test
    @return If the filter selects the event
   */
  public abstract boolean match(Event e);

  /**
    Get a human-readable name for the calendar represented by the filter
    @return a human-readable name for the calendar represented by the filter
   */
  public abstract String getName();

  /**
    Get an internal name for the calendar represented by the filter
    @return an internal name for the calendar represented by the filter
   */
  public String getInternalName()
  {
    return "Regular expression: " + getName();
  }
}
