package edu.washington.cac.calfacade.impl;

import edu.washington.cac.calendar.MyCalendar;
import edu.washington.cac.calendar.data.CaldataException;
import edu.washington.cac.calendar.data.DuplicateItemException;
import edu.washington.cac.calendar.data.Event;
import edu.washington.cac.calendar.data.EventRef;
import edu.washington.cac.calendar.data.Events;
import edu.washington.cac.calendar.data.EventsI;
import edu.washington.cac.calendar.data.ItemException;
import edu.washington.cac.calendar.data.ItemAccessException;
import edu.washington.cac.calendar.data.Keyword;
import edu.washington.cac.calendar.data.Keywords;
import edu.washington.cac.calendar.data.Location;
import edu.washington.cac.calendar.data.Locations;
import edu.washington.cac.calendar.MyCalendar;
import edu.washington.cac.calendar.data.NoSuchItemException;
import edu.washington.cac.calendar.data.PersonalEvents;
import edu.washington.cac.calendar.data.PublicEvents;
import edu.washington.cac.calendar.data.Sponsor;
import edu.washington.cac.calendar.data.Sponsors;
import edu.washington.cac.calendar.data.User;
import edu.washington.cac.calendar.db.Caldata;
import edu.washington.cac.calendar.filter.Calendar;
import edu.washington.cac.calendar.filter.Filter;
import edu.washington.cac.calendar.filter.FilteredEvents;
import edu.washington.cac.calendar.filter.LastmodGEFilter;
import edu.washington.cac.calendar.filter.PublicCalendars;
import edu.washington.cac.calendar.filter.SearchFilter;
import edu.washington.cac.calendar.filter.SelectAllFilter;
import edu.washington.cac.calendar.MyCalendar;
import edu.washington.cac.calenv.CalEnv;
import edu.washington.cac.calfacade.shared.AdminGroups;
import edu.washington.cac.calfacade.shared.CalendarVO;
import edu.washington.cac.calfacade.shared.CalFacadeAccessException;
import edu.washington.cac.calfacade.shared.CalFacadeException;
import edu.washington.cac.calfacade.shared.Calintf;
import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.EventRefVO;
import edu.washington.cac.calfacade.shared.filter.FilterVO;
import edu.washington.cac.calfacade.shared.GuidVO;
import edu.washington.cac.calfacade.shared.KeywordVO;
import edu.washington.cac.calfacade.shared.LocationVO;
import edu.washington.cac.calfacade.shared.RecurVO;
import edu.washington.cac.calfacade.shared.SponsorVO;
import edu.washington.cac.calfacade.shared.SynchInfoVO;
import edu.washington.cac.calfacade.shared.SynchStateVO;
import edu.washington.cac.calfacade.shared.UserAuth;
import edu.washington.cac.calfacade.shared.UserVO;

import java.sql.Date;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Vector;
import org.apache.log4j.Logger;

/** All sql exceptions should be trapped by the real calendar api and
 * converted to calendar exceptions.
 */
import java.sql.SQLException;
import java.sql.Timestamp;

/** This class acts as an interface to the calendar.
 *
 * @author Mike Douglass   douglm@rpi.edu
 */
public class CalintfImpl implements Calintf {
  private boolean debug;

  /** Ensure we don't open while open
   */
  private boolean isOpen;

  /** Flag changes to one or the other - this won't work if we have write access
   * to other users data - however this shouldn't be needed with transparent caching
   */
  private boolean publicModified;
  private boolean personalModified;

  /** Real user
   */
  private User authUser;

  /** And the real user value object
   */
  private UserVO authUserVO;

  /** User for whom we maintain this facade
   */
  private User user;

  /** And the value object
   */
  private UserVO userVO;

  /** The user authorisation object
   */
  private UserAuth userAuth;

  /** The administrative groups object.
   */
  private AdminGroups adminGroups;

  /** True if we are administrators
   */
  private boolean publicAdmin;

  /** True if this is for synchronization
   */
  private boolean synch;

  /** Non-null if this is for synchronization. Identifies the client end.
   */
  private String synchId;

  private EventsI eventsCache;

  /** Current filter for public calendars
   */
  private Filter curFilter;

  /** These two are set from properties
   */
  private boolean adminCanEditAllPublicLocations;
  private boolean adminCanEditAllPublicSponsors;

  private int maxDbStringLength;

  /** The default filter for a guest */
  private static Filter defaultFilter = new SelectAllFilter();

  private SortedSet calendars;

  /** Current calendar for public calendars - null if no calendar (filter) set
   */
  private CalendarVO curCalendar;

  /** Current search string - null if no search (filter) set
   */
  private String curSearch;

  /** We cache keyword attrs here for the moment
   */
  private HashMap keyAttrs = new HashMap();

  private transient Logger log;

  /** When we were created for debugging */
  private Timestamp objTimestamp;

  /** ===================================================================
   *                   initialisation
   *  =================================================================== */

  /** Must be called to initialise the new object.
   *
   * @param authUser    String authenticated user of the application
   * @param rights      int access rights value as defined by
   *                    edu.washington.cac.calfacade.shared.UserAuth
   * @param user        String user we are acting as. If null we use authUser
   * @param isGuest     boolean true if user is guest
   * @param publicAdmin boolean true if this is a public events admin app
   * @param synchId     non-null if this is for synchronization. Identifies the
   *                    client end.
   * @param debug       boolean true to turn on debugging trace
   */
  public void init(String authUser,
                   int rights,
                   String user,
                   boolean isGuest,
                   boolean publicAdmin,
                   String synchId,
                   boolean debug) throws CalFacadeException {
    this.debug = debug;
    objTimestamp = new Timestamp(System.currentTimeMillis());

    if (synch && publicAdmin) {
      throw new CalFacadeException("synch only valid for non admin");
    }

    this.synch = synchId != null;
    this.synchId = synchId;
    this.publicAdmin = publicAdmin;

    if (user == null) {
      user = authUser;
    }

    if (debug) {
      getLogger().debug("CalIntfImpl.init - authUser=" + authUser +
                        " publicAdmin=" + publicAdmin +
                        " user=" + user +
                        " synch=" + synch);
    }

    open();

    try {
      if (isGuest) {
        this.user = new User();
        userVO = VOFactory.getUserVO(this.user);
      } else {
        this.authUser = new User(authUser);
        this.authUser.setRights(rights);
        this.authUser.setPublicEvents(publicAdmin);

        try {
          this.authUser.logon();
        } catch (Exception e) {
          throw new CalFacadeException("Unable to logon authUser " + authUser + ": " +
                              e.getMessage());
        }

        getLogger().info("Authenticated user " + authUser +
                         " logged on with rights " + rights);
        authUserVO = VOFactory.getUserVO(this.authUser);

        setUser(user);
      }

      try {
        CalEnv env = CalEnv.getEnv(debug);

        adminCanEditAllPublicLocations = isSuper() ||
            env.adminCanEditAllPublicLocations();

        adminCanEditAllPublicSponsors = isSuper() ||
            env.adminCanEditAllPublicSponsors();

        maxDbStringLength = CalEnv.getEnv().maxDbStringLength();
      } catch (Throwable t) {
        throw new CalFacadeException(t);
      }
    } finally {
      close();
    }
  }

  /** Switch to the given non-null user. The access rights are determined by
   * the previous call to init.
   *
   * @param val         String user id
   */
  public void setUser(String val) throws CalFacadeException {
    refreshEvents();

    user = new User(val);
//    user.setRights(rights);
    user.setPublicEvents(publicAdmin);

    try {
      user.logon();
    } catch (Exception e) {
      throw new CalFacadeException("Unable to set user " + val + ": " +
                          e.getMessage());
    }

    if (debug) {
      log.debug("User " + user + " set in calintf");
    }

    userVO = VOFactory.getUserVO(user);
  }

  /* ====================================================================
   *                   Misc methods
   * ==================================================================== */

  /** Signal the start of a sequence of operations. These overlap transactions
   * in that there may be 0 to many transactions started and ended within an
   * open/close call and many open/close calls within a transaction.
   *
   * <p>The open close methods are mainly associated with web style requests
   * and open will usually be called early in the incoming request handling
   * and close will always be called on the way out allowing the interface an
   * opportunity to reacquire (on open) and release (on close) any resources
   * such as connections.
   */
  public synchronized void open() throws CalFacadeException {
    if (debug) {
      log.debug("Open for " + objTimestamp);
    }

    if (isOpen) {
      throw new CalFacadeException("Already open");
    }

    isOpen = true;

    if (userAuth != null) {
      userAuth.reinitialise(this);
    }

    if (adminGroups != null) {
      adminGroups.init(this);
    }
  }

  /** Call on the way out after handling a request..
   */
  public synchronized void close() throws CalFacadeException {
    if (debug) {
      log.debug("Close for " + objTimestamp);
    }

    if (!isOpen) {
      return;
    }

    if (publicModified) {
      touchPublic();
    } else if (personalModified) {
      touchPersonal();
    }

    isOpen = false;
  }

  /** Start a (possibly long-running) transaction. In the web environment
   * this might do nothing. The endTransaction method should in some way
   * check version numbers to detect concurrent updates and fail with an
   * exception.
   */
  public void beginTransaction() throws CalFacadeException {
    checkOpen();
  }

  /** End a (possibly long-running) transaction. In the web environment
   * this should in some way check version numbers to detect concurrent updates
   * and fail with an exception.
   */
  public void endTransaction() throws CalFacadeException {
    checkOpen();
  }

  /** Call if there has been an error during an update process.
   */
  public void rollbackTransaction() throws CalFacadeException {
    checkOpen();
  }

  /** Returns a value object representing the current user.
   *
   * @return UserVO       representing the current user
   */
  public UserVO getUserVO() throws CalFacadeException {
    return userVO;
  }

  /** Add a user to the database
   *
   * @param user   UserVO object to add.
   */
  public void addUser(UserVO user) throws CalFacadeException {
    checkOpen();
    try {
      User u = new User(user.getAccount());

      int id = new Caldata().add(u, false);

      user.setId(id);
    } catch (SQLException se) {
      handleSQLException(se);
    }
  }

  /** Returns a value object representing the user with the given id.
   *
   * @param id            int id of user entry
   * @return UserVO       representing the current user
   */
  public UserVO getUser(int id) throws CalFacadeException {
    checkOpen();
    return getPubEvents().getUser(id);
  }

  /** Returns a value object representing the user with the given account name.
   *
   * @param account       string account name of user entry
   * @return UserVO       representing the current user
   */
  public UserVO getUser(String account) throws CalFacadeException {
    checkOpen();
    return getPubEvents().getUser(account);
  }

  /** Get a UserAuth object which allows the application to determine what
   * special rights the user has.
   *
   * <p>Note that the returned object may require a one time initialisation
   * call to set it up correctly.
   * @see edu.washington.cac.calfacade.shared.UserAuth#initialise(String, int)
   * @see edu.washington.cac.calfacade.shared.UserAuth#initialise(String, Object)
   *
   * @return UserAuth    implementation.
   */
  public UserAuth getUserAuth() throws CalFacadeException {
    checkOpen();
    if (userAuth != null) {
      userAuth.reinitialise(this);
      return userAuth;
    }

    try {
      String userAuthClassName = CalEnv.getEnv().getUserAuthClassName();

      if (userAuthClassName == null) {
        throw new CalFacadeException("No property defined in calenv for " +
                            "getUserAuthClassName()");
      }

      Object o = Class.forName(userAuthClassName).newInstance();

      if (o == null) {
        throw new CalFacadeException("Class " + userAuthClassName +
                            " not found");
      }

      if (!(o instanceof UserAuth)) {
        throw new CalFacadeException("Class defined in calenv for " +
                            "getUserAuthClassName() is not a subclass of " +
                            UserAuth.class.getName());
      }

      userAuth = (UserAuth)o;
    } catch (CalFacadeException ce) {
      throw ce;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }

    return userAuth;
  }

  /** Get an AdminGroups object which allows the application to handle
   * administrative groups.
   *
   * @return AdminGroups    implementation.
   */
  public AdminGroups getAdminGroups() throws CalFacadeException {
    checkOpen();
    if (adminGroups != null) {
      return adminGroups;
    }

    try {
      String adminGroupsClassName = CalEnv.getEnv().getAdminGroupsClassName();

      if (adminGroupsClassName == null) {
        throw new CalFacadeException("No property defined in calenv for " +
                            "getAdminGroupsClassName()");
      }

      Object o = Class.forName(adminGroupsClassName).newInstance();

      if (o == null) {
        throw new CalFacadeException("Class " + adminGroupsClassName +
                            " not found");
      }

      if (!(o instanceof AdminGroups)) {
        throw new CalFacadeException("Class defined in calenv for " +
                        "getAdminGroupsClassName() is not a subclass of " +
                        AdminGroups.class.getName());
      }

      adminGroups = (AdminGroups)o;
      adminGroups.init(this);
    } catch (CalFacadeException ce) {
      throw ce;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }

    return adminGroups;
  }

  /** Refresh the users events cache
   */
  public void refreshEvents() throws CalFacadeException {
    checkOpen();
    eventsCache = null;
    keyAttrs = new HashMap();
  }

  /* ====================================================================
   *                   Global parameters
   * ==================================================================== */

  /** Get the value of the current public events timestamp.
   *
   * @return long       current timestamp value
   */
  public long getPublicLastmod() throws CalFacadeException {
    checkOpen();
    return PublicEvents.getPublicEvents().getLastmod();
  }

  /** Get a name uniquely.identifying this system. This should take the form <br/>
   *   name@host
   * <br/>where<ul>
   * <li>name identifies the particular calendar system at the site</li>
   * <li>host part identifies the domain of the site.</li>..
   * </ul>
   *
   * @return String    globally unique system identifier.
   */
  public String getSysid() throws CalFacadeException {
    checkOpen();
    // **********************************************************
    // **********************************************************
    // **********************************************************
    //    fake it for the moment - this willcome out of some
    //    global app attributes.
    // **********************************************************
    // **********************************************************
    // **********************************************************
    // **********************************************************
    // **********************************************************
    return "calendar@mysite.edu";
  }

  /* ====================================================================
   *                   Calendars and search
   * ==================================================================== */

  /** Return a list of calendar we can present to the client as a list of
   * available calendars.
   *
   * @return Collection   of FilterVO sorted by name
   */
  public Collection getCalendars() throws CalFacadeException {
    checkOpen();
    if (calendars == null) {
      SortedSet cals = PublicCalendars.mainCalendars();

      calendars = VOFactory.getCalendarVOs(cals);
    }

    return calendars;
  }

  /** Set a search filter using the suppplied search string
   *
   * @param val    String search parameters
   */
  public void setSearch(String val) throws CalFacadeException {
    checkOpen();
    if (val == null) {
      curCalendar = null;
      curSearch = null;
      curFilter = defaultFilter;
    } else {
      curCalendar = null;
      curSearch = val;
      curFilter = new SearchFilter(val);
    }

    /* We can either do a setFilter if es is a FilteredEvents object
       or just drop the events object forcing a refresh
     */
    eventsCache = null; // force refresh
  }

  /** Return the current search string
   *
   * @return  String     search parameters
   */
  public String getSearch() {
    return curSearch;
  }

  /** Set the calendar we are interested in. This is represented by the name
   * of a calendar.
   *
   * @param  val     int id of calendar - 0 for none
   * @return boolean true for OK, false for unknown calendar
   */
  public boolean setCalendar(int val) throws CalFacadeException {
    checkOpen();
    if (val == 0) {
      curCalendar = null;
      curSearch = null;
      curFilter = defaultFilter;

      return true;
    }

    /** Try to locate the calendar
     */
    try {
      Calendar c = PublicCalendars.getPublicCalendars().getCalendar(val);
      curCalendar = VOFactory.getCalendarVO(c);
      curSearch = null;
      curFilter = c.getFilter();
    } catch (NoSuchItemException nsie) {
      curFilter = null;
    }

    if (curFilter == null) {
      curCalendar = null;
      curFilter = defaultFilter;

      return false;
    }

    /* We can either do a setFilter if es is a FilteredEvents object
       or just drop the events object forcing a refresh
     */
    eventsCache = null; // force refresh
    return true;
  }

  /** Get the name of the current calendar we are displaying
   *
   * @return String    name of calendar or null for all events
   */
  public String getCalendarTitle() throws CalFacadeException {
    checkOpen();
    if (curCalendar == null) {
      return null;
    }

    return curCalendar.getTitle();
  }

  /** This method is called to add a calendar. No consistency checks are
   * carried out at the moment.
   *
   * @param val  String values for db
   */
  public void addCalendarDef(String val) throws CalFacadeException {
    checkOpen();
    try {
      new Caldata().addCalendar(val);
    } catch (SQLException se) {
      handleSQLException(se);
    }

    updated(true);
  }

  /** Add a filter to the database. All references must be set.
   */
  public void addFilter(FilterVO val) throws CalFacadeException {
    checkOpen();

    if (val != null) {
      getPubEvents().addFilter(val);
    } else {
      // Flag done
      updated(true);
    }
  }

  /** Update a filter
   *
   * @param  val           FilterVO object to upate
   * @exception CalFacadeException If there's a problem
   */
  public void updateFilter(FilterVO val) throws CalFacadeException {
    checkOpen();
    throw new CalFacadeException("Unimplemented");
  }

  /* * Set a filter which returns events starting after a given Timestamp.
   *
   * @param val   Timestamp object.
   * /
  public void setLastmodGECalendar(Timestamp val) throws CalFacadeException {
    curFilter = new LastmodGEFilter(val);
    curSearch = curFilter.getName();

    /* We can either do a setFilter if es is a FilteredEvents object
       or just drop the events object forcing a refresh
     * /
    eventsCache = null; // force refresh
  }

  /* * Set a filter which returns events starting after a given date of the
   * form yyyymmdd.
   *
   * @param date   String of form yyyymmdd.
   * /
  public void setLastmodGEDateCalendar(String date) throws CalFacadeException {
    if ((date == null) || (date.length() != 8)) {
      throw new CalFacadeException("Invalid lastmod date " + date);
    }

    String tsdate = date.substring(0, 4) + "-" +
                    date.substring(4, 6) + "-" +
                    date.substring(6) +
                    " 00:00:00.000000000";
    setLastmodGECalendar(Timestamp.valueOf(tsdate));
  }

  /* * Set a filter which returns events starting after a given date and time
   * of the form yyyymmddhhmm.
   *
   * @param dateTime   String of form yyyymmddhhmm.
   * /
  public void setLastmodGEDateTimeCalendar(String dateTime)
      throws CalFacadeException {
    if ((dateTime == null) || (dateTime.length() != 12)) {
      throw new CalFacadeException("Invalid lastmod date " + dateTime);
    }

    String tsdate = dateTime.substring(0, 4) + "-" +
                    dateTime.substring(4, 6) + "-" +
                    dateTime.substring(6, 8) + " " +
                    dateTime.substring(8, 10) + ":" +
                    dateTime.substring(10) +
                    ":00.000000000";
    setLastmodGECalendar(Timestamp.valueOf(tsdate));
  }
  */

  /** ===================================================================
   *                   Subscriptions
   *  =================================================================== */

  /** Ensure that the users subscriptions match a set of calendars
   *
   * @param crs      TreeSet of Integer calendar ids
   * @return int[]   Number of calendars subscribed to, and number of
   *                 calendars unsubscribed to
   */
  public int[] fixSubscriptions(TreeSet crs) throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();

    if (!(es instanceof PersonalEvents)) {
      throw new CalFacadeException("Not a personal events calendar");
    }

    try {
      return ((PersonalEvents)es).fixSubscriptions(crs);
    } catch (ItemAccessException iae) {
      throw new CalFacadeException(iae);
    } catch (SQLException se) {
      handleSQLException(se);
      return null;
    }
  }

  /** Determine if user is subscribed to a calendar
   *
   * @param id       int calendar id
   * @return boolean true if subscribed to calendars
   */
  public boolean getSubscribed(int id) throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();

    if (!(es instanceof PersonalEvents)) {
      throw new CalFacadeException("Not a personal events calendar");
    }

    try {
      return ((PersonalEvents)es).containsSubscription(id);
    } catch (NoSuchItemException nsie) {
      return false;
    }
  }

  /** Get the subscriptions for the user
   *
   * @param user        UserVO object
   */
  public void getSubscriptions(UserVO user) throws CalFacadeException {
    checkOpen();
    Collection subs = getPubEvents().getSubscriptions(user);

    user.setSubscriptions(subs);
  }

  /** ===================================================================
   *                   Keywords
   *  =================================================================== */

  /** Return all keywords for this user. They will have their attributes set.
   * For a public admin client this is the same as getPublicKeywords().
   *
   * @return Collection  of keywords value objects
   */
  public Collection getKeywords() throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();

    Keywords ks = es.getKeywords();
    Iterator it = ks.sortedElements();

    Vector v = new Vector();

    while (it.hasNext()) {
      v.addElement(it.next());
    }

    Collection keys = VOFactory.getKeywordVOs(v);
    if (keys == null) {
      return null;
    }

    it = keys.iterator();
    while (it.hasNext()) {
      KeywordVO kw = (KeywordVO)it.next();

      kw.setAttributes(getKeywordAttrs(kw.getId()));
    }

    return keys;
  }

  /** Add a Keyword to the database. The id will be set in the parameter
   * object.
   *
   * @param val   KeywordVO object to be added
   */
  public void addKeyword(KeywordVO val) throws CalFacadeException {
    checkOpen();
//    EventsI es = getEventsI();
    Keyword k = null;

    requireSuper();

    if (val.getCreator() == null) {
      val.setCreator(userVO);
    }

    try {
      k = CalFactory.getKeyword(val);
//      k.setCaldata(new Caldata());
//      es.getKeywords().add(k);
      k.add(k.getCreator());

    } catch (SQLException se) {
      handleSQLException(se);
    }

    int id = k.getId();
    val.setId(id);

    updated(val.isPublic());
  }

  /** Add a Keyword to the database and give it the assigned key.
   * Only for restores
   *
   * @param val   KeywordVO object to be added with key set
   */
  public void addKeywordWithKey(KeywordVO val) throws CalFacadeException {
    checkOpen();
    throw new CalFacadeException("Unimplemented");
  }

  /** Find a Keyword with the given id
   *
   * @param id     int id of the Keyword
   * @return KeywordVO object representing the location in question
   *                     null if it doesn't exist.
   */
  public KeywordVO getKeyword(int id) throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();

    try {
      Keyword k = es.getKeywords().getKeyword(id);

      KeywordVO kw = VOFactory.getKeywordVO(k);
      if (kw == null) {
        return null;
      }

      kw.setAttributes(getKeywordAttrs(kw.getId()));

      return kw;
    } catch (NoSuchItemException nsie) {
      return null;
    }
  }

  /** Check to see if a Keyword is referenced.
   *
   * @param val      KeywordVO object to check
   * @return boolean true if the Keyword is referenced somewhere
   */
  public boolean checkKeywordRefs(KeywordVO val) throws CalFacadeException {
    checkOpen();
    return getPubEvents().keywordReffed(val.getId());
  }

  /** ===================================================================
   *                   Public Keywords
   *  =================================================================== */

  /** Return all public keywords
   *
   * @return Collection  of keywords value objects
   */
  public Collection getPublicKeywords() throws CalFacadeException {
    checkOpen();
    //Keywords ks = PublicEvents.getPublicKeywords(user);
    // Until access control works in back end
    checkPublicAccess(readAccess);
    Keywords ks = PublicEvents.getPublicKeywords(/*user*/);
    Iterator it = ks.sortedElements();

    Vector v = new Vector();

    while (it.hasNext()) {
      v.addElement(it.next());
    }

    Collection keys = VOFactory.getKeywordVOs(v);
    if (keys == null) {
      return null;
    }

    it = keys.iterator();
    while (it.hasNext()) {
      KeywordVO kw = (KeywordVO)it.next();

      kw.setAttributes(getKeywordAttrs(kw.getId()));
    }

    return keys;
  }

  /** Return given public keyword
   *
   * @param id     int id of the keyword
   * @return KeywordVO[]  array of keywords value objects
   */
  public KeywordVO getPublicKeyword(int id) throws CalFacadeException {
    checkOpen();
    //Keywords ks = PublicEvents.getPublicKeywords(user);
    // Until access control works in back end
    checkPublicAccess(readAccess);
    Keywords ks = PublicEvents.getPublicKeywords(/*user*/);

    try {
      Keyword k = ks.getKeyword(id);

      KeywordVO kw = VOFactory.getKeywordVO(k);
      if (kw == null) {
        return null;
      }

      kw.setAttributes(getKeywordAttrs(kw.getId()));

      return kw;
    } catch (NoSuchItemException nsie) {
      return null;
    }
  }

  /** Replace a keyword in the database.
   *
   * @param val   KeywordVO object to be replaced
   */
  public void replaceKeyword(KeywordVO val) throws CalFacadeException {
    checkOpen();
//    EventsI es = getEventsI();
    requireSuper();

    try {
      Keyword k = CalFactory.getKeyword(val);

      k.replace(k.getCreator());
    } catch (SQLException se) {
      handleSQLException(se);
    }

    updated(val.isPublic());
  }

  /** Delete the given keyword
   *
   * @param val      KeywordVO object to be deleted
   * @return boolean false if it didn't exist, true if it was deleted.
   */
  public boolean deleteKeyword(KeywordVO val) throws CalFacadeException {
    checkOpen();
    requireSuper();

    EventsI es = getEventsI();

    try {
      Keyword k = es.getKeywords().getKeyword(val.getId());
      k.delete(k.getCreator());

      // Delete any attributes
      getPubEvents().deleteKeywordAttrs(val.getId());
    } catch (SQLException se) {
      handleSQLException(se);
      return false;  // never reached
    } catch (NoSuchItemException nsie) {
      return false;
    }

    updated(val.isPublic());

    return true;
  }

  /* ====================================================================
   *                   Keyword attributes
   * ==================================================================== */

  /** Add a keyword attribute name/value pair for a keyword
   *
   * @param val             KeywordVO object to update
   * @param name            String attribute name
   * @param aval            String attribute value
   */
  public void addKeywordAttr(KeywordVO val,
                             String name,
                             String aval) throws CalFacadeException {
    checkOpen();
    requireSuper();
    getPubEvents().addKeywordAttr(val.getId(), name, aval);

    updated(val.isPublic());
  }

  /** Remove a keyword attribute name/value pair for a keyword
   *
   * @param val             KeywordVO object to update
   * @param name            String attribute name
   * @param aval            String attribute value
   */
  public void removeKeywordAttr(KeywordVO val,
                                String name,
                                String aval) throws CalFacadeException {
    checkOpen();
    requireSuper();
    getPubEvents().removeKeywordAttr(val.getId(), name, aval);

    updated(val.isPublic());
  }

  /* ====================================================================
   *                   Locations
   * ==================================================================== */

  /** Return all locations for this user. Returns an empty collection for
   * no locations. If this is a public admin instance, this will return all
   * public locations owned by the current user. Otherwise it returns all
   * personal locations owned by the user.
   *
   * @return Collection        of location value objects
   */
  public Collection getLocations() throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();

    Collection locs;

    if (!publicAdmin) {
      Locations ls = es.getLocations();
      Iterator it = ls.sortedElements();
      locs = new Vector();

      /* It appears that the public locations get dragged in as well -
          the inclusion of filtered events seems to include all public events.
          Remove them here
       */

      while (it.hasNext()) {
        Location l = (Location)it.next();

        if (!l.isPublic() || (l.getId() < Location.DELETED_LOCATION_ID)) {
          locs.add(l);
        }
      }

      return VOFactory.getLocationVOs(locs);
    }

    if (adminCanEditAllPublicLocations) {
      // return all
      return getPublicLocations();
    }

    Collection c = getPublicLocations();
    locs = new Vector();

    Iterator it = c.iterator();
    while (it.hasNext()) {
      LocationVO loc = (LocationVO)it.next();

      if (loc.getCreator().equals(userVO)) {
        locs.add(loc);
      }
    }

    return locs;
  }

  /** Add a location to the database. The id will be set in the parameter
   * object.
   *
   * @param val   LocationVO object to be added
   */
  public void addLocation(LocationVO val) throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();
    Location l = null;

    if (val.getCreator() == null) {
      val.setCreator(userVO);
    }

    try {
      l = CalFactory.getLocation(val);
      l.add(user);
/* This method leaves us with a location wwith the guest owner

      es.getLocations().add(l);*/
//    } catch (ItemAccessException iae) {
//      throw new CalFacadeException(iae);
    } catch (SQLException se) {
      handleSQLException(se);
    }

    int id = l.getId();
    val.setId(id);

    updated(val.isPublic());
  }

  /** Replace a location in the database.
   *
   * @param val   LocationVO object to be replaced
   */
  public void replaceLocation(LocationVO val) throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();

    try {
      Location l = CalFactory.getLocation(val);

      l.replace(l.getCreator());
    } catch (SQLException se) {
      handleSQLException(se);
    }

    updated(val.isPublic());
  }

  /** Find a location with the given id
   *
   * @param id     int id of the location
   * @return LocationVo object representing the location in question
   *                     null if it doesn't exist.
   */
  public LocationVO getLocation(int id) throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();

    try {
      Location l = es.getLocations().getLocation(id);
      return VOFactory.getLocationVO(l);
    } catch (NoSuchItemException nsie) {
      return null;
    }
  }

  /** Delete a location
   *
   * @param val      LocationVO object to be deleted
   * @return boolean false if it didn't exist, true if it was deleted.
   */
  public boolean deleteLocation(LocationVO val) throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();

    try {
      Locations ls = es.getLocations();
      Location l = ls.getLocation(val.getId());
      if (l.isPublic()) {
        return false;
      }
      ls.delete(l);
    } catch (SQLException se) {
      handleSQLException(se);
    } catch (ItemException ie) {
      return false;
    }

    updated(val.isPublic());

    return true;
  }

  /** Check to see if a location is referenced.
   *
   * @param val      LocationVO object to be checked
   * @return boolean true if the location is referenced somewhere
   */
  public boolean checkLocationRefs(LocationVO val) throws CalFacadeException {
    checkOpen();
    return getPubEvents().locationReffed(val.getId());
  }

  /** ===================================================================
   *                   Public Locations
   *  =================================================================== */

  /** Return public locations - excluding reserved entries. Returns an empty
   * collection for no locations.
   *
   * @return Collection        of location value objects
   */
  public Collection getPublicLocations() throws CalFacadeException {
    checkOpen();
    return getPublicLocations(false);
  }

  /** Return public locations - optionally including reserved entries.
   * Returns an empty collection for no locations.
   *
   * @param includeReserved boolean true include reserved, false exclude
   * @return Collection        of location value objects
   */
  public Collection getPublicLocations(boolean includeReserved)
      throws CalFacadeException {
    checkOpen();
    //Locations ls = PublicEvents.getPublicLocations(user);
    // Until access control works in back end
    checkPublicAccess(readAccess);
    Locations ls = PublicEvents.getPublicLocations(/*user*/);
    Iterator it = ls.sortedElements();

    Vector v = new Vector();

    while (it.hasNext()) {
      Location l = (Location)it.next();

      if (includeReserved || (l.getId() > LocationVO.maxReservedId)) {
        v.addElement(l);
      }
    }

    return VOFactory.getLocationVOs(v);
  }

  /** Return given public location
   *
   * @param id          int id of the location
   * @return LocationVo object representing the location in question
   *                     null if it doesn't exist.
   */
  public LocationVO getPublicLocation(int id) throws CalFacadeException {
    checkOpen();
    //Locations ls = PublicEvents.getPublicLocations(user);
    // Until access control works in back end
    checkPublicAccess(readAccess);
    Locations ls = PublicEvents.getPublicLocations(/*user*/);

    try {
      Location l = ls.getLocation(id);
      return VOFactory.getLocationVO(l);
    } catch (NoSuchItemException nsie) {
      return null;
    }
  }

  /* * Delete given public location
   *
   * @param val      LocationVO object to be deleted
   * @return boolean false if it didn't exist, true if it was deleted.
   * /
  public boolean deletePublicLocation(LocationVO val) throws CalFacadeException {
    //Locations ls = PublicEvents.getPublicLocations(user);
    // Until access control works in back end
    checkPublicAccess(writeAccess);
    Locations ls = PublicEvents.getPublicLocations(/*user* /);

    try {
      Location l = ls.getLocation(val.getId());
      l.delete(l.getCreator());
    } catch (SQLException se) {
      handleSQLException(se);
    } catch (NoSuchItemException nsie) {
      return false;
    }

    return true;
  }*/

  /** ===================================================================
   *                   Sponsors
   *  =================================================================== */

  /** Return all sponsors for this user. Returns an empty collection for
   * no sponsors. If this is a public admin instance, this will return all
   * public sponsors owned by the current user. Otherwise it returns all
   * personal sponsors owned by the user.
   *
   * <p>(Note the current implementation does not allow forthe creation of
   * personal sponsors.
   *
   * @return Collection     of sponsors
   */
  public Collection getSponsors() throws CalFacadeException {
    checkOpen();
    if (!publicAdmin) {
      throw new CalFacadeException("Unimplemented");
    }

    if (adminCanEditAllPublicSponsors) {
      // return all
      return getPublicSponsors();
    }

    Collection c = getPublicSponsors();
    Vector sps = new Vector();

    Iterator it = c.iterator();
    while (it.hasNext()) {
      SponsorVO sp = (SponsorVO)it.next();

      if (sp.getCreator().equals(userVO)) {
        sps.add(sp);
      }
    }

    return sps;
  }

  /** Find a sponsor with the given id
   *
   * @param id     int id of the sponsor
   * @return SponsorVo object representing the sponsor in question
   *                     null if it doesn't exist.
   */
  public SponsorVO getSponsor(int id) throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();

    try {
      Sponsor s = es.getSponsors().getSponsor(id);
      return VOFactory.getSponsorVO(s);
    } catch (NoSuchItemException nsie) {
      return null;
    }
  }

  /** Add a sponsor to the database. The id will be set in the parameter
   * object. Only for public events.
   *
   * @param val   SponsorVO object to be added
   */
  public void addSponsor(SponsorVO val) throws CalFacadeException {
    checkOpen();
    checkPublicAccess(writeAccess);
    EventsI es = getEventsI();
    Sponsor s = null;

    if (val.getCreator() == null) {
      val.setCreator(userVO);
    }

    try {
      s = CalFactory.getSponsor(val);
      s.add(user);
/* This method leaves us with a location wwith the guest owner

      es.getSponsors().add(s);*/
//    } catch (ItemAccessException iae) {
//      throw new CalFacadeException(iae);
    } catch (SQLException se) {
      handleSQLException(se);
    }

    int id = s.getId();
    val.setId(id);

    updated(val.isPublic());
  }

  /** Replace a sponsor in the database.
   *
   * @param val   SponsorVO object to be replaced
   */
  public void replaceSponsor(SponsorVO val) throws CalFacadeException {
    checkOpen();
    checkPublicAccess(writeAccess);
    EventsI es = getEventsI();

    try {
      Sponsor s = CalFactory.getSponsor(val);

      s.replace(s.getCreator());
    } catch (SQLException se) {
      handleSQLException(se);
    }

    updated(val.isPublic());
  }

  /** Delete a sponsor
   *
   * @param val      SponsorVO object to be deleted
   * @return boolean false if it didn't exist, true if it was deleted.
   */
  public boolean deleteSponsor(SponsorVO val) throws CalFacadeException {
    checkOpen();
    checkPublicAccess(writeAccess);
    EventsI es = getEventsI();

    try {
      Sponsor s = es.getSponsors().getSponsor(val.getId());
      s.delete(s.getCreator());
    } catch (SQLException se) {
      handleSQLException(se);
    } catch (NoSuchItemException nsie) {
      return false;
    }

    updated(val.isPublic());

    return true;
  }

  /** Check to see if a sponsor is referenced.
   *
   * @param val      SponsorVO object to be checked
   * @return boolean true if the sponsor is referenced somewhere
   */
  public boolean checkSponsorRefs(SponsorVO val) throws CalFacadeException {
    checkOpen();
    return getPubEvents().sponsorReffed(val.getId());
  }

  /* ====================================================================
   *                   Public Sponsors
   * ==================================================================== */

  /** Return public sponsors - excluding reserved entries. Returns an empty
   * collection for no sponsors.
   *
   * @return Collection        of sponsor value objects
   */
  public Collection getPublicSponsors() throws CalFacadeException {
    checkOpen();
    return getPublicSponsors(false);
  }

  /** Return all sponsors - optionally including reserved entries.
   * Returns an empty collection for no sponsors.
   *
   * @param includeReserved boolean true include reserved, false exclude
   * @return Collection     of sponsors
   */
  public Collection getPublicSponsors(boolean includeReserved)
      throws CalFacadeException {
    checkOpen();
    //Sponsors ss = PublicEvents.getPublicSponsors(user);
    // Until access control works in back end
    checkPublicAccess(readAccess);
    Sponsors ss = PublicEvents.getPublicSponsors(/*user*/);
    Iterator it = ss.sortedElements();

    Vector v = new Vector();

    while (it.hasNext()) {
      Sponsor sp = (Sponsor)it.next();

      if (includeReserved || (sp.getId() > SponsorVO.maxReservedId)) {
        v.addElement(sp);
      }
    }

    return VOFactory.getSponsorVOs(v);
  }

  /** Return given public sponsor
   *
   * @param id         int id of the sponsor
   * @return SponsorVo object representing the sponsor in question
   *                     null if it doesn't exist.
   */
  public SponsorVO getPublicSponsor(int id) throws CalFacadeException {
    checkOpen();
    //Sponsors ss = PublicEvents.getPublicSponsors(user);
    // Until access control works in back end
    checkPublicAccess(readAccess);
    Sponsors ss = PublicEvents.getPublicSponsors(/*user*/);

    try {
      Sponsor s = ss.getSponsor(id);
      return VOFactory.getSponsorVO(s);
    } catch (NoSuchItemException nsie) {
      return null;
    }
  }

  /* ====================================================================
   *                   Events
   * ==================================================================== */

  /** Return a single event for the current user.
   *
   * @param   eventId   int id of the event
   * @return  EventVO   value object representing event.
   */
  public EventVO getEvent(int eventId) throws CalFacadeException {
    checkOpen();
    if (eventId < 0) {
      throw new CalFacadeException("Bad eventid " + eventId);
    }

    EventsI es = getEventsI();

    try {
      Event e;

      if (es instanceof PersonalEvents) {
         /* Try to extract it from the current cache of events. If it's not
            there try it as a public event. There are plenty of places we want
            to retrieve a publci event before it's in the users cache and in
            any case, we may want to display a publci event without adding it.
          */
        try {
          e = (Event)es.get(eventId);
        } catch (NoSuchItemException dummy) {
          try {
            e = PublicEvents.getPublicEvents().getEvent(eventId, true);
          } catch (NoSuchItemException nsie) {
            throw new CalFacadeException(nsie);
          }
        }
      } else {
        e = es.getEvent(eventId, true);
      }

      checkAccess(readAccess, e);

      return VOFactory.getEventVO(e, es, this, userVO.getAccount());
    } catch (NoSuchItemException nsie) {
      return null;
    }
  }

  /** Return filtered events for the current user.
   *
   * @return Collection of event value objects
   */
  public Collection getEvents() throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();

    if (es instanceof PersonalEvents) {
      try {
        ((PersonalEvents)es).loadAll();
      } catch (ItemException ie) {
        throw new CalFacadeException(ie);
      } catch (SQLException se) {
        handleSQLException(se);
      }
    }

    Collection evs = es.values();

    Collection evvos = VOFactory.getEventVOs(evs, es, this, userVO.getAccount());
//    } catch (ItemException iae) {
//      handleUnserializableException(iae);
//    }

    return evvos;
  }

  /** Return the events for the given day as an array of value objects
   *
   * @param   date    MyCalendar object defining day
   * @return  Collection  one days events, empty for no events.
   */
  public Collection getDaysEvents(java.util.Date date) throws CalFacadeException {
//  public Collection getDaysEvents(MyCalendarVO date) throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();

    try {
//      Event[] evs = es.oneDaysEvents(CalFactory.getMyCalendar(date));
      Event[] evs = es.oneDaysEvents(new MyCalendar(date));
      // change events object???
      return VOFactory.getEventVOs(evs, es, this, userVO.getAccount());
    } catch (ItemException iae) {
      handleUnserializableException(iae);
    } catch (SQLException se) {
      handleSQLException(se);
    }

    return new Vector();
  }

  /** Add an event to the database. The id and uid will be set in the parameter
   * object.
   *
   * @param val   EventVO object to be added
     @exception CalFacadeException If there's a db problem or problem with
         the event
   */
  public void addEvent(EventVO val) throws CalFacadeException {
    checkOpen();
    try {
      EventsI es = getEventsI();

      requireNotGuest();

      if (val.getCreator() == null) {
        val.setCreator(userVO);
      }

      Event e = CalFactory.getEvent(val);

      /* We can't use PublicEvents yet to create events - the user gets set
         to null by the calls to getPublicKeywords etc. However, we can't give
         them a user either as that stops personal calendars working.
       */
      e.add(e.getCreator());

      int id = e.getId();
      val.setId(id);

      /* Force a load of the event
       */
      es.getEvent(id, true);
//      es.add(e);

      assignGuid(val);
    } catch (SQLException se) {
      handleSQLException(se);
//    } catch (ItemException ie) {
//      handleUnserializableException(ie);
//      return 0;	// never reached
    } catch (NoSuchItemException nsie) {
      handleUnserializableException(nsie);
    }

    updated(val.isPublic());
  }

  /** Update an event in the database.
   *
   * @param val   EventVO object to be replaced
   * @exception CalFacadeException If there's a db problem or problem with
   *     the event
   */
  public void updateEvent(EventVO val) throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();

    try {
      Event e = CalFactory.getEvent(val);

      checkAccess(writeAccess, e);

      /** Can't use public events here for the same reason as adds
       */
      e.replace(user);
//      es.replace(e);


/* **********************************************************
    Was this needed?
      refreshEvents();
*/

//    } catch (ItemException ie) {
//      handleUnserializableException(ie);
    } catch (SQLException se) {
      handleSQLException(se);
    }

    updated(val.isPublic());
  }

  /** Fix up the event keywords after an update or addition.
   *
   * @param val         updated EventVO object
   * @param oldEvent    EventVO object before changes applied. null for a
   *                    new event.
   */
  public void fixKeywords(EventVO val,
                          EventVO oldEvent) throws CalFacadeException {
    checkOpen();
    /* If keywords changed, we need to change event_keywords entry
     */

    Iterator it = val.iterateKeywords();
    int eid = val.getId();

    while (it.hasNext()) {
      KeywordVO k = (KeywordVO)it.next();

      if ((oldEvent == null) || (!oldEvent.hasKeyword(k))) {
        // Added the category
        getPubEvents().addEventKeyword(eid, k.getId());
      }
    }

    if (oldEvent != null) {
      it = oldEvent.iterateKeywords();

      while (it.hasNext()) {
        KeywordVO k = (KeywordVO)it.next();

        if (!val.hasKeyword(k)) {
          // Deleted the category
          getPubEvents().deleteEventKeyword(eid, k.getId());
        }
      }
    }
  }

  /** Delete an event. Set any referring synch states to deleted.
   *
   * @param val      EventVO object to be deleted
   * @return boolean false if it didn't exist, true if it was deleted.
   * @exception CalFacadeException If there's a database access problem
   */
  public boolean deleteEvent(EventVO val) throws CalFacadeException {
    checkOpen();
    try {
      EventsI es = getEventsI();
      Event e = es.getEvent(val.getId(), true);

      if (e == null) {
        return false;
      }

      checkAccess(writeAccess, e);
      if (es instanceof PublicEvents) {
        /* Currently deletes don't work with that class -
           same problem as above */
        getPubEvents().deleteEvent(val.getId());
        getPubEvents().deleteEventKeywords(val.getId());
      } else {
        es.delete(e);
      }

      /* This event was really deleted so we need to set any synch states to
       * indicate this is the case.
       */
      setSynchState(val, SynchStateVO.DELETED);
    } catch (ItemAccessException e) {
      handleUnserializableException(e);
    } catch (ItemException e) {
      return false;
    } catch (SQLException se) {
      handleSQLException(se);
    }

    updated(val.isPublic());

    return true;
  }

  /** Assign a guid to an event which must exist in the db. A noop if this
   * event already has a guid.
   *
   * @param val      EventVO object
   */
  public void assignGuid(EventVO val) throws CalFacadeException {
    checkOpen();
    if (val == null) {
      return;
    }

    if (val.getGuid() != null) {
      return;
    }

    int id = val.getId();

    if (id < 0) {
      throw new CalFacadeException("Attemptto assign guid to non-existant event.");
    }

    val.setGuid(new GuidVO(id, getSysid()).toString());
  }

  /* ====================================================================
   *                   Public Events
   * ==================================================================== */

  /** Return the public events for the current user. This is an administrative
   * function for the public events admin client.
   *
   * @return Collection  populated event value objects
   */
  public Collection findPublicEvents() throws CalFacadeException {
    checkOpen();
    //EventsI es = PublicEvents.getPublicEvents(user);
    // Until access control works in back end
    if ((!publicAdmin) || (user.getName() == null)) {
      throw new CalFacadeAccessException();
    }

    EventsI es = PublicEvents.getPublicEvents();

    Collection evs = getPubEvents().findPublicEvents(user.getName(), null, null);
    return VOFactory.getEventVOs(evs, es, this, userVO.getAccount());
  }

  /** Return the public events for the current user within the given date
   * range. This is an administrative function for the public events admin
   * client.
   *
   * @param startDate    java.util.Date start date - may be null
   * @param endDate      java.util.Date end of period - may be null.
   * @return Collection  populated event value objects
   */
  public Collection findPublicEvents(java.util.Date startDate,
                                     java.util.Date endDate)
          throws CalFacadeException {
    checkOpen();
    //EventsI es = PublicEvents.getPublicEvents(user);
    // Until access control works in back end
    if ((!publicAdmin) || (user.getName() == null)) {
      throw new CalFacadeAccessException();
    }

    EventsI es = PublicEvents.getPublicEvents();

    Date sdt = null;
    Date edt = null;

    if (startDate != null) {
      sdt = new Date(startDate.getTime());
    }

    if (endDate != null) {
      edt = new Date(endDate.getTime());
    }

    Collection evs = getPubEvents().findPublicEvents(user.getName(), sdt, edt);
    return VOFactory.getEventVOs(evs, es, this, userVO.getAccount());
  }

  /* * Return the public events for the given day as an array of value objects
   *
   * @param   date    MyCalendar object defining day
   * @return  EventVO[]  one days events or null for no events.
   * /
  public EventVO[] getDaysPublicEvents(MyCalendarVO date)
      throws CalFacadeException {
    //EventsI es = PublicEvents.getPublicEvents(user);
    // Until access control works in back end
    EventsI es = PublicEvents.getPublicEvents();

    try {
      Event[] evs = es.oneDaysEvents(CalFactory.getMyCalendar(date));
      EventVO[] evos = VOFactory.getEventVOs(evs, es);

      return evos;
    } catch (ItemException iae) {
      handleUnserializableException(iae);
    } catch (SQLException se) {
      handleSQLException(se);
    }

    return null;
  }*/

  /* * ===================================================================
   *                   Event keywords
   *  =================================================================== * /

  /* * Given an event and keyword add an entry to the event keyword table
   *
   * @param event       EventVo object
   * @param keyword     KeywordVO object
   * @exception CalFacadeException If there's a database access problem
   * /
  public void addEventKeyword(EventVO event, KeywordVO keyword)
      throws CalFacadeException {
    getPubEvents().addEventKeyword(event.getId(), keyword.getId());
  }

  /* * Given an event and keywordid delete an entry from the event keyword table
   *
   * @param event       EventVo object
   * @param keyword     KeywordVO object
   * @exception CalFacadeException If there's a database access problem
   * /
  public void deleteEventKeyword(EventVO event, KeywordVO keyword)
      throws CalFacadeException {
    getPubEvents().deleteEventKeyword(event.getId(), keyword.getId());
  }*/

  /* ====================================================================
   *                   EventRefs
   * ==================================================================== */

  /** Create an event ref for the current user.
   *
   * @param  event      EventVO object
   * @return  boolean   false for no such item.
   */
  public boolean addEventRef(EventVO event) throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();
    if (!(es instanceof PersonalEvents)) {
      // Watch for attempts to fool public client
      return false;
    }

    EventRefVO ervo = new EventRefVO(event, userVO);

    EventRef er = CalFactory.getEventRef(ervo);

    try {
//      PublicEvents pe = PublicEvents.getPublicEvents(user);

//      Event ev = pe.getEvent(eventId, true);
      ((PersonalEvents)es).addRef(er, null); //new MyCalendar(ev.getStartdate()));
    } catch (ItemAccessException iae) {
      throw new CalFacadeException(iae);
    } catch (NoSuchItemException e) {
      return false;
    } catch (CaldataException cde) {
      handleUnserializableException(cde);
    } catch (SQLException se) {
      if (isConstraintException(se)) {
        // seems likely this was caused by duplicating an item
        throw new CalFacadeException(se.getMessage());
      }
      handleSQLException(se);
    }

    updated(false);

    return true;
  }

  /** Delete an event ref for the current user.
   *
   * @param  event          EventVO object
   * @return  boolean       false for no such item.
   */
  public boolean deleteEventRef(EventVO event) throws CalFacadeException {
    checkOpen();
    EventsI es = getEventsI();

    EventRefVO ervo = new EventRefVO(event, userVO);

    EventRef er = CalFactory.getEventRef(ervo);

    try {
      ((PersonalEvents)es).deleteRef(er);

      /* This looks like an event was really deleted so we need to set any
       * synch states only for this user to indicate this is the case.
       */
      setSynchState(userVO, event, SynchStateVO.DELETED);
    } catch (SQLException se) {
      handleSQLException(se);
      return false;
    } catch (NoSuchItemException e) {
      return false;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }

    updated(false);

    return true;
  }

  /** Delete a subscribed event for the current user.
   *
   * <p>This is a partially implemented feature which currently only changes
   * the synch state. What we need is in effect a reverse event ref - I guess
   * a new purpose, which indicates the event is hidden for this user.
   *
   * @param  event          EventVO object
   * @return  boolean       false for no such item.
   */
  public boolean deleteSubscribedEvent(EventVO event) throws CalFacadeException {
    checkOpen();
//    EventsI es = getEventsI();

    try {

      /* For this user and device it looks like an event was really deleted
       * so we need to set that synch state to CLIENT_DELETED. For any other
       * synch states for this user set it to CLIENT_DELETED_UNDELIVERED so
       * that other devices will see the deletion once  only.
       */
      setSynchState(userVO, event, SynchStateVO.CLIENT_DELETED_UNDELIVERED);
      setSynchState(userVO, synchId, event, SynchStateVO.CLIENT_DELETED);
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    }

    updated(false);

    return true;
  }

  /* ====================================================================
   *                   Synchronization
   * ==================================================================== */

  /** Get the synch info for the current device and user.
   *
   * @return SynchInfoVO object or null.
   */
  public SynchInfoVO getSynchInfo() throws CalFacadeException {
    checkOpen();
    return getPubEvents().getSynchInfo(userVO, synchId);
   }

  /** Add the synch info for the current device and user.
   *
   * @param val    SynchInfoVO object to add
   */
  public void addSynchInfo(SynchInfoVO val) throws CalFacadeException {
    checkOpen();
    getPubEvents().addSynchInfo(val);
   }

  /** Update the synch info for the current device and user. There must be
   * exactly one object to update.
   *
   * @param val    SynchInfoVO object to update
   */
  public void updateSynchInfo(SynchInfoVO val) throws CalFacadeException {
    checkOpen();
    getPubEvents().updateSynchInfo(val);
   }

  /** Return synchronization state for the given event in the current
   * synchronization context as defined by the initialisation parameters,
   * that is the user and synchId. If the event has never been involved in a
   * synch process there will be no related synch state object.
   *
   * @param ev            EventVO object .
   * @return SynchStateVO object or null.
   */
  public SynchStateVO getSynchState(EventVO ev)
      throws CalFacadeException {
    checkOpen();
    return getPubEvents().getSynchState(ev, userVO, synchId);
  }

  /** Get the deleted synch states for the current user
   *
   * @returns Collection of synch state objects (possibly empty)
   */
  public Collection getDeletedSynchStates() throws CalFacadeException {
    checkOpen();
    return getPubEvents().getDeletedSynchStates(userVO, synchId);
  }

  /** Add the synch state for the given event and user.
   *
   * @param val      SynchStateVO object.
   */
  public void addSynchState(SynchStateVO val)
      throws CalFacadeException {
    checkOpen();
    getPubEvents().addSynchState(val);
  }

  /** Update the synch state for the given event and user. There must be
   * exactly one object to update.
   *
   * @param val    SynchStateVO object with associated event and user
   */
  public void updateSynchState(SynchStateVO val)
      throws CalFacadeException {
    checkOpen();
    getPubEvents().updateSynchState(val);
  }

  /** Update the synch state for the given event. There may be many objects
   * to update.
   *
   * @param ev      EventVO representing associated event
   * @param state   int new state - defined in SynchStateVO
   * @return int    number of updated entries.
   */
  public int setSynchState(EventVO ev, int state)
        throws CalFacadeException {
    checkOpen();
    return getPubEvents().setSynchState(ev, state);
  }

  /** Get the synch data associated with the given object..
   *
   * @param val    SynchStateVO object
   */
  public void getSynchData(SynchStateVO val) throws CalFacadeException {
    checkOpen();
    getPubEvents().getSynchData(val);
  }

  /** Set the synch state and data for the given user and event.
   * There may be zero to many objects to update.
   *
   * @param val    SynchStateVO object with associated event and user
   */
  public void setSynchData(SynchStateVO val) throws CalFacadeException {
    checkOpen();
    getPubEvents().setSynchData(val);
  }

  /** Set the synch state for the given user and event. There may be zero to many
   * objects to update.
   *
   * @param user    UserVO object
   * @param ev      EventVO representing associated event
   * @param state   int new state - defined in SynchStateVO
   * @return int    number of updated entries.
   */
  public int setSynchState(UserVO user, EventVO ev, int state)
        throws CalFacadeException {
    checkOpen();
    return getPubEvents().setSynchState(user, ev, state);
  }

  /** Set the synch state for the given user, synchId and event. There must
   * be one and only one object to update.
   *
   * @param user    UserVO object
   * @param synchId String synch id
   * @param ev      EventVO representing associated event
   * @param state   int new state - defined in SynchStateVO
   */
  public void setSynchState(UserVO user, String synchId, EventVO ev, int state)
        throws CalFacadeException {
    checkOpen();
    getPubEvents().setSynchState(user, synchId, ev, state);
  }

  /** Update the synch states for the current user and device.
   *
   * <p>This is called at the termination of synchronization to set any state
   * objects appropriately.
   *
   * <p><table cols="2">
   * <th><td>Old state</td><td>New state</td></th>
   * <tr><td>UNKNOWN</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>SYNCHRONIZED</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>NEW</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>MODIFIED</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>DELETED</td><td>entry deleted</td></tr>
   * <tr><td>CLIENT_DELETED</td><td>unchanged</td></tr>
   * <tr><td>CLIENT_DELETED_UNDELIVERED</td><td>CLIENT_DELETED</td></tr>
   * <tr><td>CLIENT_MODIFIED</td><td>unchanged</td></tr>
   * </table>
   */
  public void updateSynchStates() throws CalFacadeException {
    checkOpen();
    getPubEvents().removeDeletedSynchStates(userVO, synchId);
    getPubEvents().updateSynchStates(userVO, synchId);
  }

  /* ====================================================================
   *                       Recurrences
   * This is currently under development at RPI
   * ==================================================================== */

  /** Add a recurrence to the database.
   *
   * @param val      RecurVO object.
   */
  public void addRecur(RecurVO val) throws CalFacadeException {
    checkOpen();
    getPubEvents().addRecur(val);
  }

  /** Get a recurrence associated with an event.
   *
   * @param ev      Associated EventVO
   * @return RecurVO  recurrence object or null
   */
  public RecurVO getRecur(EventVO ev) throws CalFacadeException {
    checkOpen();
    return getPubEvents().getRecur(ev);
  }

  /** Remove a recurrence associated with an event.
   *
   * @param ev      Associated EventVO
   */
  public void deleteRecur(EventVO ev) throws CalFacadeException {
    checkOpen();
    getPubEvents().deleteRecur(ev);
  }

  /** Update a recurrence to the database.
   *
   * @param val      RecurVO object.
   */
  public void updateRecur(RecurVO val) throws CalFacadeException {
    checkOpen();
    getPubEvents().updateRecur(val);
  }

  /* ====================================================================
                       Private methods
     ==================================================================== */

  /** Update lastmod timestamp.
   *
   * @exception CalFacadeException If there's a database access problem
   */
  private void touchPublic() throws CalFacadeException {
    EventsI es = getEventsI();

    if ((es != null) && (es instanceof PublicEvents)) {
      ((PublicEvents)es).forceFlush();
    }
    getPubEvents().touch();
    refreshEvents();
  }

  /** Update lastmod timestamp.
   *
   * @exception CalFacadeException If there's a database access problem
   */
  private void touchPersonal() throws CalFacadeException {
    refreshEvents();
  }

  private void checkOpen() throws CalFacadeException {
    if (!isOpen) {
      throw new CalFacadeException("Calintf call when closed");
    }
  }

  /** At the moment we have to force a flush and reread at almost every
   * public modification as entries are not making their way into the
   * cache. Breaking personal and public apart should help.
   */
  private void updated(boolean publicEntity) throws CalFacadeException {
    if (publicEntity) {
//      publicModified = true;
      touchPublic();
    } else {
      personalModified = true;

      // Force a reread for the moment
//      refreshEvents();
    }
  }

  /** Get the keyword attributes entry for a keyword id
   * For the moment we'll cache them here
   *
   * @param keywordid       int id of the keyword entry
   * @return Set            of keyword attributes
   */
  private Set getKeywordAttrs(int keywordid) throws CalFacadeException {
    Integer ikey = new Integer(keywordid);

    Set ka = (Set)keyAttrs.get(ikey);

    if (ka != null) {
      return ka;
    }

    ka = getPubEvents().getKeywordAttrs(keywordid);
    keyAttrs.put(ikey, ka);

    return ka;
  }

  /** Start of an SQL state that indicates an integrity contraint was
     violated.

     This should really be in the calendar api - we should not see SQL
     exceptions at the presentation level.
   */
  /* Note:  I cannot find the X/Open SQL spec on which these codes
     are based; I base this seemingly arbitrary number on an IBM
     list of SQLstate error codes, which you can find at:
     http://www.redbooks.ibm.com/tstudio/db2_400/CLI/sqlmsg/allhtml.htm

     Note that other lists of codes I could find (i.e., Informix's
     cursory list) at least agree that if it starts
     with '23', it's an 'integrity constraint violation'.
   */
  private static final String INTEGRITY_PREFIX = "23";

  /**
    @return Is the given exception one that indicates an integrity
    constraint was violated?
    @param e The exception in question
   */
  private boolean isConstraintException(SQLException e) {
    return e.getSQLState() != null &&
           e.getSQLState().startsWith(INTEGRITY_PREFIX);
  }

  private PubEvents pubEvents;

  private PubEvents getPubEvents() {
    if (pubEvents == null) {
      pubEvents = new PubEvents(debug);
    }

    return pubEvents;
  }

  private EventsI getEventsI() {
    try {
      if (eventsCache == null) {
        if (publicAdmin) {
          //eventsCache = PublicEvents.getPublicEvents(user);
          // Until access control works in back end
          eventsCache = PublicEvents.getPublicEvents();

          if (debug) {
            getLogger().debug("Got PublicEvents object for " + user);
          }
        } else if (user.isGuest()) {
          if (curFilter == null) {
            curFilter = defaultFilter;
          }
          eventsCache = new FilteredEvents(curFilter);

          if (debug) {
            getLogger().debug("Got FilteredEvents object");
          }
        } else {
          eventsCache = new PersonalEvents(user);
          ((PersonalEvents)eventsCache).loadAll();

          if (debug) {
            getLogger().debug("Got PersonalEvents object for " + user);
          }
        }
      }

      return eventsCache;
    } catch (Throwable t) {
      getLogger().error(t);
      return null;
    }
  }

  private static final int readAccess = 0;
  private static final int writeAccess = 1;

  /** This does limited access checking to bar access to public objects
   * other than events.
   * If the current user is not a public admin user we throw an exception.
   *
   * @param access      int desired access (read/write)
   */
  private void checkPublicAccess(int access) throws CalFacadeException {
    if (access == readAccess) {
      return;
    }

    if (publicAdmin) {
      return;
    }

    throw new CalFacadeAccessException();
  }

  /** This does access checking for events.
   *
   * @param access      int desired access (read/write)
   * @param e           Event object
   */
  private void checkAccess(int access, Event e) throws CalFacadeException {
    if ((authUserVO != null) && authUserVO.isSuperUser()) {
      // Nobody can stop us - BWAAA HAA HAA
      return;
    }

    String creator = null;
    if (e.getCreator() != null) {
     creator = e.getCreator().getName();
    }
    checkAccess(access, e.isPublic(), creator);
  }

  /** This does the access check for all objects.
   *
   * @param access      int desired access (read/write)
   * @param publicObj   true for a public object
   * @param creator     String creator of object
   */
  private void checkAccess(int access, boolean publicObj, String creator)
          throws CalFacadeException {
    boolean reading = access == readAccess;

    if (user.isGuest()) {
      if (reading && publicObj) {
        return;
      }

      throw new CalFacadeAccessException();
    }

    if (publicAdmin) {
      if (!publicObj) {
        //We probably want to allow read access by superuser
        throw new CalFacadeAccessException();
      }

      if (reading) {
        // read anything public in admin
        return;
      }
    } else {
      if (publicObj) {
        if (reading) {
          // Why not - it's public
          return;
        }

        throw new CalFacadeAccessException();
      }
    }

    // Have to be owner
    if (userVO.getAccount().equals(creator))  {
      return;
    }

    throw new CalFacadeAccessException();
  }

  /** Ensure the current user has super user access.
   */
  private void requireSuper() throws CalFacadeException {
    if (publicAdmin && (authUserVO != null) && authUserVO.isSuperUser()) {
      return;
    }

    throw new CalFacadeAccessException();
  }

  /** Ensure the current user is not a guest.
   */
  private void requireNotGuest() throws CalFacadeException {
    if (!user.isGuest())  {
      return;
    }

    throw new CalFacadeAccessException();
  }

  /** See if current authorised user has super user access.
   */
  private boolean isSuper() throws CalFacadeException {
    if (publicAdmin && (authUserVO != null) && authUserVO.isSuperUser()) {
      return true;
    }

    return false;
  }

  private void handleSQLException(SQLException se) throws CalFacadeException {
    /* SQLExceptions are supposed to be Serializable, but
       Mike D. believes SQLExceptions are not ALWAYS serializable.
       Not sure what that means, but I imagine it causes problems
       for JBoss and/or Tomcat sometimes.
     */
    /* XXX May want to just make this call handleUnserializableException, if
       the only problem with SQLExceptions is their unserializable nature
     */
    throw new CalFacadeException(se.getMessage());
  }

  /**
    Handle an unserializable exception
    @param e The unserializable exception
   */
  private void handleUnserializableException(Exception e)
      throws CalFacadeException {
    getLogger().error(e);
    throw new CalFacadeException(e.getMessage());
  }

  /** Get a logger for messages
   */
  private Logger getLogger() {
    if (log == null) {
      log = Logger.getLogger(this.getClass());
    }

    return log;
  }
}

