package edu.washington.cac.calfacade.impl;

import edu.washington.cac.calendar.data.Event;
import edu.washington.cac.calendar.data.Recurrence;
import edu.washington.cac.calendar.data.User;
import edu.washington.cac.calendar.db.DBResources;
import edu.washington.cac.calendar.db.Fields;
import edu.washington.cac.calendar.db.PubEventsQueries;
import edu.washington.cac.calendar.db.Queries;
import edu.washington.cac.calendar.db.Tables;
import edu.washington.cac.calendar.db.WhereClauses;
import edu.washington.cac.calfacade.impl.CalFactory;
import edu.washington.cac.calfacade.shared.CalFacadeException;
import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.filter.AndFilterVO;
import edu.washington.cac.calfacade.shared.filter.CreatorFilterVO;
import edu.washington.cac.calfacade.shared.filter.FilterVO;
import edu.washington.cac.calfacade.shared.filter.KeywordFilterVO;
import edu.washington.cac.calfacade.shared.filter.LocationFilterVO;
import edu.washington.cac.calfacade.shared.filter.NotFilterVO;
import edu.washington.cac.calfacade.shared.filter.OrFilterVO;
import edu.washington.cac.calfacade.shared.filter.SponsorFilterVO;
import edu.washington.cac.calfacade.shared.IntArrayList;
import edu.washington.cac.calfacade.shared.KeywordVO;
import edu.washington.cac.calfacade.shared.KeywordAttrVO;
import edu.washington.cac.calfacade.shared.RecurRuleVO;
import edu.washington.cac.calfacade.shared.RecurVO;
import edu.washington.cac.calfacade.shared.SynchDataVO;
import edu.washington.cac.calfacade.shared.SynchInfoVO;
import edu.washington.cac.calfacade.shared.SynchStateVO;
import edu.washington.cac.calfacade.shared.UserVO;

import java.sql.Date;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;
import org.apache.log4j.Logger;

/** PubEvents:
 *  Object to handle sql statments to add/update/delete events in db tables.
 *
 * <p>I have gathered together the queries found in the pubevents modules
 * and intend combining them.
 *
 * <p>My intent is to simplify the process. The query rate is unlikely to
 * be high enough to stress the system so not getting all the columns
 * is probably an unneccessary complication.
 *
 * <p>Also, attempting to retrieve everything with a single query probably
 * causes portability problems.
 *
 * @author Leman Chung, Greg Barnes
 * @author Mike Douglass douglm@rpi.edu
 * @version 2.2
 */
public class PubEvents extends EventData {
  /** Ideally this would trigger the debugging log in the underlying jdbc
   * implementation.
   */
  private boolean debug = false;

  private transient Logger log;

  public PubEvents() {
    this(false);
  }

  public PubEvents(boolean debug) {
    super();
    this.debug = debug;
  }

  /* ====================================================================
   *                   Touch - update lastmods table
   * ==================================================================== */

  /** Update lastmod timestamp.
   *
   * @exception CalFacadeException If there's a database access problem
   */
  public void touch() throws CalFacadeException {
    PreparedStatement s = null;

    try {
      String sql = "update lastmods set lastmod = ?" +
                      "where name =?";

      s = getPreparedStatement(sql);
      s.setTimestamp(1,
                     new java.sql.Timestamp(new java.util.Date().getTime()));
      s.setString(2, "pubevents");

      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /* ====================================================================
   *                            Filters
   * ==================================================================== */

  /** Add a filter to the database. All references must be set.
   */
  public void addFilter(FilterVO val) throws CalFacadeException {
    /* Set this according to class of filter.
     */
    char typeFlag;
    int ref = 0;

    if (val instanceof AndFilterVO) {
      typeFlag = 'A';
    } else if (val instanceof KeywordFilterVO) {
      typeFlag = 'K';

      ref = ((KeywordFilterVO)val).getKeyword().getId();
    } else if (val instanceof CreatorFilterVO) {
      typeFlag = 'C';

      ref = ((CreatorFilterVO)val).getCreator().getId();
    } else if (val instanceof LocationFilterVO) {
      typeFlag = 'L';

      ref = ((LocationFilterVO)val).getLocation().getId();
    } else if (val instanceof OrFilterVO) {
      typeFlag = 'O';
    } else if (val instanceof SponsorFilterVO) {
      typeFlag = 'S';

      ref = ((SponsorFilterVO)val).getSponsor().getId();
    } else if (val instanceof NotFilterVO) {
      typeFlag = 'N';

      ref = ((NotFilterVO)val).getFilter().getId();
    } else {
      throw new CalFacadeException("Unimplemented filter " + val);
    }

    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
            "INSERT INTO calendars " +
              "(calendarid, type, name, ref_num, parent) " +
              "VALUES (?,?,?,?,?)");

      s.setInt(1, val.getId());
      s.setString(2, String.valueOf(typeFlag));
      s.setString(3, val.getName());
//      s.setString(4, val.getTitle());
//      s.setString(5, val.getDescription());
      s.setInt(4, ref);
      if (val.getParent() == null) {
        s.setInt(5, 0);
      } else {
        s.setInt(5, val.getParent().getId());
      }

      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /* ====================================================================
   *                            Sponsors
   * ==================================================================== */

  /** SQL to check that an event exists with a particular sponsor */
  private static final String SELECT_SPONSOR_FROM_EVENTS =
      "select " + Fields.ESPONSOR_ID.getSelectName() + " from "
      + Tables.EVENTS.getName() + " where "
      + Fields.ESPONSOR_ID.getStmtName() + "=?";

  /** SQL to check that a filter exists with a particular sponsor */
  private static final String SELECT_SPONSOR_FROM_CALENDARS =
      "select REF_NUM from "
      + Tables.CALENDARS.getName() + " where "
      + "TYPE='S' AND REF_NUM=?";

  /** See if the given sponsor id is referenced
   *
   *  @param  sponsorid   int sponsor id we are looking for
   *  @return boolean     true if found
   *  @exception CalFacadeException If there's a database access problem
   */
  public boolean sponsorReffed(int sponsorid) throws CalFacadeException {
      PreparedStatement s = null;
      ResultSet rs = null;

    try {
      s = getPreparedStatement(SELECT_SPONSOR_FROM_EVENTS);
      s.setInt(1, sponsorid);
      rs = s.executeQuery();

      if (rs.next()) {
        return true;
      }

      rs.close();
      close();

      s = getPreparedStatement(SELECT_SPONSOR_FROM_CALENDARS);
      s.setInt(1, sponsorid);
      rs = s.executeQuery();

      return rs.next();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close(); // Will free s
      s = null;
    }
  }

  /* ====================================================================
   *                            Locations
   * ==================================================================== */

  /** SQL to check whether a location is still used by any event */
  private static final String SELECT_LOCATION_FROM_EVENTS =
      "select " + Fields.ELOCATION_ID.getSelectName() + " from "
      + Tables.EVENTS.getName() + " where "
      + Fields.ELOCATION_ID.getStmtName() + "=?";

  /** SQL to check that a filter exists with a particular sponsor */
  private static final String SELECT_LOCATION_FROM_CALENDARS =
      "select REF_NUM from "
      + Tables.CALENDARS.getName() + " where "
      + "TYPE='L' AND REF_NUM=?";

  /** See if the given location id exists in events
   *
   *  @param  locationid   int location id we are looking for
   *  @return boolean     true if found
   *  @exception CalFacadeException If there's a database access problem
   */
  public boolean locationReffed(int locationid) throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement(SELECT_LOCATION_FROM_EVENTS);
      s.setInt(1, locationid);
      rs = s.executeQuery();

      if (rs.next()) {
        return true;
      }

      rs.close();
      close();

      s = getPreparedStatement(SELECT_LOCATION_FROM_CALENDARS);
      s.setInt(1, locationid);
      rs = s.executeQuery();

      return rs.next();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close(); // Will free s
      s = null;
    }
  }

  /* ====================================================================
   *                            Events
   * ==================================================================== */

  /** This should come from PubEventsQueries I guess. This needs to have
   * all the fields in an EventVO class.
   */
  private static final String eventQuerySelectFields =
                          Fields.EEVENT_ID.getSelectName() + ", "
                        + Fields.ESTARTDATE.getSelectName() + ", "
                        + Fields.ESTARTTIME.getSelectName() + ", "
                        + Fields.EENDDATE.getSelectName() + ", "
                        + Fields.EENDTIME.getSelectName() + ", "
                        + Fields.ESHORTDESC.getSelectName() + ", "
                        + Fields.ELONGDESC.getSelectName() + ", "
                            + Fields.ELONGDESC1.getSelectName() + ", "
                         + Fields.EPUBLIC.getSelectName() + ", "
                         + Fields.ELINK.getSelectName() + ", "
                         + Fields.ECOST.getSelectName() + ", "
                         + Fields.ESPONSOR_ID.getSelectName() + ", "
                         + Fields.ELOCATION_ID.getSelectName() + ", "
                         + Fields.ESEQ.getSelectName() + ", "
                         + Fields.ECREATOR.getSelectName() + ", "
                         + Fields.ERECURRING_STATUS.getSelectName() + ", "
                            + Fields.ELASTMOD.getSelectName() + ", "
                            + Fields.ECREATED.getSelectName();

  private static final String populateEventQueryPrefix =
                            "select " + eventQuerySelectFields +
                            " from " +
                            Tables.EVENTS.getName() + " where " +
                            Fields.EPUBLIC.getStmtName() + "='T' and " +
                            Fields.ECREATOR.getStmtName() + "=? ";

  private static final String populateEventQueryCheckRangeStartDate =
                            " and ( " +
                            Fields.ESTARTDATE.getStmtName() + ">=? or " +
                            Fields.EENDDATE.getStmtName() + ">=? ) ";

  private static final String populateEventQueryCheckRangeEndDate =
                            " and ( " +
                            Fields.ESTARTDATE.getStmtName() + "<=? or " +
                            Fields.EENDDATE.getStmtName() + "<=? ) ";

  private static final String populateEventQueryCheckRangeStartEndDate =
                            " and ( " +
                            "( " +
                            Fields.ESTARTDATE.getStmtName() + ">=? AND " +
                            Fields.ESTARTDATE.getStmtName() + "<=? ) or ( " +
                            Fields.EENDDATE.getStmtName() + ">=? ) and " +
                            Fields.EENDDATE.getStmtName() + "<=? ) ) ";

  private static final String populateEventStartdateOrder =
                            " order by " +
                            Fields.ESTARTDATE.getStmtName();

  /** Return an array of public events for the given creator, optionally
   *  within the given date range.
   *
   * <p>Either or both of the date range may be null.
   *
   * @param creator       if non-null must match
   * @param rangeStart    Date - optional start date
   * @param rangeEnd      Date - optional end date
   * @return Collection   event objects
   * @exception CalFacadeException If there's a database access problem
   */
  public Collection findPublicEvents(String creator,
                                  Date rangeStart,
                                  Date rangeEnd) throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      StringBuffer sql = new StringBuffer(populateEventQueryPrefix);

      if ((rangeStart != null) && (rangeEnd == null)) {
        sql.append(populateEventQueryCheckRangeStartDate);
      } else if ((rangeStart == null) && (rangeEnd != null)) {
        sql.append(populateEventQueryCheckRangeEndDate);
      } else if ((rangeStart != null) && (rangeEnd != null)) {
        sql.append(populateEventQueryCheckRangeStartEndDate);
      }

      sql.append(populateEventStartdateOrder);

      if (debug) {
        System.out.println("About to execute query:");
        System.out.println(sql.toString());
      }

      s = getPreparedStatement(sql.toString());

      s.setString(1, creator);

      if ((rangeStart != null) && (rangeEnd == null)) {
        s.setDate(2, rangeStart);
        s.setDate(3, rangeStart);
      } else if ((rangeStart == null) && (rangeEnd != null)) {
        s.setDate(2, rangeEnd);
        s.setDate(3, rangeEnd);
      } else if ((rangeStart != null) && (rangeEnd != null)) {
        s.setDate(2, rangeStart);
        s.setDate(3, rangeStart);
        s.setDate(4, rangeEnd);
        s.setDate(5, rangeEnd);
      }

      rs = s.executeQuery();
      Vector v = new Vector();

      while (rs.next()) {
        Event ev = new Event(rs.getInt(Fields.EEVENT_ID.getName()),
            rs.getString(Fields.ESHORTDESC.getName()),
            rs.getString(Fields.ELONGDESC.getName()),
            // rs.getString(Fields.ELONGDESC1.getName()),
            rs.getDate(Fields.ESTARTDATE.getName()),
            rs.getTime(Fields.ESTARTTIME.getName()),
            rs.getDate(Fields.EENDDATE.getName()),
            rs.getTime(Fields.EENDTIME.getName()),
            rs.getString(Fields.EPUBLIC.getName()).equals(WhereClauses.TRUE),
            rs.getString(Fields.ELINK.getName()),
            new User(rs.getString(Fields.ECREATOR.getName())),
            rs.getString(Fields.ECOST.getName()),
            Recurrence.create(rs.getString(Fields.ERECURRING_STATUS.getName()),
                              rs.getInt(Fields.EEVENT_ID.getName())),
            rs.getInt(Fields.ELOCATION_ID.getName()),
            rs.getInt(Fields.ESPONSOR_ID.getName()),
            rs.getInt(Fields.ESEQ.getName()),
            rs.getTimestamp(Fields.ELASTMOD.getName()),
            rs.getTimestamp(Fields.ECREATED.getName()));

        v.addElement(ev);
      }

      close();

      /* Now populate with keyword ids
       */

      Iterator it = v.iterator();
      while (it.hasNext()) {
        Event ev = (Event)it.next();

        getEventKeywordIds(ev);
      }

      return v;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close();
    }
  }

  /** SQL to delete an event */
  private static final String DELETE_EVENTS =
      "delete from " + Tables.EVENTS + " where " +
      Fields.EEVENT_ID.getInsertName() + "=?";

  /** Delete an event
   *
   * @param eventid   int id of event to delete
   * @exception CalFacadeException If there's a database access problem
   */
  public void deleteEvent(int eventid) throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(DELETE_EVENTS);
      s.setInt(1, eventid);
      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /* ====================================================================
   *                            Marked Events
   * This apparently is on its way out.
   * ==================================================================== */

  /** Query to see if event is marked.
  */
  private static final String checkMarked =
      "select " +
         Fields.MEVENT_ID.getSelectName() +
      " from " +
         Tables.MARKEDEVENTS.getName() +
      " where " +
         Fields.MEVENT_ID.getSelectName() + "=?";

  /** Determine if an event exists in the marked table
   *
   *  @param   eventid    ID of the event
   *  @return  boolean    True if it's there
   *  @exception CalFacadeException If there's a database access problem
   */
  public boolean isMarkedEvent(int eventid) throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement(checkMarked);
      s.setInt(1, eventid);
      rs = s.executeQuery();

      return rs.next();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close(); // Will free s
      s = null;
    }
  }

  /* ====================================================================
   *                       Keyword attributes
   * ==================================================================== */

  /** Get the keyword attributes entry for a keyword id
   *
   * @param keywordid       int id of the keyword entry
   * @return Set            of KeywordAttrVO
   */
  public Set getKeywordAttrs(int keywordid) throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement("select " +
                               "kaid, " +
                               "kakeywordid, " +
                               "kaname, " +
                               "kaval " +
                               "from keyword_attrs where " +
                               "kakeywordid =?");

      s.setInt(1, keywordid);
      rs = s.executeQuery();

      TreeSet ts = new TreeSet();

      while (rs.next()) {
        ts.add(new KeywordAttrVO(rs.getInt("kaid"), rs.getString("kaname"),
                      rs.getString("kaval")));
      }

      return ts;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close();
    }
  }

  /** Delete all keyword attrs entries for a keyword
   *
   * @param keywordid         int Id of the keyword entry
   */
  public void deleteKeywordAttrs(int keywordid) throws CalFacadeException {
    String sql = "delete from " +
                 "keyword_attrs " +
                 " where " +
                 "kakeywordid" + "=?";

    try {
      PreparedStatement s = getPreparedStatement(sql);
      s.setInt(1, keywordid);
      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Add a keyword attribute name/value pair for a keyword
   *
   * @param keywordid       int id of the keyword entry
   * @param name            String attribute name
   * @param val             String attribute value
   */
  public void addKeywordAttr(int keywordid,
                             String name,
                             String val) throws CalFacadeException {
    String findsql = "select kaid from " +
                        "keyword_attrs " +
                        " where " +
                        "kakeywordid" + "=?" + " and " +
                        "kaname" + "=?" + " and " +
                        "kaval" + "=?";

    String insertsql = "insert into keyword_attrs " +
                       "(kakeywordid, kaname, kaval) " +
                        " values " +
                        "(?, ?, ?)";

    ResultSet rs = null;

    try {
      PreparedStatement s = getPreparedStatement(findsql);
      s.setInt(1, keywordid);
      s.setString(2, name);
      s.setString(3, val);

      rs = s.executeQuery();
      if (rs.next()) {
        // Already exists
        return;
      }

      close();

      s = getPreparedStatement(insertsql);
      s.setInt(1, keywordid);
      s.setString(2, name);
      s.setString(3, val);

      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close();
    }
  }

  /** Remove a keyword attribute name/value pair for a keyword
   *
   * @param keywordid       int id of the keyword entry
   * @param name            String attribute name
   * @param val             String attribute value
   */
  public void removeKeywordAttr(int keywordid,
                                String name,
                                String val) throws CalFacadeException {
    String deletesql = "delete from " +
                        "keyword_attrs " +
                        " where " +
                        "kakeywordid" + "=?" + " and " +
                        "kaname" + "=?" + " and " +
                        "kaval" + "=?";
    try {
      PreparedStatement s = getPreparedStatement(deletesql);
      s.setInt(1, keywordid);
      s.setString(2, name);
      s.setString(3, val);

      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /* ====================================================================
   *                            Keywords
   * ==================================================================== */

  /** SQL to check whether a keyword is still used by any event */
  private static final String SELECT_KEYWORD_FROM_EVENTS =
      "select " + Fields.EKKEYWORD_ID.getSelectName() + " from "
      + Tables.EVENTKEYWORDS.getName() + " where "
      + Fields.EKKEYWORD_ID.getStmtName() + "=?";

  /** SQL to check that a filter exists with a particular keyword */
  private static final String SELECT_KEYWORD_FROM_CALENDARS =
      "select REF_NUM from "
      + Tables.CALENDARS.getName() + " where "
      + "TYPE='K' AND REF_NUM=?";

  /** See if the given keyword id exists in events
   *
   *  @param  id          int keyword id we are looking for
   *  @return boolean     true if found
   *  @exception CalFacadeException If there's a database access problem
   */
  public boolean keywordReffed(int id) throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement(SELECT_KEYWORD_FROM_EVENTS);
      s.setInt(1, id);
      rs = s.executeQuery();

      if (rs.next()) {
        return true;
      }

      rs.close();
      close();

      s = getPreparedStatement(SELECT_KEYWORD_FROM_CALENDARS);
      s.setInt(1, id);
      rs = s.executeQuery();

      return rs.next();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close(); // Will free s
      s = null;
    }
  }

  /* ====================================================================
   *                       Event keywords
   * ==================================================================== */

  /** SQL to add an entry into the event_keywords table */
  private static final String ADD_EVENT_KEYWORD =
      "insert into " + Tables.EVENTKEYWORDS + " (" +
      Fields.EKEVENT_ID.getInsertName() + ", " +
      Fields.EKKEYWORD_ID.getInsertName() + ") values (?, ?)";

  /** Given a eventid and keywordid add an entry to the event keyword table
   *
   * @param eventid   Id of event
   * @param keywordid Id of keyword
   * @exception CalFacadeException If there's a database access problem
   */
  public void addEventKeyword(int eventid, int keywordid)
      throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(ADD_EVENT_KEYWORD);
      s.setInt(1, eventid);
      s.setInt(2, keywordid);
      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** SQL to delete an entry in the event_keywords table */
  private static final String DELETE_EK =
      "delete from " + Tables.EVENTKEYWORDS + " where " +
      Fields.EKEVENT_ID.getInsertName() + "=? and " +
      Fields.EKKEYWORD_ID.getInsertName() + "=?";

  /** Given a eventid and a keyword id delete an entry from the event keyword table
   *
   * @param eventid   Id of event
   * @param keywordid Id of keyword
   * @exception CalFacadeException If there's a database access problem
   */
  public void deleteEventKeyword(int eventid, int keywordid)
      throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(DELETE_EK);
      s.setInt(1, eventid);
      s.setInt(2, keywordid);
      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** SQL to delete all entries for an event in the event_keywords table */
  private static final String DELETE_EKS =
      "delete from " + Tables.EVENTKEYWORDS + " where " +
      Fields.EKEVENT_ID.getInsertName() + "=?";

  /** Given a eventid and a keyword id delete all entries from the event
   * keyword table
   *
   * @param eventid   Id of event
   * @param keywordid Id of keyword
   * @exception CalFacadeException If there's a database access problem
   */
  public void deleteEventKeywords(int eventid) throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(DELETE_EKS);
      s.setInt(1, eventid);
      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /* ====================================================================
   *                       Users
   * ==================================================================== */

  /** SQL to find an entry in the users table by id */
  private static final String FIND_USER =
      "select " +
        Fields.UUSER_ID.getSelectName() + ", " +
        Fields.UUSERNAME.getSelectName() + ", " +
        Fields.UCREATED.getSelectName() + ", " +
        Fields.ULAST_LOGON.getSelectName() + ", " +
        Fields.ULAST_ACCESS.getSelectName() + ", " +
        Fields.ULAST_MODIFY.getSelectName() +
      " from " + Tables.USERS + " where " +
      Fields.UUSER_ID.getInsertName() + "=?";

  /** SQL to find an entry in the users table by account */
  private static final String FIND_USER_BY_ACCOUNT =
      "select " +
        Fields.UUSER_ID.getSelectName() + ", " +
        Fields.UUSERNAME.getSelectName() + ", " +
        Fields.UCREATED.getSelectName() + ", " +
        Fields.ULAST_LOGON.getSelectName() + ", " +
        Fields.ULAST_ACCESS.getSelectName() + ", " +
        Fields.ULAST_MODIFY.getSelectName() +
      " from " + Tables.USERS + " where " +
      Fields.UUSERNAME.getInsertName() + "=?";

  public UserVO getUser(int id) throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement(FIND_USER);
      s.setInt(1, id);
      rs = s.executeQuery();

      if (!rs.next()) {
        return null;
      }

      UserVO user = new UserVO();

      user.setId(rs.getInt(Fields.UUSER_ID.getName()));
      user.setAccount(rs.getString(Fields.UUSERNAME.getName()));
      user.setCreated(rs.getTimestamp(Fields.UCREATED.getName()));
      user.setLogon(rs.getTimestamp(Fields.ULAST_LOGON.getName()));
      user.setLastAccess(rs.getTimestamp(Fields.ULAST_ACCESS.getName()));
      user.setLastModify(rs.getTimestamp(Fields.ULAST_MODIFY.getName()));

      return user;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  public UserVO getUser(String account) throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement(FIND_USER_BY_ACCOUNT);
      s.setString(1, account);
      rs = s.executeQuery();

      if (!rs.next()) {
        return null;
      }

      UserVO user = new UserVO();

      user.setId(rs.getInt(Fields.UUSER_ID.getName()));
      user.setAccount(rs.getString(Fields.UUSERNAME.getName()));
      user.setCreated(rs.getTimestamp(Fields.UCREATED.getName()));
      user.setLogon(rs.getTimestamp(Fields.ULAST_LOGON.getName()));
      user.setLastAccess(rs.getTimestamp(Fields.ULAST_ACCESS.getName()));
      user.setLastModify(rs.getTimestamp(Fields.ULAST_MODIFY.getName()));

      return user;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Get the subscriptions for the user
   *
   * @param user        UserVO object
   * @returns Collection of subscriptions
   */
  public Collection getSubscriptions(UserVO user) throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      int id = user.getId();
      s = getPreparedStatement(
              "select * from subscriptions where userid=?");
      s.setInt(1, id);
      rs = s.executeQuery();

      Vector subs = new Vector();

      while (rs.next()) {
        FilterVO f = new FilterVO();

        f.setId(rs.getInt("calendarid"));
        subs.addElement(f);
      }

      return subs;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /* ====================================================================
   *                       Synch state
   * ==================================================================== */

  /** Get the synch info for the given device and user.
   *
   * @param user       UserVO representing user
   * @param deviceId   deviceId identifying the synch conversation .
   * @return SynchInfoVO object or null.
   */
   public SynchInfoVO getSynchInfo(UserVO user, String deviceId)
             throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement(
              "select * from synchinfo where userid=? " +
                                      "and deviceid=?");
      s.setInt(1, user.getId());
      s.setString(2, deviceId);
      rs = s.executeQuery();

      if (debug) {
        debugmsg("**********************Add synchinfo: " + user + " " + deviceId);
      }

      if (!rs.next()) {
        return null;
      }

      SynchInfoVO si = new SynchInfoVO();

      si.setUser(user);
      si.setDeviceId(rs.getString("deviceid"));
      si.setCalendarId(rs.getInt("calendarid"));
      si.setLastsynch(rs.getTimestamp("lastsynch"));

      return si;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
   }

  /** Add the synch info for the current device and user.
   *
   * @param val    SynchInfoVO object to add
   */
   public void addSynchInfo(SynchInfoVO val) throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "insert into synchinfo values (?,?,?,?)");

      s.setInt(1, val.getUser().getId());
      s.setString(2, val.getDeviceId());
      s.setInt(3, val.getCalendarId());
      s.setTimestamp(4, val.getLastsynch());

      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
   }

  /** Update the synch info for the current device and user. There must be
   * exactly one object to update.
   *
   * @param val    SynchInfoVO object to update
   */
   public void updateSynchInfo(SynchInfoVO val) throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "update synchinfo set calendarid=?, " +
                                     "lastsynch=? " +
                                "where userid=?" +
                                      "and deviceid=?");

      s.setInt(1, val.getCalendarId());
      s.setTimestamp(2, val.getLastsynch());
      s.setInt(3, val.getUser().getId());
      s.setString(4, val.getDeviceId());

      int rows = s.executeUpdate();
      if (rows > 1) {
        throw new CalFacadeException("Update to synchinfo affected " + rows);
      }
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
   }

  /** Return the synch state for the given event and user. If one doesn't exist
   * return null.
   *
   * @param ev         EventVO representing associated event
   * @param user       UserVO representing user
   * @param deviceId   deviceId identifying the synch conversation .
   * @return SynchStateVO object or null.
   */
  public SynchStateVO getSynchState(EventVO ev, UserVO user, String deviceId)
        throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      /* We need a statement something like:
          select ss.state,ss.guid,sd.eventdata from synchstate ss
                                   where ss.eventid=13
                                      and ss.userid=16
                                      and ss.deviceid='SynchTest1'
                                      left outer join synchdata sd on sd.userid=ss.userid
                                       and sd.eventid=ss.eventid
                                       */
      s = getPreparedStatement(
              "select state,guid from synchstate " +
                                   "where eventid=? " +
                                      "and userid=? " +
                                      "and deviceid=?");
      s.setInt(1, ev.getId());
      s.setInt(2, user.getId());
      s.setString(3, deviceId);
      rs = s.executeQuery();

      if (!rs.next()) {
        return null;
      }

      SynchStateVO ss = new SynchStateVO();

      ss.setEvent(ev);
      ss.setUser(user);
      ss.setSynchState(rs.getInt("state"));
      ss.setGuid(rs.getString("guid"));

      close();

      s = getPreparedStatement("select eventData from synchdata " +
                                   "where eventid=? " +
                                      "and userid=?");

      s.setInt(1, ev.getId());
      s.setInt(2, user.getId());

      rs = s.executeQuery();

      if (rs.next()) {
        ss.setSynchData(new SynchDataVO(user, ev.getId(),
                                        rs.getString("eventData")));
      }

      return ss;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Get the deleted synch states for the user and device
   *
   * @param user        UserVO object
   * @returns Collection of synch state objects (possibly empty)
   */
  public Collection getDeletedSynchStates(UserVO user,
                                          String deviceId) throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      int id = user.getId();
      s = getPreparedStatement(
              "select guid from synchstate where userid=? " +
                                      "and deviceid=?" +
                                      "and state=?");
      s.setInt(1, id);
      s.setString(2, deviceId);
      s.setInt(3, SynchStateVO.DELETED);
      rs = s.executeQuery();

      Vector sss = new Vector();

      while (rs.next()) {
        SynchStateVO ss = new SynchStateVO();

        ss.setUser(user);
        ss.setSynchState(SynchStateVO.DELETED);
        ss.setGuid(rs.getString("guid"));

        sss.addElement(ss);
      }

      return sss;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Add the synch state for the given event and user.
   *
   * @param val      SynchStateVO object.
   */
  public void addSynchState(SynchStateVO val)
        throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "insert into synchstate " +
                          "(userid,deviceid,eventid,guid,state)" +
                          "values (?,?, ?,?,?)");
      s.setInt(1, val.getUser().getId());
      s.setString(2, val.getDeviceId());
      s.setInt(3, val.getEvent().getId());
      s.setString(4, val.getGuid());
      s.setInt(5, val.getSynchState());
      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Update the synch state for the given event and user. There must be
   * zero or one object to update.
   *
   * <p>Zero is OK because an event may never have taken part in a
   * synchronization operation.
   *
   * @param val    SynchStateVO object with associated event and user
   */
  public void updateSynchState(SynchStateVO val)
        throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "update synchstate set state=?," +
                                    "guid=? " +
                                 "where eventid=? and userid=?" +
                                      "and deviceid=?");

      s.setInt(1, val.getSynchState());
      s.setString(2, val.getGuid());
      s.setInt(3, val.getEvent().getId());
      s.setInt(4, val.getUser().getId());
      s.setString(5, val.getDeviceId());

      int rows = s.executeUpdate();
      if (rows > 1) {
        throw new CalFacadeException("Update to synchstate affected " + rows);
      }
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Update the synch data for the given event.
   *
   * <p>If the value has no synch data - do nothing. We could try a delete
   *
   * <p>If there is some synch data try a fetch to see if we should update
   * or add..
   *
   * @param val    SynchStateVO object with associated event and user
   * @return true if data changed or added.
   */
  private boolean updateSynchData(SynchStateVO val)
        throws CalFacadeException {
    SynchDataVO sd = val.getSynchData();

    if (sd == null) {
      return false;
    }

    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement(
              "select eventData from synchdata where userid=? " +
                                      "and eventid=?");
      s.setInt(1, val.getUser().getId());
      s.setInt(2, sd.getEventId());
      rs = s.executeQuery();

      if (rs.next()) {
        String ed = rs.getString("eventData");

        if (ed.equals(sd.getEventData())) {
          // No need to update
          return false;
        }

        close();
        s = getPreparedStatement(
              "update synchdata set eventData=? " +
                                 "where eventid=? and userid=?");

        s.setString(1, sd.getEventData());
        s.setInt(2, sd.getEventId());
        s.setInt(3, val.getUser().getId());

        s.executeUpdate();
        return true;
      }

      // Need to add.
      close();

      s = getPreparedStatement(
              "insert into synchdata values (?,?,?)");

      s.setInt(1, val.getUser().getId());
      s.setInt(2, sd.getEventId());
      s.setString(3, sd.getEventData());

      s.executeUpdate();

      return true;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Get the synch data associated with the given object..
   *
   * @param val    SynchStateVO object
   */
  public void getSynchData(SynchStateVO val) throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement(
              "select eventData from synchdata where userid=? " +
                                      "and eventid=?");
      s.setInt(1, val.getUser().getId());
      s.setInt(2, val.getEvent().getId());
      rs = s.executeQuery();

      if (!rs.next()) {
        val.setSynchData(null);
        return;
      }

      String ed = rs.getString("eventData");

      SynchDataVO sd = new SynchDataVO(val.getUser(),
                                       val.getEvent().getId(),
                                       ed);

      val.setSynchData(sd);
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Set the synch state and data for the given user and event.
   * There may be zero to many objects to update.
   *
   * @param val    SynchStateVO object with associated event and user
   */
  public int setSynchData(SynchStateVO val) throws CalFacadeException {
    /* First update the synchdata */
    if (!updateSynchData(val)) {
      // Nothing changed
      return 0;
    }

    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "update synchstate set state=? " +
                                 "where eventid=? and userid=?");
      s.setInt(1, SynchStateVO.CLIENT_MODIFIED_UNDELIVERED);
      s.setInt(2, val.getEvent().getId());
      s.setInt(3, val.getUser().getId());

      return s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Set the synch state for the given event. There may be zero to many
   * objects to update.
   *
   * @param ev     EventVO representing associated event
   * @param state  int new state - defined in SynchStateVO
   * @return int   number of updated entries.
   */
  public int setSynchState(EventVO ev, int state)
        throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "update synchstate set state=? where eventid=?");
      s.setInt(1, state);
      s.setInt(2, ev.getId());

      return s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Set the synch state for the given user and event. There may be zero to many
   * objects to update.
   *
   * @param user   UserVO object
   * @param ev     EventVO representing associated event
   * @param state  int new state - defined in SynchStateVO
   * @return int   number of updated entries.
   */
  public int setSynchState(UserVO user, EventVO ev, int state)
        throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "update synchstate set state=? " +
                                 "where userid=? " +
                                       "and eventid=?");
      s.setInt(1, state);
      s.setInt(2, user.getId());
      s.setInt(3, ev.getId());

      return s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Set the synch state for the given user, synchId and event. There must
   * be one and only one object to update.
   *
   * @param user    UserVO object
   * @param synchId String synch id
   * @param ev      EventVO representing associated event
   * @param state   int new state - defined in SynchStateVO
   */
  public void setSynchState(UserVO user, String synchId, EventVO ev, int state)
        throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "update synchstate set state=? " +
                                 "where userid=? " +
                                      "and deviceid=? " +
                                       "and eventid=?");
      s.setInt(1, state);
      s.setInt(2, user.getId());
      s.setString(3, synchId);
      s.setInt(4, ev.getId());

      int rows = s.executeUpdate();
      if (rows > 1) {
        throw new CalFacadeException("Update to synchstate affected " + rows);
      }
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Update the synch states for the given user and device.
   *
   * <p>This is called at the termination of synchronization to set any state
   * objects appropriately.
   *
   * <p><table cols="2">
   * <th><td>Old state</td><td>New state</td></th>
   * <tr><td>UNKNOWN</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>SYNCHRONIZED</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>NEW</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>MODIFIED</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>DELETED</td><td>leave unchanged</td></tr>
   * <tr><td>CLIENT_DELETED</td><td>unchanged</td></tr>
   * <tr><td>CLIENT_DELETED_UNDELIVERED</td><td>CLIENT_DELETED</td></tr>
   * <tr><td>CLIENT_MODIFIED</td><td>unchanged</td></tr>
   * <tr><td>CLIENT_MODIFIED_UNDELIVERED</td><td>CLIENT_MODIFIED</td></tr>
   * </table>
   */
  public void updateSynchStates(UserVO user,
                                String deviceId) throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "update synchstate set state=? " +
                                 "where userid=? " +
                                       "and deviceid=?" +
                                       "and state<=?");
      s.setInt(1, SynchStateVO.SYNCHRONIZED);
      s.setInt(2, user.getId());
      s.setString(3, deviceId);
      s.setInt(4, SynchStateVO.MODIFIED);

      s.executeUpdate();

      close();

      s = getPreparedStatement(
              "update synchstate set state=? " +
                                 "where userid=? " +
                                       "and deviceid=?" +
                                       "and state=?");
      s.setInt(1, SynchStateVO.CLIENT_DELETED);
      s.setInt(2, user.getId());
      s.setString(3, deviceId);
      s.setInt(4, SynchStateVO.CLIENT_DELETED_UNDELIVERED);

      s.executeUpdate();

      close();

      s = getPreparedStatement(
              "update synchstate set state=? " +
                                 "where userid=? " +
                                       "and deviceid=?" +
                                       "and state=?");
      s.setInt(1, SynchStateVO.CLIENT_MODIFIED);
      s.setInt(2, user.getId());
      s.setString(3, deviceId);
      s.setInt(4, SynchStateVO.CLIENT_MODIFIED_UNDELIVERED);

      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Remove deleted synch states for the given user and device.
   *
   */
  public void removeDeletedSynchStates(UserVO user,
                                      String deviceId) throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "delete from synchstate " +
                                 "where userid=? " +
                                       "and deviceid=?" +
                                       "and state=?");
      s.setInt(1, user.getId());
      s.setString(2, deviceId);
      s.setInt(3, SynchStateVO.DELETED);

      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /* ====================================================================
   *                       Recurrences
   * This is currently under development at RPI
   * ==================================================================== */

  /** Add a recurrence to the database.
   *
   * @param val      RecurVO object.
   */
  public void addRecur(RecurVO val) throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "insert into recurrences " +
                 "(eventid, startdate, starttime, " +
                      "enddate, endtime, expanded) " +
                 "values (?,?,?,?,?,?)");

      s.setInt(1, val.getEvent().getId());
      s.setDate(2, val.getEvent().getStartdate());
      s.setTime(3, val.getEvent().getStarttime());
      s.setDate(4, null);
      s.setTime(5, null);
      if (val.getExpanded()) {
        s.setString(6, "T");
      } else {
        s.setString(6, "F");
      }

      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }

    addRecurRules(val);
  }

  /** Remove a recurrence associated with an event.
   *
   * @param ev      Associated EventVO
   */
  public void deleteRecur(EventVO ev) throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "delete from recurrences where eventid=?");
      s.setInt(1, ev.getId());

      s.executeUpdate();

      close();

      s = getPreparedStatement(
              "delete from rdates where eventid=?");
      s.setInt(1, ev.getId());

      s.executeUpdate();

      close();

      s = getPreparedStatement(
              "delete from rrules where eventid=?");
      s.setInt(1, ev.getId());

      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Update a recurrence in the database.
   *
   * @param val      RecurVO object.
   */
  public void updateRecur(RecurVO val) throws CalFacadeException {
    EventVO ev = val.getEvent();

    deleteRecur(ev);
    addRecur(val);
  }

  /** Get a recurrence associated with an event.
   *
   * @param ev      Associated EventVO
   * @return RecurVO  recurrence object or null
   */
  public RecurVO getRecur(EventVO ev) throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;
    RecurVO res = null;

    try {
      s = getPreparedStatement(
              "select * from recurrences where eventid=?");

      s.setInt(1, ev.getId());

      rs = s.executeQuery();
      if (!rs.next()) {
        return null;
      }

      res = makeRecur(ev, rs);
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close();
    }

    res.setExclusionRules(getRecurRule(res, false));
    res.setExclusionDates(getRecurDates(res, false));
    res.setRecurrenceRules(getRecurRule(res, true));
    res.setRecurrenceDates(getRecurDates(res, true));

    return res;
  }

  /* ====================================================================
   *                       Private
   * ==================================================================== */

  private static final String getEventKeywordIdsSql =
                 "select " +
                     Fields.EKKEYWORD_ID.getSelectName() +
                  " from " +
                     Tables.EVENTKEYWORDS.getName() +
                  " where " +
                     Fields.EKEVENT_ID.getStmtName() + "=?";

  /** Add keyword ids to event
   *
   * @param   ev        the event
   */
  private void getEventKeywordIds(Event ev)
    throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement(getEventKeywordIdsSql);
      s.setInt(1, ev.getId());
      rs = s.executeQuery();
      while (rs.next()) {
        ev.addKeyword(rs.getInt(Fields.EKKEYWORD_ID.getName()));
      }
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close();
    }
  }

  /** Convert (possibly null) IntArrayList to a string.
   */
  private String valueOf(IntArrayList val) {
    if (val == null) {
      return null;
    }

    return val.encode();
  }

  /** Convert a possible null string to an IntArrayList
   */
  private IntArrayList intArrayVal(String val) {
    if (val == null) {
      return null;
    }

    return IntArrayList.decode(val);
  }

  /** Make an unpopulated recurrence from current result.
   *
   * @param ev      Associated EventVO
   * @param rs      ResultSet positioned at a row
   * @return RecurVO  recurrence object
   */
  public RecurVO makeRecur(EventVO ev, ResultSet rs) throws Throwable {
    RecurVO res = new RecurVO();

    res.setEvent(ev);
    res.setExpanded("T".equals(rs.getString("expanded")));

    return res;
  }

  /** Add recurrence and exclusion dates to the database.
   *
   * @param val      RecurVO object.
   */
  private void addRecurDates(RecurVO val)
        throws CalFacadeException {
    Collection dates = val.getExclusionDates();
    Iterator it = dates.iterator();
    while (it.hasNext()) {
      addRecurDate(val, (java.util.Date)it.next(), false);
    }

    dates = val.getRecurrenceDates();
    it = dates.iterator();
    while (it.hasNext()) {
      addRecurDate(val, (java.util.Date)it.next(), true);
    }
  }

  /** Add recurrence and exclusion rules to the database.
   *
   * @param val      RecurVO object.
   */
  private void addRecurRules(RecurVO val)
        throws CalFacadeException {
    Collection rules = val.getExclusionRules();
    Iterator it = rules.iterator();
    while (it.hasNext()) {
      addRecurRule((RecurRuleVO)it.next());
    }

    rules = val.getRecurrenceRules();
    it = rules.iterator();
    while (it.hasNext()) {
      addRecurRule((RecurRuleVO)it.next());
    }
  }

  /** Add a recurrence date to the database.
   *
   * @param val        RecurVO object.
   * @param dt         java.util.Date object.
   * @param recurrence true for a recurrence (as against exclusion).
   */
  private void addRecurDate(RecurVO val, java.util.Date dt,
                            boolean recurrence)
        throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "insert into rdates " +
                 "(eventid, type, rdate, rtime) " +
                 "values (?,?,?,?)");

      s.setInt(1, val.getEvent().getId());
      if (recurrence) {
        s.setString(2, "R");
      } else {
        s.setString(2, "E");
      }

      if (dt == null) {
        s.setDate(3, null);
        s.setTime(4, null);
      } else {
        long millis = dt.getTime();

        s.setDate(3, new Date(millis));
        s.setTime(4, new Time(millis));
      }

      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Get the recurrence/exclusion dates for the recurrence.
   *
   * @param val       RecurVO object.
   * @param recur     true for recurrence, false for exclusion
   * @return Collection of java.util.Date
   */
  private Collection getRecurDates(RecurVO val, boolean recurrence)
        throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement(
              "select * from rdates where eventid=? and type=?" +
                  " order by ruleid");

      s.setInt(1, val.getEvent().getId());
      if (recurrence) {
        s.setString(2, "R");
      } else {
        s.setString(2, "E");
      }

      Vector v = new Vector();
      rs = s.executeQuery();
      while (rs.next()) {
        Date dt = rs.getDate("rdate");
        Time tm = rs.getTime("rtime");

        if (dt != null) {
          long millis = dt.getTime();
          if (tm != null) {
            millis += tm.getTime();
          }
          v.add(new java.util.Date(millis));
        }
      }

      return v;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close();
    }
  }

  /** Add a recurrence rule to the database.
   *
   * @param val        RecurRuleVO object.
   * @param recurrence true for a recurrence (as against exclusion).
   */
  private void addRecurRule(RecurRuleVO val)
        throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(
              "insert into rrules " +
                 "(ruleid, eventid, type, freq, " +
                      "untildate, untiltime, count, interval, " +
                      "bysecond, byMinute, byHour, byDay, byMonthday, " +
                      "byYearday, byWeekno, byMonth, bySetpos, wkst, ruleText) " +
                 "values (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)");

      s.setInt(1, val.getId());
      s.setInt(2, val.getRecur().getEvent().getId());
      if (val.getRecurrence()) {
        s.setString(3, "R");
      } else {
        s.setString(3, "E");
      }
      s.setInt(4, val.getFreq());

      java.util.Date dt = val.getUntil();
      if (dt == null) {
        s.setDate(5, null);
        s.setTime(6, null);
      } else {
        long millis = dt.getTime();

        s.setDate(5, new Date(millis));
        s.setTime(6, new Time(millis));
      }

      s.setInt(7, val.getCount());
      s.setInt(8, val.getInterval());
      s.setString(9, valueOf(val.getBySecond()));
      s.setString(10, valueOf(val.getByMinute()));
      s.setString(11, valueOf(val.getByHour()));
      s.setString(12, valueOf(val.getByDay()));
      s.setString(13, valueOf(val.getByMonthday()));
      s.setString(14, valueOf(val.getByYearday()));
      s.setString(15, valueOf(val.getByWeekno()));
      s.setString(16, valueOf(val.getByMonth()));
      s.setString(17, valueOf(val.getBySetpos()));
      s.setInt(18, val.getWkst());
      s.setString(19, val.getRuleText());

      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Get the recurrence/exclusion rule set for the recurrence.
   *
   * @param val       RecurVO object.
   * @param recur     true for recurrence, false for exclusion
   * @return Collection of RecurRuleVO
   */
  private Collection getRecurRule(RecurVO val, boolean recurrence)
        throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement(
              "select * from rrules where eventid=? and type=?" +
                  " order by ruleid");

      s.setInt(1, val.getEvent().getId());
      if (recurrence) {
        s.setString(2, "R");
      } else {
        s.setString(2, "E");
      }

      Vector v = new Vector();
      rs = s.executeQuery();
      while (rs.next()) {
        RecurRuleVO rule = new RecurRuleVO();

        rule.setRecur(val);

        rule.setId(rs.getInt("ruleid"));
        rule.setRecurrence("R".equals(rs.getString("type")));
        rule.setFreq(rs.getInt("freq"));

        Date dt = rs.getDate("untildate");
        Time tm = rs.getTime("untiltime");

        if (dt != null) {
          long millis = dt.getTime();
          if (tm != null) {
            millis += tm.getTime();
          }
          rule.setUntil(new java.util.Date(millis));
        }

        rule.setCount(rs.getInt("count"));
        rule.setInterval(rs.getInt("interval"));

        rule.setBySecond(intArrayVal(rs.getString("bySecond")));
        rule.setByMinute(intArrayVal(rs.getString("byMinute")));
        rule.setByHour(intArrayVal(rs.getString("byHour")));
        rule.setByDay(intArrayVal(rs.getString("byDay")));
        rule.setByMonthday(intArrayVal(rs.getString("byMonthday")));
        rule.setByYearday(intArrayVal(rs.getString("byYearday")));
        rule.setByWeekno(intArrayVal(rs.getString("byWeekno")));
        rule.setByMonth(intArrayVal(rs.getString("byMonth")));
        rule.setBySetpos(intArrayVal(rs.getString("bySetpos")));

        rule.setWkst(rs.getInt("wkst"));
        rule.setRuleText(rs.getString("ruleText"));

        v.add(rule);
      }

      return v;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close();
    }
  }

  /** Get a logger for messages
   */
  private Logger getLogger() {
    if (log == null) {
      log = Logger.getLogger(this.getClass());
    }

    return log;
  }

  private void debugmsg(String msg) {
    getLogger().debug(msg);
  }

  private void trace(String msg) {
    debugmsg("trace: " + msg);
  }

  /* ====================================================================
   *                       Misc
   * ==================================================================== */

  /** Print the queries used in the class to standard output
   *
   * @param args    String[] ignored
   */
  public static void main(String[] args) {
    System.out.println(SELECT_SPONSOR_FROM_EVENTS);
    System.out.println(SELECT_LOCATION_FROM_EVENTS);
    System.out.println(DELETE_EVENTS);
    System.out.println(ADD_EVENT_KEYWORD);
    System.out.println(DELETE_EK);
  }
}
