package edu.washington.cac.calfacade.shared;

import java.io.Serializable;
import java.sql.Date;
import java.sql.Time;
import java.util.Vector;

/** Value object to represent an authorised calendar user - that is a user
 * with some special privilege. This could also be represented by users
 * with roles.
 *
 *   @author Mike Douglass douglm@rpi.edu
 *  @version 1.0
 */
public class AuthUserVO implements Serializable {
  private int id;
  protected UserVO user;  // Related user entry

  protected AuthUserPrefsVO prefs;  // Related user prefs entry

  private int usertype;
  private char dbtype;
  private String lastname;
  private String firstname;
  private String phone;
  private String email;
  private String dept;

  /* ====================================================================
   *                   Constructors
   * ==================================================================== */

  /** No-arg constructor
   */
  public AuthUserVO() {
  }

  /** Create an authuser vo by specifying all its fields
   *
   * @param  user           UserVO object representing user
   * @param  usertype       int type of user
   * @param  lastname       String
   * @param  firstname      String
   * @param  phone          String
   * @param  email          String
   * @param  dept           String
    */
  public AuthUserVO(UserVO user,
                    int usertype,
                    String lastname,
                    String firstname,
                    String phone,
                    String email,
                    String dept) {
    this.user = user;
    this.usertype = usertype;
    this.lastname = lastname;
    this.firstname = firstname;
    this.phone = phone;
    this.email = email;
    this.dept = dept;
  }

  /* ====================================================================
   *                   Bean methods
   * ==================================================================== */

  /**    Set the id
   * @param val   id
   */
  public void setId(int val) {
    id = val;
  }

  /** Get the unique id
   *
   * @return int    the unique id
   */
  public int getId() {
    return id;
  }

  public void setUser(UserVO val) {
    user = val;
  }

  public UserVO getUser() {
    return user;
  }

  public void setPrefs(AuthUserPrefsVO val) {
    prefs = val;
  }

  public AuthUserPrefsVO getPrefs() {
    return prefs;
  }

  public void setUsertype(int val) {
    usertype = val;
    dbtype = toDbtype(usertype);
  }

  public int getUsertype() {
    return usertype;
  }

  public void setDbtype(char val) {
    dbtype = val;
    usertype = adjustUsertype(toType(dbtype));
  }

  public char getDbtype() {
    return dbtype;
  }

  public void setLastname(String val) {
    lastname = val;
  }

  public String getLastname() {
    return lastname;
  }

  public void setFirstname(String val) {
    firstname = val;
  }

  public String getFirstname() {
    return firstname;
  }

  public void setPhone(String val) {
    phone = val;
  }

  public String getPhone() {
    return phone;
  }

  public void setEmail(String val) {
    email = val;
  }

  public String getEmail() {
    return email;
  }

  public void setDept(String val) {
    dept = val;
  }

  public String getDept() {
    return dept;
  }

  /* ====================================================================
   *                      Convenience methods
   * ==================================================================== */

  public boolean isAlertUser() {
    return ((usertype & UserAuth.alertUser) != 0);
  }

  public boolean isPublicEventUser() {
    return ((usertype & UserAuth.publicEventUser) != 0);
  }

  public boolean isSuperUser() {
    return ((usertype & UserAuth.superUser) != 0);
  }

  /* ====================================================================
   *                        Private methods
   * ==================================================================== */

  /** Convert internal db type to external user type.
   */
  private int toType(char dbType) {
    if (dbType == ' ') {
      return UserAuth.noPrivileges;
    }

    if (dbType == 'A') {
      return UserAuth.superUser;
    }

    if (dbType == 'C') {
      return UserAuth.contentAdminUser;
    }

    if (dbType == 'S') {
      return UserAuth.alertUser + UserAuth.publicEventUser;
    }

    if (dbType == 'P') {
      return UserAuth.publicEventUser;
    }

    if (dbType == 'a') {
      return UserAuth.alertUser;
    }

    return UserAuth.noPrivileges;
  }

  /** Convert external user type to internal db type.
   *
   * @return String   internal type - null for no privileges.
   */
  private char toDbtype(int usertype) {
    if (usertype == UserAuth.noPrivileges) {
      return ' ';
    }

    if ((usertype & UserAuth.superUser) != 0) {
      return 'A';
    }

    if (usertype == UserAuth.contentAdminUser) {
      return 'C';
    }

    if (usertype == UserAuth.publicEventUser) {
      return 'P';
    }

    if (usertype == UserAuth.alertUser) {
      return 'a';
    }

    if (usertype == (UserAuth.alertUser + UserAuth.publicEventUser)) {
      // semi-priv
      return 'S';
    }

    return ' ';
  }

  /** Adjust the user type to take account of site specific requirements.
   *
   * <p>For example, some sites may allow any user who can add public events
   * the authority to add alerts. Others may wish to keep those roles
   * separate.
   *
   * <p>This routine also should ensure all the super user privileges are
   * set.
   *
   * @param userType    int unadjusted user type.
   * @return int   adjusted user type.
   * @exception CalFacadeException   For invalid usertype values.
   */
  private int adjustUsertype(int userType) {
    if ((userType & ~UserAuth.allAuth) != 0) {
      throw new RuntimeException("Invalid authorised user type: " + userType);
    }

    if ((userType & UserAuth.superUser) != 0) {
      return UserAuth.allAuth;
    }

    /** We'll allow any combination of the remaining flags.
     */
    return userType;
  }

  /* ====================================================================
   *                        Object methods
   * ==================================================================== */

  public String toString() {
    StringBuffer sb = new StringBuffer();

    sb.append("AuthUserVO{user=");
    sb.append(user);
    sb.append(", usertype=");
    sb.append(usertype);
    sb.append("}");

    return sb.toString();
  }

  public Object clone() {
    UserVO u = (UserVO)getUser().clone();

    AuthUserVO au = new AuthUserVO(u, getUsertype(), getLastname(),
                                   getFirstname(),
                                   getPhone(),
                                   getEmail(),
                                   getDept());

    AuthUserPrefsVO aup = getPrefs();

    if (aup != null) {
      aup = (AuthUserPrefsVO)aup.clone();
      aup.setAuthUser(au);

      au.setPrefs(aup);
    }

    return au;
  }
}
