package edu.washington.cac.calfacade.shared;

import java.io.Serializable;
import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;

/** An Event in UWCal. This value object does no consistency or validity
 * checking. It does not modify the dates.
 *
 *  @version 1.0
 */
public class EventVO implements Comparable, Comparator, Serializable {
  /** Some constants. These are copies of the constants in the data objects.
   *  I want to remove all references to data classes in the facade. We
   *  should probably move those constants to here or to a global
   *  constant defs class.
   */

  public final static int NoLocationId = 1;
  public final static int UnknownLocationId = 2;

  /** Unique ID of a location that has been deleted (An event can
   *   use a location created by another user, and if the user
   *   deletes the location, the 'deleted' location can be used
   *   to indicate this event to this user.)
   */
  public final static int DeletedLocationId = 3;

  public final static int DefaultLocationId = NoLocationId;

  public final static int NoSponsorId = 1;
  public final static int UnknownSponsorId = 2;
  public final static int DeletedSponsorId = 3;

  public final static int DefaultSponsorId = NoSponsorId;

  private int id;
  private String shortdesc;

  /** This should be set by the back end code.
   */
  public static final int maxLongdescLength = 500;

  private String longdesc;
  private Date startdate;
  private Time starttime;
  private Date enddate;
  private Time endtime;
  private String link;
  private char status = 'F';
  private String cost;

  /** User who created this event */
  private UserVO creator;

  private boolean isPublic;

  private int seq;

  private Timestamp lastmod;

  private Timestamp created;

  /** Information about whether this event is a recurring event, and
   * if it is, how it recurs.
   *  null means no recurrence
   */
  private RecurrenceVO recurrence;
  private char recurrenceStatus = 'N'; // Derived from above

  /** This object, if present, holds rfc2445 recurrence information.
   *  ********* Under development *********
   */
  private RecurVO recur;

  /** A Set of KeywordVO objects
   */
  private Collection keywords = null;

  /** This may or may not be set to populate fully
   */
  private SponsorVO sponsor;

  /** This may or may not be set to populate fully
   */
  private LocationVO location;

  /** editable is set at retrieval to indicate an event owned by the current
   * user. This only has significance for the personal calendar.
   */
  private boolean editable;

  public final static int kindEntry = 0;
  public final static int kindAdded = 1;
  public final static int kindUndeletable = 2;

  /** kind tells us something about where the event comes from.
   * <br/>
   *  0 - actual event entry<br/>
   *  1 - 'added' event - from a reference<br/>
   *  2 - from a subscription
   */
  private int kind;

  private static final String[] kindStr = {
    "entry",
    "reffed",
    "subscribed",
  };

  /** This is a generated field. It does not appear in the db and only exists
   * for events which exists in the database. UWCal guids take to form: <br/>
   *   id-name@host
   * <br/>where<ul>
   * <li>id is the internal event id</li>
   * <li>name identifies the particular calendar system at the site</li>
   * <li>host part identifies the domain of the site.</li>..
   * </ul>
   */
  private String guid;

  public EventVO() {
    created = new Timestamp(System.currentTimeMillis());
  }

  /** Create an event by specifying all its fields

      @param id           Unique ID for the event
      @param shortdesc    String  Short description
      @param longdesc     String Long description
      @param startdate    Date start Date
      @param starttime    Time starting Time (can be null)
      @param enddate      Date end Date (can be null)
      @param endtime      Time end Time (can be null)
      @param isPublic     boolean true if the event is viewable by everyone
      @param link         String URL with more info
      @param creator      UserVO creator of the event
      @param cost         String Cost to attend
      @param recurrence   RecurrenceVO information
      @param location     event's location
      @param sponsor      event's sponsor
      @param seq          int sequence number
      @param lastmod      Timestamp last modification time
      @param created      Timestamp creation time
    */
  public EventVO(int id,
               String shortdesc,
               String longdesc,
               Date startdate,
               Time starttime,
               Date enddate,
               Time endtime,
               boolean isPublic,
               String link,
               UserVO creator,
               String cost,
               RecurrenceVO recurrence,
               LocationVO location,
               SponsorVO sponsor,
               int seq,
               Timestamp lastmod,
               Timestamp created) {
    this.id = id;
    this.shortdesc = shortdesc;
    this.longdesc = longdesc;
    this.startdate = startdate;
    this.starttime = starttime;
    this.enddate = enddate;
    this.endtime = endtime;
    this.isPublic = isPublic;
    this.link = link;
    this.creator = creator;
    this.cost = cost;
    this.recurrence = recurrence;
    this.location = location;
    this.sponsor = sponsor;
    this.seq = seq;
    this.lastmod = lastmod;
    this.created = created;
  }

  /** ===================================================================
   *                      Bean methods
   *  =================================================================== */

  /**
    Set the id for this event
    @param eventid Event id
   */
  public void setId(int val) {
    id = val;
  }

  /** Get the event's unique id
   *
   * @return int    the event's unique id
   */
  public int getId() {
    return id;
  }

  /** Set the event's short description
   *
   * @param val    String event's short description
   */
  public void setShortdesc(String val) {
    shortdesc = val;
  }

  /** Get the event's short description
   *
   * @return String   event's short description
   */
  public String getShortdesc() {
    return shortdesc;
  }

  /** Set the event's long description
   *
   * @param val    String event's short description
   */
  public void setLongdesc(String val) {
    longdesc = val;
  }

  /** Get the event's long description
   *
   *  @return String   event's long description
   */
  public String getLongdesc() {
    return longdesc;
  }

  /** Set the event's starting date (not time)
   *
   *  @param  val   Date event's starting date (not time)
   */
  public void setStartdate(Date val) {
    startdate = val;
  }

  /** Get the event's starting date (not time)
   *
   *  @return The event's starting date (not time)
   */
  public Date getStartdate() {
    return startdate;
  }

  /** Set the event's starting time (not date)
   *
   *  @param  val   Time event's starting time (not date)
   */
  public void setStarttime(Time val) {
    starttime = val;
  }

  /** Get the event's starting time (not date)
   *
   *  @return The event's starting time (not date)
   */
  public Time getStarttime() {
    return starttime;
  }

  /** Set the event's ending date (not time)
   *
   *  @param  val   Date event's ending date (not time)
   */
  public void setEnddate(Date val) {
    enddate = val;
  }

  /** Get the event's ending date (not time)
   *
   * @return The event's ending date (not time)
   */
  public Date getEnddate() {
    return enddate;
  }

  /** Set the event's ending time (not date)
   *
   *  @param  val   Time event's ending time (not date)
   */
  public void setEndtime(Time val) {
    endtime = val;
  }

  /** Get the event's ending time (not date)
   *
   * @return Time    The event's ending time (not date)
   */
  public Time getEndtime() {
    return endtime;
  }

  /**
    Set uwcal recurrence information for this event
    @param val       RecurrenceVO information
   */
  public void setRecurrence(RecurrenceVO val) {
    recurrence = val;
  }

  /** Get uwcal recurrence information for this event
   *
   * @return ReccurrenceVO information
   */
  public RecurrenceVO getRecurrence() {
    return recurrence;
  }

  /** Set the event's recurrence status
   *
   *  @param val     recurrence status
   */
  public void setRecurrenceStatus(char val) {
     recurrenceStatus = val;
  }

  /** Get the event's recurrence status
   *
   * @return the event's recurrence status
   */
  public char getRecurrenceStatus() {
    return  recurrenceStatus;
  }

  /**
    Set rfc2445 recurrence information for this event
    @param val       RecurVO information
   */
  public void setRecur(RecurVO val) {
    recur = val;
  }

  /** Get rfc2445 recurrence information for this event
   *
   * @return ReccurVO information
   */
  public RecurVO getRecur() {
    return recur;
  }

  /** Set the event public flag
   *  @param val    boolean true if the event is public
   */
  public void setPublic(boolean val) {
    isPublic = val;
  }

  /** Get the event public flag
   *
   *  @return boolean    true if the event is public
   */
  public boolean getPublic() {
    return isPublic;
  }

  /** Is the event public?
   *  @return boolean    true if the event is public
   */
  public boolean isPublic() {
    return isPublic;
  }

  /** Set the event's URL
   *
   *  @param link   string URL
   */
  public void setLink(String link) {
    this.link = link;
  }

  /** Get the event's URL
   *
   * @return the event's URL
   */
  public String getLink() {
    return link;
  }

  /** Set the event's status
   *
   *  @param val     status
   */
  public void setStatus(char val) {
    status = val;
  }

  /** Get the event's status
   *
   * @return the event's status
   */
  public char getStatus() {
    return status;
  }

  /** Set the creator
   *
   * @param val      UserVO creator of the event
   */
  public void setCreator(UserVO val) {
    creator = val;
  }

  /** Return the creator of the event
   *
   * @return UserVO       Event's creator
   */
  public UserVO getCreator() {
    return creator;
  }

  /** Set the event's cost
   *
   * @param val    String event's cost
   */
  public void setCost(String val) {
    cost = val;
  }

  /** Get the event's cost
   *
   * @return String   the event's cost
   */
  public String getCost() {
    return cost;
  }

  /** Set the seq for this event
   *
   * @param val   sequence number
   */
  public void setSeq(int val) {
    seq = val;
  }

  /** Get the events seq
   *
   * @return int    the events seq
   */
  public int getSeq() {
    return seq;
  }

  public void setLastmod(Timestamp val) {
    lastmod = val;
  }

  public Timestamp getLastmod() {
    return lastmod;
  }

  public void setCreated(Timestamp val) {
    created = val;
  }

  public Timestamp getCreated() {
    return created;
  }

  /** Set the events sponsor
   *
   * @param val  SponsorVO events sponsor
   */
  public void setSponsor(SponsorVO val) {
    sponsor = val;
  }

  /**
    Get the event's sponsor
    @return SponsorVO   event's sponsor
   */
  public SponsorVO getSponsor() {
    return sponsor;
  }

  public void setLocation(LocationVO val) {
    location = val;
  }

  public LocationVO getLocation() {
    return location;
  }

  public void setEditable(boolean val) {
    editable = val;
  }

  public boolean getEditable() {
    return editable;
  }

  public void setKind(int val) {
    kind = val;
  }

  public int getKind() {
    return kind;
  }

  /** Set the event's guid
   *
   * @param val    String event's guid
   */
  public void setGuid(String val) {
    guid = val;
  }

  /** Get the event's guid
   *
   * @return String   event's guid
   */
  public String getGuid() {
    return guid;
  }

  public void setKeywords(Collection val) {
    keywords = val;
  }

  public Collection getKeywords() {
    if (keywords == null) {
      keywords = new TreeSet();
    }

    return keywords;
  }

  /** ===================================================================
   *                   Convenience methods
   *  =================================================================== */

  /**
    Is this a recurring event?
    @return Is this a recurring event?
   */
  public boolean isRecurring() {
    if (recurrence == null) {
      return false;
    }

    return recurrence.getRecurs();
  }

  /** Get an <code>Iterator</code> over the event's keywords
   *
   * @return Iterator  over the event's keywords
   */
  public Iterator iterateKeywords() {
    return getKeywords().iterator();
  }

  /** clear the event's keywords
   */
  public void clearKeywords() {
    getKeywords().clear();
  }

  public void addKeyword(KeywordVO val) {
    Collection ks = getKeywords();

    if (!ks.contains(val)) {
      ks.add(val);
    }
  }

  /** Get the first keyword or null.
   *
   * @return KeywordVO first keyword or null.
   */
  public KeywordVO getFirstKeyword() {
    Iterator it = iterateKeywords();

    if (!it.hasNext()) {
      return null;
    }

    return (KeywordVO)it.next();
  }

  /* * Set the ith keyword.
   *
   * @param  i         int index of keyword
   * @return KeywordVO ith keyword or null.
   * /
  public void setKeyword(int i, KeywordVO val) throws Throwable {
    TreeSet ks = (TreeSet)getKeywords();

    if (i >= ks.size()) {
      throw new Exception("Keyword index out of range");
    }

    ks.removeElementAt(i);
    ks.insertElementAt(val, i);
  }

  /* * Return the ith keyword or null.
   *
   * @param  i         int index of keyword
   * @return KeywordVO ith keyword or null.
   * /
  public KeywordVO getKeyword(int i) {
    TreeSet ks = (TreeSet)getKeywords();

    if (i >= ks.size()) {
      return null;
    }

    return (KeywordVO)ks.elementAt(i);
  }

  /* * return the corresponding KeywordVO object if the given id is within
   * the keywords list
   *
   * @param  id    int keyword id
   * @return KeywordVO  keyword value object
   * /
  public KeywordVO findKeyword(int id) {
    Iterator it = iterateKeywords();

    while (it.hasNext()) {
      KeywordVO k = (KeywordVO)it.next();

      if (id == k.getId()) {
        return k;
      }
    }

    return null;
  }*/

  /** Check the keywords for the same id
   *
   * @param k          Keyword to test
   * @return boolean   true if the event has a particular keyword
   */
  public boolean hasKeyword(KeywordVO k) {
    return getKeywords().contains(k);
  }

  /* ====================================================================
   *                   Object methods
   * ==================================================================== */

  /** Override the default and sort on startdate, starttime, enddate, endtime
   */
  public int compare(Object o1, Object o2) {
    if (!(o1 instanceof EventVO)) {
      return -1;
    }

    if (!(o2 instanceof EventVO)) {
      return 1;
    }

    EventVO e1 = (EventVO)o1;
    EventVO e2 = (EventVO)o2;

    int res = e1.startdate.compareTo(e2.startdate);
    if (res != 0) {
      return res;
    }

    res = e1.starttime.compareTo(e2.starttime);
    if (res != 0) {
      return res;
    }

    res = e1.enddate.compareTo(e2.enddate);
    if (res != 0) {
      return res;
    }

    return e1.endtime.compareTo(e2.endtime);
  }

  /** Use our compare method.
   */
  public int compareTo(Object o2) {
    return compare(this, o2);
  }

  public int hashCode() {
    return id;
  }

  /** We consider two events equal if their ids match
   */
  public boolean equals(Object obj) {
    if (!(obj instanceof EventVO)) {
      return false;
    }

    EventVO that = (EventVO)obj;

    return id == that.id;
  }

  public String toString() {
    StringBuffer sb = new StringBuffer();

    sb.append("EventVO{id=");
    sb.append(id);
    sb.append(", creator=");

    if (creator == null) {
      sb.append("**********null************");
    } else {
      sb.append(creator.getAccount());
    }
    sb.append(", isPublic=");
    sb.append(isPublic);
    sb.append(", lastmod=");
    sb.append(lastmod);
    sb.append(", seq=");
    sb.append(seq);
    sb.append(", created=");
    sb.append(created);
    sb.append(", startdate=");
    sb.append(startdate);
    sb.append(", starttime=");
    sb.append(starttime);
    sb.append(", enddate=");
    sb.append(enddate);
    sb.append(", editable=");
    sb.append(editable);
    sb.append(", kind=");
    sb.append(kindStr[kind]);
    sb.append(", guid=");
    sb.append(guid);
    sb.append(", shortdesc=");
    sb.append(shortdesc);
    sb.append(", recurrence(uwcal)=");
    sb.append(recurrence);
    sb.append(", recur=");
    sb.append(recur);
    sb.append("}");

    return sb.toString();
  }

  public Object clone() {
    EventVO ev = new EventVO(id, shortdesc, longdesc,
                             startdate, starttime, enddate, endtime,
                             isPublic, link,
                             (UserVO)creator.clone(),
                             cost,

                             /* This ought to be cloned but it brings with it a
                                whole set of instances. Leave for the moment */
                             recurrence,
                             null,
                             null,
                             seq, lastmod, created);
    LocationVO loc = getLocation();
    if (loc != null) {
      loc = (LocationVO)loc.clone();
    }
    SponsorVO sp = getSponsor();
    if (sp != null) {
      sp = (SponsorVO)sp.clone();
    }

    ev.setLocation(loc);
    ev.setSponsor(sp);

    Iterator it = iterateKeywords();
    TreeSet ks = new TreeSet();

    while (it.hasNext()) {
      KeywordVO k = (KeywordVO)it.next();

      ks.add((KeywordVO)k.clone());
    }

    ev.setKeywords(ks);

    return ev;
  }
}
