package edu.washington.cac.calfacade.shared;

import java.io.Serializable;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.Comparator;
import java.util.Set;

/** Value object to represent a calendar principal. Principals may be users,
 * groups or other special obects. Principals may own objects within the
 * system or simply identify a client to the system.
 *
 * <p>We need to address a problem that might occur with groups. If we choose
 * to allow a group all the facilities of a sinlge user (subscriptions,
 * ownership etc) then we need to be careful with names and their uniqueness.
 *
 * <p>That is to say, the name will probably not be unique, for example, I might
 * have the id douglm and be a member of the group douglm.
 *
 * <p>Allowing groups all the rights of a user gives us trhe current functionality
 * of administrative groups as well as the functions we need for departmental
 * sites.
 *
 *   @author Mike Douglass douglm@rpi.edu
 *  @version 1.0
 */
public class PrincipalVO implements Comparator, Serializable {
  private int id;

  private static final int invalidId = -1;

  /** The internal name by which this principal is identified, unique within
   * its class
   */
  private String account;  // null for guest

  protected Timestamp created;

  /** Last time we saw this principal appear in our system.
   */
  protected Timestamp logon;

  /** Last time we saw this principal did something in our system.
   */
  protected Timestamp lastAccess;

  /** Last time we saw this principal modified something in our system.
   */
  protected Timestamp lastModify;

  /** Populated by an explicit call - for the moment
   */
  protected Collection subscriptions;

  /** Value from edu.washington.cac.calfacade.shared.UserAuth
      Only affects PublicEvents administration
   */
  private int rights;

  /** True if this user is trying to acces public events.
   */
  private boolean publicEvents;

  /* ====================================================================
   *                   Constructors
   * ==================================================================== */

  /** Create a guest PrincipalVO
   */
  public PrincipalVO() {
    this(null, 0, 0, false);
  }

  /** Create a user vo usually based on a single string name from an object,
   * e.g., an event.
   *
   * @param  name            String user id for database
   */
  public PrincipalVO(String name) {
    this(name, 0, 0, false);
  }

  /** Create a user vo by specifying all its fields
   *
   * @param  account         String user account for database
    */
  public PrincipalVO(String account,
                     int id,
                     int rights,
                     boolean publicEvents) {
    this.account = account;
    this.id = id;
    this.rights = rights;
    this.publicEvents = publicEvents;
  }

  public void setId(int val) {
    id = val;
  }

  public int getId() {
    return id;
  }

  public void setAccount(String val) {
    account = val;
  }

  public String getAccount() {
    return account;
  }

  public void setCreated(Timestamp val) {
    created = val;
  }

  public Timestamp getCreated() {
    return created;
  }

  public void setLogon(Timestamp val) {
    logon = val;
  }

  public Timestamp getLogon() {
    return logon;
  }

  public void setLastAccess(Timestamp val) {
    lastAccess = val;
  }

  public Timestamp getLastAccess() {
    return lastAccess;
  }

  public void setLastModify(Timestamp val) {
    lastModify = val;
  }

  public Timestamp getLastModify() {
    return lastModify;
  }

  /** Set of calendars principal is subscribed to
   *
   * @param val        Collection of CalendarVO
   */
  public void setSubscriptions(Collection val) {
    subscriptions = val;
  }

  /** Get the calendars principal is subscribed to
   *
   * @return Collection    of CalendarVO
   */
  public Collection getSubscriptions() {
    return subscriptions;
  }

  public boolean isGuest() {
    return account == null;
  }

  /** Set the rights this user has
   *
   * @param val    int access rights value as defined by
   *               edu.washington.cac.calfacade.shared.UserAuth
   */
  public void setRights(int val) {
    rights = val;
  }

  /** Get the rights this user has
   *
   * @return int   value as defined by
   *               edu.washington.cac.calfacade.shared.UserAuth
   */
  public int getRights() {
    return rights;
  }

  /** Show if this user accessing public events
   */
  public void setPublicEvents(boolean val) {
    publicEvents = val;
  }

  /** Is this user accessing public events
   */
  public boolean getPublicEvents() {
    return publicEvents;
  }

  /** Check for priv user
   *
   * @return boolean    true for super user
   */
  public boolean isSuperUser() {
    return (rights & UserAuth.superUser) != 0;
  }

  /** Check for alert user
   *
   * @return boolean    true for alert user
   */
  public boolean isAlertUser() {
    return (rights & UserAuth.alertUser) != 0;
  }

  /** Check for public events owner user
   *
   * @return boolean    true for public events owner user
   */
  public boolean isOwnerUser() {
    return (rights & UserAuth.publicEventUser) != 0;
  }

  /** Check for content admin user
   *
   * @return boolean    true for content admin user
   */
  public boolean isContentAdminUser() {
    return (rights & UserAuth.contentAdminUser) != 0;
  }

  /* ====================================================================
   *                   Copying methods
   * ==================================================================== */

  public void copyTo(PrincipalVO val) {
    /* We do not copy the attached subscriptions if present. These need to
       be cloned explicitly or we might set up a clone loop.
    */
    val.setAccount(getAccount());
    val.setId(getId());
    val.setCreated(getCreated());
    val.setLogon(getLogon());
    val.setLastAccess(getLastAccess());
    val.setLastModify(getLastModify());
    val.setRights(getRights());
    val.setPublicEvents(getPublicEvents());
  }

  /* ====================================================================
   *                   Object methods
   * ==================================================================== */

  /** Override the default and sort on account
   */
  public int compare(Object o1, Object o2) {
    if (!(o1 instanceof PrincipalVO)) {
      return -1;
    }

    if (!(o2 instanceof PrincipalVO)) {
      return 1;
    }

    PrincipalVO p1 = (PrincipalVO)o1;
    PrincipalVO p2 = (PrincipalVO)o2;

    if (p1.account == null) {
      // Should not happen
      throw new RuntimeException("Null account for UserVO object id=" +
                                 p1.id);
    }

    if (p2.account == null) {
      // Should not happen
      throw new RuntimeException("Null account for UserVO object id=" +
                                 p2.id);
    }

    return p1.account.compareTo(p2.account);
  }

  public String toString() {
    StringBuffer sb = new StringBuffer();

    sb.append("UserVO{id=");
    sb.append(id);
    sb.append(", account=");
    sb.append(account);
    sb.append(", created=");
    sb.append(created);
    sb.append(", logon=");
    sb.append(logon);
    sb.append(", lastAccess=");
    sb.append(lastAccess);
    sb.append(", lastModify=");
    sb.append(lastModify);
    sb.append("}");

    return sb.toString();
  }

  /** Check for equality
   *
   * @param  o          Object to compare.
   * @return true       if object represents same user.
   */
  public boolean equals(Object o) {
    if (!(o instanceof PrincipalVO)) {
      return false;
    }

    PrincipalVO that = (PrincipalVO)o;

    if (account == null) {
      return that.account == null;
    }

    return account.equals(that.account);
  }

  public int hashCode() {
    int hc = 1;

    if (account != null) {
      hc = account.hashCode();
    }

    return hc;
  }

  public Object clone() {
    /* We do not clone the attached subscriptions if present. These need to
       be cloned explicitly or we might set up a clone loop.
    */
    PrincipalVO p = new PrincipalVO();
    copyTo(p);

    return p;
  }
}
