package edu.washington.cac.calfacade.shared.filter;

import edu.washington.cac.calfacade.shared.CalFacadeDefs;
import edu.washington.cac.calfacade.shared.EventVO;

import java.io.Serializable;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

/**
  A filter selects events (and possibly other entities) that fulfill
  certain criteria.  For example, "All events   that have the keyword
  'Lecture'".

  Some filters can have any number of children such as an OrFilter.

  @author Greg Barnes
  @version 1.1
 */
public class FilterVO implements Comparable, Comparator, Serializable {
  protected int id = CalFacadeDefs.unsavedItemKey;

  /** The internal name of the filter
   */
  protected String name;

  /** A printable title for the calendar
   */
  protected String title;

  /** Some sort of description - may be null
   */
  protected String description;

  protected boolean showChildren;

  protected FilterVO parent;

  /** The children of the filter */
  protected Collection children;

  public FilterVO() {
  }

  public FilterVO(String name, FilterVO parent) {
    this.name =  name;
    this.parent =  parent;
  }

  public FilterVO(int id, String name, FilterVO parent) {
    this.id =  id;
    this.name =  name;
    this.parent =  parent;
  }

  /* ====================================================================
   *                   Bean methods
   * ==================================================================== */

  public void setId(int val) {
    id = val;
  }

  public int getId() {
    return id;
  }

  /** Set the name
   *
   * @param val    String name
   */
  public void setName(String val) {
    name = val;
  }

  /** Get the name
   *
   * @return String   name
   */
  public String getName() {
    return name;
  }

  /** Set the title
   *
   * @param val    String title
   */
  public void setTitle(String val) {
    title = val;
  }

  /** Get the title
   *
   * @return String   title
   */
  public String getTitle() {
    return title;
  }

  /** Set the description
   *
   * @param val    description
   */
  public void setDescription(String val) {
    description = val;
  }

  /** Get the description
   *
   *  @return String   description
   */
  public String getDescription() {
    return description;
  }

  /** Should the children of this calendar be displayed?
   *
   * @param val        boolean true if calendar should be displayed
   */
  public void setShowChildren(boolean val) {
    showChildren = val;
  }

  /** Should the children of this calendar be displayed?
   * @return boolean     true if calendar should be displayed
   */
  public boolean getShowChildren() {
    return showChildren;
  }

  /** Set the parent for this calendar
   *
   * @param val   FilterVO parent object
   */
  public void setParent(FilterVO val) {
    parent = val;
  }

  /** Get the parent
   *
   * @return FilterVO    the parent
   */
  public FilterVO getParent() {
    return parent;
  }

  /**  Set the set of children
   *
   * @param   val   Collection of children for this filter
   */
  public void setChildren(Collection val) {
    children = val;
  }

  /**  Get the set of children
   *
   * @return Collection   FilterVO children for this filter
   */
  public Collection getChildren() {
    if (children == null) {
      children = new TreeSet();
    }

    return children;
  }

  /* ====================================================================
   *                   Convenience methods
   * ==================================================================== */

  /**  Add a child to the set of children
   *
   * @param val     FilterVO child
   */
  public void addChild(FilterVO val) {
    getChildren().add(val);
  }

  /* ====================================================================
   *                   Abstract methods
   * ==================================================================== */

  /** Test if this event passes the test
   *
   * @param e Event to test
   * @return If the filter selects the event
   */
   public boolean match(EventVO e) {throw new Error("Unimplemented");}

  /* ====================================================================
   *                   Copying methods
   * ==================================================================== */

  protected void copyTo(FilterVO val) {
    val.setId(getId());
    val.setName(getName());
    val.setTitle(getTitle());
    val.setDescription(getDescription());
    val.setShowChildren(getShowChildren());
    val.setParent(getParent()); // We clone downwards

    Iterator it = getChildren().iterator();

    while (it.hasNext()) {
      val.addChild((FilterVO)((FilterVO)it.next()).clone());
    }
  }

  /** Clone a complete collection of filters.
   */
  public static Collection clone(Collection val) {
    Set s = new TreeSet();

    if (val != null) {
      Iterator it = val.iterator();

      while (it.hasNext()) {
        s.add(((FilterVO)it.next()).clone());
      }
    }

    return s;
  }

  /* ====================================================================
   *                   Object methods
   * ==================================================================== */

  /** We'll provide default equals and compare methods based on the
   * id. Overide them for different ordering.
   */
  public int compare(Object o1, Object o2) {
    if (!(o1 instanceof FilterVO)) {
      return -1;
    }

    if (!(o2 instanceof FilterVO)) {
      return 1;
    }

    FilterVO f1 = (FilterVO)o1;
    FilterVO f2 = (FilterVO)o2;

    if (f1.id < f2.id) {
      return -1;
    }

    if (f1.id > f2.id) {
      return 1;
    }

    return 0;
  }

  /** Compare this filter and an object
   *
   * @param  o    object to compare.
   * @return int -1, 0, 1
   */
  public int compareTo(Object o) {
    if (!(o instanceof FilterVO)) {
      return -1;
    }

    FilterVO that = (FilterVO)o;

    if (id < that.id) {
      return -1;
    }

    if (id > that.id) {
      return 1;
    }

    return 0;
  }

  /** The id does not take part in these comparisons because we're presumably
   * doing this to see if two objects are the same without doing a db retrieval.
   * That is, one of the objects in a comparison may not have the id set.
   *
   * <p>This is relevant for hashcode as the equals method requires that
   * hashcode return equal values for both objects in a comparison that are
   * to be considered equal.
   *
   * <p>As we have both objects to hand in the equals method we can check to
   * see if both ids are non-zero and use them in the comparison.
   */
  public int hashCode() {
    int hc = 1;

    if (id != 0) {
      hc *= id;
    }

    if (name != null) {
      hc *= name.hashCode();
    }

    if (title != null) {
      hc *= title.hashCode();
    }

    if (description != null) {
      hc *= description.hashCode();
    }

    return hc;
  }

  /** Is this filter the same as another object?
   * @param o        Object to check against
   * @return boolean true this filter is the same as the object
   */
  public boolean equals(Object obj) {
    if (!(obj instanceof CalendarVO)) {
      return false;
    }

    CalendarVO that = (CalendarVO)obj;

    if ((id != 0) && (that.id != 0)) {
      if (id != that.id) {
        return false;
      }
    }

    if (name == null) {
      if (that.name != null) {
        return false;
      }
    } else if (!name.equals(that.name)) {
      return false;
    }

    /*
    if (title == null) {
      if (that.title != null) {
        return false;
      }
    } else if (!title.equals(that.title)) {
      return false;
    }

    if (description == null) {
      if (that.description != null) {
        return false;
      }
    } else if (!description.equals(that.description)) {
      return false;
    }*/

    return true;
  }

  public String toString() {
    StringBuffer sb = new StringBuffer();

    sb.append("FilterVO{name=");
    sb.append(String.valueOf(name));
    //sb.append(", title=");
    //sb.append(String.valueOf(title));
    //sb.append(", description=");
    //sb.append(String.valueOf(description));
    sb.append(", parent=");
    if (parent == null) {
      sb.append("null");
    } else {
      sb.append(parent.getId());
    }
    sb.append("}");

    return sb.toString();
  }

  public Object clone() {
    FilterVO f = new FilterVO();

    copyTo(f);

    return f;
  }
}
