/* **********************************************************************
    Copyright 2003 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.uwcal.testsynch;

import edu.rpi.cct.uwcal.calsvci.CalSvcI;
import edu.rpi.cct.uwcal.common.URIgen;
import edu.rpi.cct.uwcal.icalendar.IcalTranslator;
import edu.rpi.cct.uwcal.synch.common.IcalSynchState;
import edu.rpi.cct.uwcal.synch.common.IcalSynchState;
import edu.rpi.cct.uwcal.synch.common.Synch;
import edu.rpi.cct.uwcal.webcommon.UWCalWebURIgen;

import edu.tests.rpi.cct.uwcal.calsvci.CalSvcTestUtil;

import edu.washington.cac.calfacade.shared.CalFacadeDefs;
import edu.washington.cac.calfacade.shared.GuidVO;
import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.filter.FilterVO;
import edu.washington.cac.calfacade.shared.filter.KeywordFilterVO;
import edu.washington.cac.calfacade.shared.filter.OrFilterVO;
import edu.washington.cac.calfacade.shared.KeywordVO;
import edu.washington.cac.calfacade.shared.LocationVO;
import edu.washington.cac.calfacade.shared.UserAuth;
import edu.washington.cac.calfacade.shared.UserVO;

import net.fortuna.ical4j.model.component.VEvent;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.TreeSet;

/** Stand alone package to drive the synchronization interface. This can
 * set up the database in a known state in preparation for tests then run
 * some basic tests and check the end state.
 *
 * <p>We'd probably do better to drive a subset of the junit test suite.
 *
 * @author Mike Douglass       douglm@rpi.edu
 */
public class SynchTest {
  private boolean debug = true;

  /** If true we (re)initialise the db by adding some specific events for
   * a known set of users.
   */
  private boolean initdb = false;

  /** If true we are running with hibernate (means we have to add some extra
   * data).
   */
  private boolean hibernate = false;

  /** if true we run a set of tests on the data. Implies initdb.
   *
   * <p>Use -runtest -noinitdb to prevent reinitialisation of the data.
   */
  private boolean runtest = false;

  private String deviceId1 = "SynchTest1";

  private String deviceId2 = "SynchTest2";

  private URIgen urigen;

  private Synch synch;

  /* data for the public calendar
   */
  private static String publicOwner = "caladmin";

  private static final int numPublicEvents = 10;

  /** There are numPublicEvents of each of these starting with this id
   */
  private int firstPublicLoc;
  private int firstPublicSp;
  private int firstPublicEvent;

  /* data for the private calendar
   */
  private static String privateUser = "caluser";

  private int firstPrivateEvent;
  private int numPrivateEvents;

  private int firstReffedEvent;

  private static String[] userCalText = {
    "BEGIN:VCALENDAR",
    "PRODID:-//Jahia Solutions//iCalendar SyncClient 1.0 MIMEDIR//EN",
    "VERSION:2.0",
    "METHOD:PUBLISH",
    "BEGIN:VEVENT",
    "DTSTART;VALUE=DATE:20050116",
    "DTEND;VALUE=DATE:20050116",
    "SUMMARY:Dummy SyncItem 0 (data0)",
    "UID:DUMMY-SYNC-ITEM-20-FOR-TESTING",
    "DTSTAMP:20040707T140415Z",
    "END:VEVENT",
    "BEGIN:VEVENT",
    "DTSTART;VALUE=DATE:20050116",
    "DTEND;VALUE=DATE:20050116",
    "SUMMARY:Dummy SyncItem 1 (data1)",
    "UID:DUMMY-SYNC-ITEM-20-FOR-TESTING",
    "DTSTAMP:20040707T140415Z",
    "END:VEVENT",
    "BEGIN:VEVENT",
    "DTSTART;VALUE=DATE:20050217",
    "DTEND;VALUE=DATE:20050218",
    "SUMMARY:Dummy SyncItem 2 (data2)",
    "UID:DUMMY-SYNC-ITEM-30-FOR-TESTING",
    "DTSTAMP:20040707T140415Z",
    "END:VEVENT",
    "BEGIN:VEVENT",
    "DTSTART;VALUE=DATE:20050215",
    "DTEND;VALUE=DATE:20050216",
    "SUMMARY:Dummy SyncItem 3 (data3)",
    "UID:9DE4B99A-D020-11D8-9530-000A958A3252",
    "DTSTAMP:20040707T140415Z",
    "END:VEVENT",
    "END:VCALENDAR",
  };

  private KeywordVO concertKey;
  private KeywordVO forumsKey;

  /* -------------------------- Admin (public) stuff        ---------------*/

  private CalSvcTestUtil svciUtil;

  private static String adminUser = "caladmin";

  /** Called after parameters are set
   */
  public void testSynch() {
    String state = "";

    try {
      urigen = new UWCalWebURIgen("http://cal.rpi.edu");

      if (initdb) {
        state = "initdb";

        doInitdb();
      }

      if (runtest) {
        state = "runtest";

        doRuntest();
      }
    } catch (Throwable t) {
      log("Exception: state=" + state);
      t.printStackTrace();
    }
  }

  void processArgs(String[] args) throws Throwable {
    if (args == null) {
      return;
    }

    for (int i = 0; i < args.length; i++) {
      if (args[i].equals("-debug")) {
        debug = true;
      } else if (args[i].equals("-ndebug")) {
        debug = false;
      } else if (args[i].equals("-initdb")) {
        initdb = true;
      } else if (args[i].equals("-hibernate")) {
        hibernate = true;
      } else if (args[i].equals("-noinitdb")) {
        initdb = false;
      } else if (args[i].equals("-runtest")) {
        runtest = true;
        initdb = true;
      } else {
        log("Illegal argument: " + args[i]);
        usage();
        throw new Exception("Invalid args");
      }
    }
  }

  boolean argpar(String n, String[] args, int i) throws Exception {
    if (!args[i].equals(n)) {
      return false;
    }

    if ((i + 1) == args.length) {
      throw new Exception("Invalid args");
    }
    return true;
  }

  public static void main(String[] args) {
    SynchTest st = null;

    try {
      st = new SynchTest();

      st.processArgs(args);

      st.testSynch();
    } catch (Throwable t) {
      t.printStackTrace();
    } finally {
      try {
//        d.close();
      } catch (Throwable t1) {
        t1.printStackTrace();
      }
    }
  }

  /* ====================================================================
   *                       Private methods.
   * ==================================================================== */

  private void usage() {
    log("Usage:");
    log("  -[n]debug     set debugging on [off]");
    log("  -initdb       set some initial data in the db");
    log("  -runtest      run some tests - implies initdb.");
    log("                Use -runtest -noinitdb to prevent reinitialisation of the data.");
    log("");
  }

  /* ====================================================================
   *                       Private location methods.
   * ==================================================================== */

  /** Add a bunch of locations for later use
   *
   * @param n     int number of locations
   * @return int  key of first added (remainder should follow)
   */
  private int addLocations(int n,
                            boolean isPublic) {
    int first = -1;
    String user;

    if (isPublic) {
      user = publicOwner;
    } else {
      user = privateUser;
    }

    for (int i = 1; i <= n; i++) {
      int key = svciUtil.addLocation("Dummy location address " + i,
                                     "Dummy location subaddress " + i,
                                     "http://dummy.link" + i + ".edu",
                                     user,
                                     isPublic);

      if (key < 0) {
        // failure
        return key;
      }

      if (first < 0) {
        first = key;
      }
    }

    return first;
  }

  /* ====================================================================
   *                       Private sponsor methods.
   * ==================================================================== */

  /** Add a bunch of sponsors for later use
   *
   * @param n     int number of sponsors
   * @return int  key of first added (remainder should follow)
   */
  private int addSponsors(int n,
                          boolean isPublic) {
    int first = -1;
    String user;

    if (isPublic) {
      user = publicOwner;
    } else {
      user = privateUser;
    }

    for (int i = 1; i <= n; i++) {
      int key = svciUtil.addSponsor("Dummy sponsor name " + i,
                                    "Dummy sponsor phone " + i,
                                    "Dummy sponsor email " + i,
                                    "http://dummy.link" + i + ".edu",
                                    user,
                                    isPublic);

      if (key < 0) {
        // failure
        return key;
      }

      if (first < 0) {
        first = key;
      }
    }

    return first;
  }

  /* ====================================================================
   *                       Private keyword methods.
   * ==================================================================== */

  /** Add a bunch of sponsors for later use
   *
   * @param n     int number of sponsors
   * @return int  key of first added (remainder should follow)
   */
  private int addKeywords(int n,
                          boolean isPublic) {
    int first = -1;
    String user;

    if (isPublic) {
      user = publicOwner;
    } else {
      user = privateUser;
    }

    for (int i = 1; i <= n; i++) {
      int key = svciUtil.addKeyword("Dummy keyword " + i,
                                    "Dummy keyword description " + i,
                                    user,
                                    isPublic);

      if (key < 0) {
        // failure
        return key;
      }

      if (first < 0) {
        first = key;
      }
    }

    return first;
  }

  private void addCalendars() throws Throwable {

    if (!hibernate) {
      svciUtil.open(true);

      String[] calendarDefs = {
         "(2, 'Main Calendars', 'O', 0, 0) ",
         "  (100, 'General Calendars', 'O', 0, 2) ",
         "    (1000, 'Academic Calendar', 'K', 35, 100) ",
         "    (1100, 'Alerts', 'K', 33, 100) ",
         "    (1200, 'Deadlines', 'K', 22, 100) ",
         "    (1300, 'Holidays', 'K', 34, 100) ",
         "  (110, 'Arts (all)', 'O', 0, 2) ",
         "    (1400, 'Concerts', 'K', 3, 110) ",
         "    (1500, 'Dance', 'K', 4, 110) ",
         "    (1600, 'Drama', 'K', 5, 110) ",
         "    (1700, 'Exhibits', 'K', 6, 110) ",
         "    (1800, 'Films', 'K', 7, 110) ",
         "    (1900, 'Operas', 'K', 12, 110) ",
         "    (2000, 'Readings', 'K', 14, 110) ",
         "  (120, 'Conferences/Meetings (all)', 'O', 0, 2) ",
         "    (2100, 'Conferences', 'K', 24, 120) ",
         "    (2200, 'Forums', 'K', 25, 120) "
      };

      /* Add the calendar defs */
      for (int i =  0; i < calendarDefs.length; i++) {
        svciUtil.getSvci(true).addCalendarDef(calendarDefs[i]);
      }

      svciUtil.close(true);

      return;
    }

    CalSvcI svci = svciUtil.getSvci(true);

    FilterVO filterRoot = new OrFilterVO(2, "Main Calendars", null);

    svciUtil.open(true);
    svci.addFilter(filterRoot);
    svciUtil.close(true);

    FilterVO filterGeneral = new OrFilterVO(100, "General Calendars", filterRoot);
    FilterVO filterArts = new OrFilterVO(110, "Arts (all)", filterRoot);
    FilterVO filterMeetings =
        new OrFilterVO(120, "Conferences/Meetings (all)", filterRoot);

    svciUtil.open(true);
    filterRoot.addChild(filterArts);
    filterRoot.addChild(filterMeetings);
    svci.addFilter(filterArts);
    svci.addFilter(filterMeetings);
    svci.updateFilter(filterRoot);
    svciUtil.close(true);

    KeywordFilterVO filterConcerts =
        new KeywordFilterVO(1400, "Concerts", filterArts, concertKey);

    KeywordFilterVO filterForums =
        new KeywordFilterVO(2200, "Forums", filterMeetings, forumsKey);

    svciUtil.open(true);
    svci.addFilter(filterConcerts);
    filterArts.addChild(filterConcerts);
    svci.updateFilter(filterArts);
    svciUtil.close(true);

    svciUtil.open(true);
    svci.addFilter(filterForums);
    filterMeetings.addChild(filterForums);
    svci.updateFilter(filterMeetings);
    svciUtil.close(true);
  }

  /* ====================================================================
   *                       Private event methods.
   * ==================================================================== */

  /** Add a bunch of events for later use
   *
   * @param n         int number of events
   * @param locn      int first location id
   * @param spn       int first sponsor id
   * @param isPublic  true to create public events
   * @return int      key of first added (remainder should follow)
   */
  private int addEvents(int n, int locn, int spn,
                        boolean isPublic) throws Throwable {
    int first = -1;
    String user;

    if (isPublic) {
      user = publicOwner;
    } else {
      user = privateUser;
    }

    for (int i = 1; i <= n; i++) {
      KeywordVO keywd;
      if (i < 3) {
        keywd = concertKey;
      } else {
        keywd = forumsKey;
      }

      int key = svciUtil.addEvent("Dummy event " + i,
                                     "Dummy event description " + i,
                                     "2005-02-06 13:00:00",
                                     "2005-02-06 14:00:00",
                                     svciUtil.getLocation(locn - 1 + i, true),
                                     svciUtil.getSponsor(spn - 1 + i, true),
                                     keywd,
                                     "http://dummy.link" + i + ".edu",
                                     user,
                                     isPublic);

      if (key < 0) {
        // failure
        return key;
      }

      if (first < 0) {
        first = key;
      }
    }

    return first;
  }

  private void updatePublic(int evid) throws Throwable {
    logResult("updatePublic event " + evid);

    CalSvcI svci = svciUtil.getSvci(publicOwner, UserAuth.superUser, true);
    svciUtil.open(true);

    EventVO ev = svci.getEvent(evid);
    ev.setLongdesc(String.valueOf(ev.getLongdesc()) + "X");
    svci.updateEvent(ev, null);

    svciUtil.close(true);
  }

  private void updatePublicSynch(int evid) throws Throwable {
    logResult("updatePublicSynch event " + evid);

    CalSvcI svci = svciUtil.getSvci(publicOwner, UserAuth.superUser, true);
    svciUtil.open(true);

    EventVO ev = svci.getEvent(evid);
    ev.setLongdesc(String.valueOf(ev.getLongdesc()) + "X");
    svci.updateEvent(ev, null);

    svciUtil.close(true);
  }

  private void addRef(int evid) throws Throwable {
    logResult("addRef for event " + evid);

    CalSvcI svci = svciUtil.getSvci(privateUser, 0, false);
    svciUtil.open(false);

    EventVO ev = svci.getEvent(evid);
    svci.addEventRef(ev);

    svciUtil.close(false);
  }

  private void deleteRef(int evid) throws Throwable {
    logResult("deleteRef for event " + evid);

    CalSvcI svci = svciUtil.getSvci(privateUser, 0, false);
    svciUtil.open(false);

    EventVO ev = svci.getEvent(evid);
    svci.deleteEventRef(ev);

    svciUtil.close(false);
  }

  /** Add an event to the db for the current user. We will get a guid and
   * merge the new guid into the supplied data
   *
   * @param  calText    String icalendar format events
   */
  private void addEvents(String[] calText) throws Throwable {
    StringBuffer sb = new StringBuffer();
    firstPrivateEvent = -1;
    String guid = null;

    for (int i = 0; i < calText.length; i++) {
      String ln = calText[i];

/*      if (ln.startsWith("UID")) {
        guid = synch.getNewGuid();

        log("Obtained new guid: " + guid);

        sb.append("UID:");
        sb.append(guid);
        if (firstPrivateEvent == -1) {
          GuidVO gv = GuidVO.valueOf(guid);
          firstPrivateEvent = gv.getId();
        }
      } else {
        sb.append(ln);
      }
*/
      sb.append(ln);
      sb.append("\r\n");
    }

    IcalTranslator trans = synch.getTranslator();

    Collection vevs = trans.toVEvent(sb.toString());

    Iterator it = vevs.iterator();

    while (it.hasNext()) {
      guid = synch.getNewGuid();
      log("Obtained new guid: " + guid);

      if (firstPrivateEvent == -1) {
        GuidVO gv = GuidVO.valueOf(guid);
        firstPrivateEvent = gv.getId();
      }

      synch.updateEvent((VEvent)it.next(), guid);
      numPrivateEvents++;
    }
  }

  /* ====================================================================
   *                       Private test methods.
   * ==================================================================== */

  private void doInitdb() throws Throwable {
    /* Public stuff first so we can refer to it for private user.
     */
    svciUtil = new CalSvcTestUtil(debug);

    svciUtil.getSvci(publicOwner, UserAuth.superUser, true);

    if (hibernate) {
      addMoreData();
    }

    svciUtil.open(true);

    concertKey = getKeyword("Concerts");
    forumsKey = getKeyword("Forums");

    svciUtil.close(true);

    /* Add the calendar defs */
    addCalendars();

    svciUtil.open(true);

    /* Add a bunch of public locations */
    int numLocs = numPublicEvents;
    firstPublicLoc = addLocations(numLocs, true);

    /* Add a bunch of public sponsors */
    int numSps = numPublicEvents;
    firstPublicSp = addSponsors(numSps, true);

    /* Add a bunch of public events */
    int numEvs = numPublicEvents;
    firstPublicEvent = addEvents(numEvs, firstPublicLoc, firstPublicSp, true);

    svciUtil.close(true);

    /* Now private stuff.
     */

    /* get a synch object for the private user.
     */

    getSynch(deviceId1);
    synch.open();
    synch.beginTransaction();

    addEvents(userCalText);
    logResult("Added " + numPrivateEvents + " starting at " + firstPrivateEvent);
    synch.commit();
    synch.close();

    /** Add two event refs */

    firstReffedEvent = firstPublicEvent + 3;
    addRef(firstReffedEvent);

    addRef(firstReffedEvent + 1);

    /* Subscribe to the concerts calendar */

    CalSvcI userSvci = svciUtil.getSvci(privateUser, 0, false);
    svciUtil.open(false);

    TreeSet subs = new TreeSet();
    subs.add(new Integer(1400));

    userSvci.fixSubscriptions(subs);

    svciUtil.close(false);

    /* *** temp ... make this a unit test - test checkKeywordRefs indirectly
    svciUtil.open(true);
    svciUtil.getSvci(true).deleteKeyword(concertKey);
    svciUtil.close(true);
    */
  }

  private KeywordVO getKeyword(String wd) {
    return svciUtil.getKeyword(wd, true);
  }

  private void addMoreData() {
    svciUtil.open(true);
    CalSvcI svci = svciUtil.getSvci(true);

    svciUtil.addKeyword("Concerts", "Concerts", publicOwner, true);
    svciUtil.addKeyword("Forums", "Forums", publicOwner, true);

    svciUtil.close(true);
  }

  /** We're set up with:
   *   a number of personal events +
   *       2  events from subscriptions +
   *       2  events from references.
   */
  private void doRuntest() throws Throwable {
    /** A get all from any device should give us numPrivateEvents + 4
       (personal events + 2 subscriptions + 2 references).
     */
    getEvents(deviceId1, true);

    /** A get all from any device should give us numPrivateEvents + 4
       (personal events + 2 subscriptions + 2 references).
     */
    getEvents(deviceId2, true);

    /* Try again - should get nothing */
    getEvents(deviceId1, false);

    /* Modify the public event */
    updatePublic(firstPublicEvent);

    /* Try again - should get one */
    getEvents(deviceId1, false);

    /* Delete the second public event (ref) */
    deleteRef(firstReffedEvent + 1);

    /* Try again - should get a deleted notification */
    getEvents(deviceId1, false);

    /* Try again - should get nothing */
    getEvents(deviceId1, false);

//    /* Delete a subscribed event via synch */
 //   deleteEvent(firstPublicEvent);

    /* Update the public event via synch */
    updatePublicSynch(firstPublicEvent, deviceId1);

    /* Try again - should get nothing */
    getEvents(deviceId1, false);

    /* Try for device 2 - should get the mod and the deleted */
    getEvents(deviceId2, false);

    /* Try again - should get nothing */
    getEvents(deviceId2, false);
  }

  private void getEvents(String dev, boolean all) throws Throwable {
    getSynch(dev);
    synch.open();
    synch.beginTransaction();

    Collection c;
    if (all) {
      c = synch.getAllVevents();
    } else {
      c = synch.getAllChangedVevents();
    }

    logResult("getEvents (all=" + all + ") for " + dev + " returned " + c.size());

    Iterator it = c.iterator();
    while (it.hasNext()) {
      IcalSynchState iss = (IcalSynchState)it.next();

      logResult("Got " + iss);
    }
    synch.commit();
    synch.close();
  }

  private void updatePublicSynch(int evid, String dev) throws Throwable {
    logResult("updatePublicSynch event " + evid);

    CalSvcI svci = svciUtil.getSvci(publicOwner, UserAuth.superUser, true);
    svci.open();

    EventVO ev = svci.getEvent(evid);

    svci.close();

    getSynch(dev);
    synch.open();
    synch.beginTransaction();

    IcalTranslator trans = synch.getTranslator();

    ev.setLongdesc(String.valueOf(ev.getLongdesc()) + "XXXX");
    LocationVO loc = ev.getLocation();
    if (loc != null) {
      loc.setAddress("Synchtest modified value");
    }

    VEvent vev = trans.toIcalEvent(ev);

    synch.updateEvent(vev, ev.getGuid());

    synch.commit();
    synch.close();
  }

  /* ====================================================================
   *                       Misc private methods.
   * ==================================================================== */

  private void getSynch(String devid) throws Throwable {
    synch = new Synch(privateUser, devid, urigen, debug);
  }

  private void logResult(String msg) {
    log("-----------------------------------------------------------------");
    log("-- " + msg);
    log("-----------------------------------------------------------------");
  }

  private void log(String msg) {
    System.out.println(msg);
  }
}

