package edu.rpi.cct.uwcal.tools.dump;

import java.sql.Timestamp;

/** Some definitions for dumped data.
 *
 * <p>The data is dumped in one section per object class. Dumped objects
 * refer to other objects by their key (usually an integer).
 *
 * <p>Restoring the data will take a couple of passes over some of it. For
 * example, the user entries ref to subscriptions which refer to filters
 * which refer to locations and sponsors and keywords.
 *
 * <p>Restore the users, save the subscriptions and then do them at the end.
 *
 * @author Mike Douglass   douglm@rpi.edu
 * @version 1.0
 */
public interface DumpDefs {
  public static final int dbTypeDb2 = 0;
  public static final int dbTypeHsql = 1;
  public static final int dbTypeMysql = 2;
  public static final int dbTypeOracle = 3;
  public static final int dbTypePostgres = 4;
  public static final int dbTypeSqlServer = 5;

  /** ===================================================================
   *                      Tag name for entire dump
   *  =================================================================== */

  public static final String dumpTag = "caldata";

  /** ===================================================================
   *                      Tag names for each section
   *  =================================================================== */

  public static final String sectionFilters = "filters";
  public static final String sectionUsers = "users";
  public static final String sectionLocations = "locations";
  public static final String sectionSponsors = "sponsors";
  public static final String sectionKeywords = "keywords";
  public static final String sectionAuthUsers = "authusers";
  public static final String sectionEvents = "events";
  public static final String sectionEventRefs = "eventRefs";
  public static final String sectionAdminGroups = "adminGroups";
  public static final String sectionDbLastmods = "dblastmods";

  /** ===================================================================
   *                      Tag names for each object
   *  =================================================================== */

  public static final String objectAliasFilter = "aliasFilter";
  public static final String objectAndFilter = "andFilter";
  public static final String objectKeywordFilter = "keyFilter";
  public static final String objectCreatorFilter = "creatorFilter";
  public static final String objectLocationFilter = "locationFilter";
  public static final String objectOrFilter = "orFilter";
  public static final String objectSponsorFilter = "sponsorFilter";
  public static final String objectNotFilter = "notFilter";

  public static final String objectUser = "user";
  public static final String objectLocation = "location";
  public static final String objectSponsor = "sponsor";
  public static final String objectKeyword = "keyword";
  public static final String objectAuthUser = "authuser";
  public static final String objectEvent = "event";
  public static final String objectEventRef = "eventRef";
  public static final String objectAdminGroup = "adminGroup";
  public static final String objectDbLastmod = "dblastmod";

  /** ===================================================================
   *                      Field types
   *  =================================================================== */

  public static final int ftypeSpecial = 0;
  public static final int ftypeInt = 1;
  public static final int ftypeTimestamp = 2;
  public static final int ftypeDate = 3;
  public static final int ftypeTime = 4;
  public static final int ftypeBoolean = 5;
  public static final int ftypeString = 6;
  public static final int ftypeChar = 7;

  /** ===================================================================
   *                      Field definition
   *  =================================================================== */

   public static class DumpFieldDef {
     String tagName;
     int ftype;

     /** If true we set the value in current Val.
      *
      * <p>If false we fail on non-whitespace
      */
     boolean hasVal = true;

     /** If true we ignore completely white space characters
      */
     boolean ignoreWhiteSpace = true;

     DumpFieldDef(String tagName, int ftype) {
       this.tagName = tagName;
       this.ftype = ftype;
     }

     DumpFieldDef(String tagName, int ftype, boolean hasVal) {
       this.tagName = tagName;
       this.ftype = ftype;
       this.hasVal = hasVal;
     }
   }

  /** ===================================================================
   *                      Field definitions per class
   *  =================================================================== */

  public static final DumpFieldDef[] userFdefs = {
    new DumpFieldDef("id", ftypeInt),
    new DumpFieldDef("account", ftypeString),
    new DumpFieldDef("created", ftypeTimestamp),
    new DumpFieldDef("logon", ftypeTimestamp),
    new DumpFieldDef("lastAccess", ftypeTimestamp),
    new DumpFieldDef("lastModify", ftypeTimestamp),
    new DumpFieldDef("subscriptions", ftypeSpecial, false),
    new DumpFieldDef("calendarid", ftypeInt)
  };

  public static final DumpFieldDef[] locationFdefs = {
    new DumpFieldDef("id", ftypeInt),
    new DumpFieldDef("creator", ftypeInt),
    new DumpFieldDef("public", ftypeBoolean),
    new DumpFieldDef("address", ftypeString),
    new DumpFieldDef("subaddress", ftypeString),
    new DumpFieldDef("link", ftypeString)
  };

  public static final DumpFieldDef[] sponsorFdefs = {
    new DumpFieldDef("id", ftypeInt),
    new DumpFieldDef("creator", ftypeInt),
    new DumpFieldDef("public", ftypeBoolean),
    new DumpFieldDef("name", ftypeString),
    new DumpFieldDef("phone", ftypeString),
    new DumpFieldDef("email", ftypeString),
    new DumpFieldDef("link", ftypeString)
  };

  public static final DumpFieldDef[] keywordFdefs = {
    new DumpFieldDef("id", ftypeInt),
    new DumpFieldDef("creator", ftypeInt),
    new DumpFieldDef("public", ftypeBoolean),
    new DumpFieldDef("word", ftypeString),
    new DumpFieldDef("title", ftypeString),
    new DumpFieldDef("description", ftypeString),
    new DumpFieldDef("attribute", ftypeSpecial, false),
    new DumpFieldDef("name", ftypeString),
    new DumpFieldDef("val", ftypeString)
  };

  public static final DumpFieldDef[] authUserFdefs = {
    new DumpFieldDef("id", ftypeInt),
    new DumpFieldDef("userType", ftypeInt),
    new DumpFieldDef("lastname", ftypeString),
    new DumpFieldDef("firstname", ftypeString),
    new DumpFieldDef("phone", ftypeString),
    new DumpFieldDef("email", ftypeString),
    new DumpFieldDef("department", ftypeString),
    new DumpFieldDef("preferences", ftypeSpecial, false),
    new DumpFieldDef("autoAddKeywords", ftypeBoolean),
    new DumpFieldDef("autoAddLocations", ftypeBoolean),
    new DumpFieldDef("autoAddSponsors", ftypeBoolean),
    new DumpFieldDef("preferredKeywords", ftypeSpecial, false),
    new DumpFieldDef("preferredLocations", ftypeSpecial, false),
    new DumpFieldDef("preferredSponsors", ftypeSpecial, false),
  };

  public static final DumpFieldDef[] eventFdefs = {
    new DumpFieldDef("id", ftypeInt),
    new DumpFieldDef("creator", ftypeInt),
    new DumpFieldDef("public", ftypeBoolean),
    new DumpFieldDef("lastmod", ftypeTimestamp),
    new DumpFieldDef("seq", ftypeInt),
    new DumpFieldDef("created", ftypeTimestamp),
    new DumpFieldDef("startdate", ftypeDate),
    new DumpFieldDef("starttime", ftypeTime),
    new DumpFieldDef("enddate", ftypeDate),
    new DumpFieldDef("endtime", ftypeTime),
    new DumpFieldDef("shortdesc", ftypeString),
    new DumpFieldDef("longdesc", ftypeString),
    new DumpFieldDef("link", ftypeString),
    new DumpFieldDef("recurringStatus", ftypeChar),
    new DumpFieldDef("cost", ftypeString),
    new DumpFieldDef("location", ftypeInt),
    new DumpFieldDef("sponsor", ftypeInt),
    new DumpFieldDef("eventKeywords", ftypeSpecial, false),
  };

  public static final DumpFieldDef[] eventRefFdefs = {
    new DumpFieldDef("eventid", ftypeInt),
    new DumpFieldDef("userid", ftypeInt),
    new DumpFieldDef("purpose", ftypeChar),
  };

  public static final DumpFieldDef[] adminGroupFdefs = {
    new DumpFieldDef("groupname", ftypeString),
    new DumpFieldDef("description", ftypeString),
    new DumpFieldDef("groupOwner", ftypeInt),
    new DumpFieldDef("ownerid", ftypeInt),
    new DumpFieldDef("groupMembers", ftypeSpecial, false),
    new DumpFieldDef("id", ftypeInt),
  };

  public static final DumpFieldDef[] dbLastmodDefs = {
    new DumpFieldDef("dbname", ftypeString),
    new DumpFieldDef("lastmod", ftypeTimestamp),
  };

  public static final DumpFieldDef[] andFilterFdefs = {
    new DumpFieldDef("id", ftypeInt),
    new DumpFieldDef("name", ftypeString),
    new DumpFieldDef("title", ftypeString),
    new DumpFieldDef("description", ftypeString),
    new DumpFieldDef("parent", ftypeInt),
  };

  public static final DumpFieldDef[] creatorFilterFdefs = {
    new DumpFieldDef("id", ftypeInt),
    new DumpFieldDef("name", ftypeString),
    new DumpFieldDef("title", ftypeString),
    new DumpFieldDef("description", ftypeString),
    new DumpFieldDef("parent", ftypeInt),
    new DumpFieldDef("ref", ftypeInt),
  };

  public static final DumpFieldDef[] keyFilterFdefs = {
    new DumpFieldDef("id", ftypeInt),
    new DumpFieldDef("name", ftypeString),
    new DumpFieldDef("title", ftypeString),
    new DumpFieldDef("description", ftypeString),
    new DumpFieldDef("parent", ftypeInt),
    new DumpFieldDef("ref", ftypeInt),
  };

  public static final DumpFieldDef[] locationFilterFdefs = {
    new DumpFieldDef("id", ftypeInt),
    new DumpFieldDef("name", ftypeString),
    new DumpFieldDef("title", ftypeString),
    new DumpFieldDef("description", ftypeString),
    new DumpFieldDef("parent", ftypeInt),
    new DumpFieldDef("ref", ftypeInt),
  };

  public static final DumpFieldDef[] orFilterFdefs = {
    new DumpFieldDef("id", ftypeInt),
    new DumpFieldDef("name", ftypeString),
    new DumpFieldDef("title", ftypeString),
    new DumpFieldDef("description", ftypeString),
    new DumpFieldDef("parent", ftypeInt),
  };

  public static final DumpFieldDef[] sponsorFilterFdefs = {
    new DumpFieldDef("id", ftypeInt),
    new DumpFieldDef("name", ftypeString),
    new DumpFieldDef("title", ftypeString),
    new DumpFieldDef("description", ftypeString),
    new DumpFieldDef("parent", ftypeInt),
    new DumpFieldDef("ref", ftypeInt),
  };

  public static final DumpFieldDef[] notFilterFdefs = {
    new DumpFieldDef("id", ftypeInt),
    new DumpFieldDef("name", ftypeString),
    new DumpFieldDef("title", ftypeString),
    new DumpFieldDef("description", ftypeString),
    new DumpFieldDef("parent", ftypeInt),
    new DumpFieldDef("ref", ftypeInt),
  };

  public static class DbLastmodVO {
    public String dbname;
    public Timestamp lastmod;

    public String getDbname() {
      return dbname;
    }

    public void setDbname(String val) {
      dbname = val;
    }

    public Timestamp getLastmod() {
      return lastmod;
    }

    public void setLastmod(Timestamp val) {
      lastmod = val;
    }
  }
}

