package edu.rpi.cct.uwcal.tools.dump;

import edu.washington.cac.calfacade.shared.AdminGroupVO;
import edu.washington.cac.calfacade.shared.AuthUserVO;
import edu.washington.cac.calfacade.shared.AuthUserPrefsVO;
import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.EventRefVO;
import edu.washington.cac.calfacade.shared.filter.AndFilterVO;
import edu.washington.cac.calfacade.shared.filter.CreatorFilterVO;
import edu.washington.cac.calfacade.shared.filter.FilterVO;
import edu.washington.cac.calfacade.shared.filter.KeywordFilterVO;
import edu.washington.cac.calfacade.shared.filter.LocationFilterVO;
import edu.washington.cac.calfacade.shared.filter.NotFilterVO;
import edu.washington.cac.calfacade.shared.filter.OrFilterVO;
import edu.washington.cac.calfacade.shared.filter.SponsorFilterVO;
import edu.washington.cac.calfacade.shared.KeywordAttrVO;
import edu.washington.cac.calfacade.shared.KeywordVO;
import edu.washington.cac.calfacade.shared.LocationVO;
import edu.washington.cac.calfacade.shared.SponsorVO;
import edu.washington.cac.calfacade.shared.UserAuth;
import edu.washington.cac.calfacade.shared.UserVO;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;

/** Class which restores via jdbc.
 *
 * @author Mike Douglass   douglm@rpi.edu
 * @version 1.0
 */
public class JdbcRestore implements RestoreIntf {
  private Connection conn;
  private String url;
  private String className;
  private String id;
  private String pw;
  private int dbType;

  /** This needs the CalEnv class to be properly set up.
   */
  private int maxDbStringLength = 0;

  public JdbcRestore(String url,
                     String className,
                     String id,
                     String pw,
                     int dbType) throws Throwable {
    this.url = url;
    this.className = className;
    this.id = id;
    this.pw = pw;
    this.dbType = dbType;
  }

  /** Call after any init phase
   */
  public void open() throws Throwable {
    Class c = Class.forName(className);
    if (c == null) {
      throw new Exception("Unable to load class " + className);
    }

    conn = DriverManager.getConnection(url, id, pw);
  }

  /** Call at end of restoring to finish up. Will restore any cached values.
   */
  public void close() throws Throwable {
    try {
      conn.close();
    } catch (Throwable t) {
    }
  }

  /** Restore users NOT subscriptions
   */
  public void restoreUser(UserVO o) throws Throwable {
    PreparedStatement ps = null;

    try {
      ps = conn.prepareStatement(
            "INSERT INTO users VALUES (?,?,?,?,?,?)");
      ps.setInt(1, o.getId());
      ps.setString(2, o.getAccount());
      ps.setTimestamp(3, o.getCreated());
      ps.setTimestamp(4, o.getLogon());
      ps.setTimestamp(5, o.getLastAccess());
      ps.setTimestamp(6,o.getLastModify());

      ps.executeUpdate();
    } finally {
      if (ps != null) {
        ps.close();
      }
    }
  }

  /** Restore an admin group and its members - though not the user entries nor
   * the authuser entries.
   */
  public void restoreAdminGroup(AdminGroupVO o) throws Throwable {
    PreparedStatement ps = null;

    try {
      ps = conn.prepareStatement(
            "INSERT INTO adminGroups VALUES (?,?,?,?)");
      ps.setString(1, o.getName());
      ps.setString(2, o.getDescription());
      ps.setString(3, o.getGroupOwner().getAccount());
      ps.setString(4, o.getOwner().getAccount());

      ps.executeUpdate();
      ps.close();

      Collection mbrs = o.getGroupMembers();

      if (mbrs != null) {
        Iterator uit = mbrs.iterator();

        while (uit.hasNext()) {
          UserVO u = (UserVO)uit.next();

          ps = conn.prepareStatement(
                "INSERT INTO adminGroupMembers VALUES (?,?)");
          ps.setString(1, o.getName());
          ps.setString(2, u.getAccount());

          ps.executeUpdate();
          ps.close();
        }
      }
    } finally {
      if (ps != null) {
        ps.close();
      }
    }
  }

  /** Restore an auth user and preferences
   */
  public void restoreAuthUser(AuthUserVO o) throws Throwable {
    PreparedStatement ps = null;

    try {
      ps = conn.prepareStatement(
            "INSERT INTO auth VALUES (?,?,?,?,?,?,?)");
      ps.setInt(1, o.getUser().getId());
      ps.setString(2, toDbtype(o.getUsertype()));
      ps.setString(3, o.getLastname());
      ps.setString(4, o.getFirstname());
      ps.setString(5, o.getPhone());
      ps.setString(6, o.getEmail());
      ps.setString(7, o.getDept());

      ps.executeUpdate();
    } finally {
      if (ps != null) {
        ps.close();
      }
    }
  }

  /** Restore an event and event keyword entries
   */
  public void restoreEvent(EventVO o) throws Throwable {
    PreparedStatement ps = null;

    try {
      ps = conn.prepareStatement(
          "INSERT INTO events VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?");
      ps.setInt(1, o.getId());
      ps.setTimestamp(2, o.getLastmod());
      ps.setInt(3, o.getSeq());
      ps.setString(4, boolVal(o.getPublic()));
      ps.setTimestamp(5, o.getCreated());
      ps.setDate(6, o.getStartdate());
      ps.setTime(7, o.getStarttime());
      ps.setDate(8, o.getEnddate());
      ps.setTime(9, o.getEndtime());
      ps.setString(10, o.getShortdesc());

      String ld = o.getLongdesc();
      String ld1 = null;

      if (maxDbStringLength > 0) {
        if ((ld != null) && (ld.length() > maxDbStringLength)) {
          ld1 = ld.substring(maxDbStringLength);
          ld = ld.substring(0, maxDbStringLength);
        }
      }
      ps.setString(11, ld);
      ps.setString(12, ld1);
      ps.setString(13, null);   // longdesc2

      ps.setString(13, o.getLink());
      ps.setString(14, "F");
      ps.setInt(15, o.getLocation().getId());
      ps.setInt(16, o.getSponsor().getId());
      ps.setString(17, o.getCreator().getAccount());
  //      sqlChar(o.getRecurringStatus());
      ps.setString(18, o.getCost());

      ps.executeUpdate();
      ps.close();

      restoreEventKeys(o);
    } finally {
      if (ps != null) {
        ps.close();
      }
    }
  }

  /** Restore event ref
   */
  public void restoreEventRef(EventRefVO o) throws Throwable {
    PreparedStatement ps = null;

    try {
      ps = conn.prepareStatement(
            "INSERT INTO eventrefs VALUES (?,?,?)");
      ps.setInt(1, o.getEventid());
      ps.setInt(2, o.getUser().getId());
      ps.setString(3, String.valueOf(o.getPurpose()));

      ps.executeUpdate();
    } finally {
      if (ps != null) {
        ps.close();
      }
    }
  }

  /** Restore keyword and attributes
   */
  public void restoreKeyword(KeywordVO o) throws Throwable {
    PreparedStatement ps = null;

    try {
      ps = conn.prepareStatement(
            "INSERT INTO keywords VALUES (?,?,?,?,?,?)");
      ps.setInt(1, o.getId());
      ps.setString(2, o.getWord());
      ps.setString(3, o.getTitle());
      ps.setString(4, o.getDescription());
      ps.setString(5, o.getCreator().getAccount());
      ps.setString(6, boolVal(o.isPublic()));

      ps.executeUpdate();
      ps.close();

      Set s = o.getAttributes();

      if ((s != null) && (s.size() > 0)) {
        Iterator it = s.iterator();

        while (it.hasNext()) {
          KeywordAttrVO kao = (KeywordAttrVO)it.next();

          ps = conn.prepareStatement(
              "INSERT INTO keyword_attrs VALUES (?,?,?,?)");
          ps.setInt(1, kao.getId());
          ps.setInt(2, kao.getKeyword().getId());
          ps.setString(3, kao.getName());
          ps.setString(4, kao.getVal());

          ps.executeUpdate();
          ps.close();
        }
      }
    } finally {
      if (ps != null) {
        ps.close();
      }
    }
  }

  /** Restore location
   */
  public void restoreLocation(LocationVO o) throws Throwable {
    PreparedStatement ps = null;

    try {
      ps = conn.prepareStatement(
            "INSERT INTO locations VALUES (?,?,?,?,?,?)");
      ps.setInt(1, o.getId());
      ps.setString(2, o.getAddress());
      ps.setString(3, o.getSubaddress());
      ps.setString(4, o.getLink());
      ps.setString(5, o.getCreator().getAccount());
      ps.setString(6, boolVal(o.isPublic()));

      ps.executeUpdate();
    } finally {
      if (ps != null) {
        ps.close();
      }
    }
  }

  /** Restore sponsor
   */
  public void restoreSponsor(SponsorVO o) throws Throwable {
    PreparedStatement ps = null;

    try {
      ps = conn.prepareStatement(
            "INSERT INTO sponsors VALUES (?,?,?,?,?,?,?)");
      ps.setInt(1, o.getId());
      ps.setString(2, o.getName());
      ps.setString(3, o.getPhone());
      ps.setString(4, o.getEmail());
      ps.setString(5, o.getLink());
      ps.setString(6, o.getCreator().getAccount());
      ps.setString(7, boolVal(o.isPublic()));

      ps.executeUpdate();
    } finally {
      if (ps != null) {
        ps.close();
      }
    }
  }

  /** Restore filter
   */
  public void restoreFilter(FilterVO o) throws Throwable {
    /* Set this according to class of filter.
     */
    char typeFlag;
    int ref = 0;

    if (o instanceof AndFilterVO) {
      typeFlag = 'A';
    } else if (o instanceof KeywordFilterVO) {
      typeFlag = 'K';

      KeywordVO k = ((KeywordFilterVO)o).getKeyword();

      ref = k.getId();
    } else if (o instanceof CreatorFilterVO) {
      typeFlag = 'C';

      UserVO u = ((CreatorFilterVO)o).getCreator();

      ref = u.getId();
    } else if (o instanceof LocationFilterVO) {
      typeFlag = 'L';

      LocationVO l = ((LocationFilterVO)o).getLocation();

      ref = l.getId();
    } else if (o instanceof OrFilterVO) {
      typeFlag = 'O';
    } else if (o instanceof SponsorFilterVO) {
      typeFlag = 'S';

      SponsorVO sp = ((SponsorFilterVO)o).getSponsor();

      ref = sp.getId();
    } else if (o instanceof NotFilterVO) {
      typeFlag = 'N';

      FilterVO f = ((NotFilterVO)o).getFilter();

      ref = f.getId();
    } else {
      throw new Exception("Unimplemented filter " + o);
    }

    PreparedStatement ps = null;

    try {
      ps = conn.prepareStatement(
            "INSERT INTO calendars VALUES (?,?,?,?,?,?,?)");
      ps.setInt(1, o.getId());
      ps.setString(2, String.valueOf(typeFlag));
      ps.setString(3, o.getName());
      ps.setString(4, o.getTitle());
      ps.setString(5, o.getDescription());
      ps.setInt(6, ref);
      ps.setInt(7, o.getParent().getId());

      ps.executeUpdate();
    } finally {
      if (ps != null) {
        ps.close();
      }
    }
  }

  /** Subscribe user to filter.
   */
  public void subscribe(UserVO user, FilterVO f) throws Throwable {
    PreparedStatement ps = null;

    try {
      ps = conn.prepareStatement(
            "INSERT INTO subscriptions VALUES (?,?)");
      ps.setInt(1, user.getId());
      ps.setInt(2, f.getId());

      ps.executeUpdate();
    } finally {
      if (ps != null) {
        ps.close();
      }
    }
  }

  /** Restore db lastmods
   */
  public void restoreDbLastmod(DumpDefs.DbLastmodVO o) throws Throwable {
    PreparedStatement ps = null;

    try {
      ps = conn.prepareStatement(
            "INSERT INTO lastmods VALUES (?,?)");
      ps.setString(1, o.getDbname());
      ps.setTimestamp(2, o.getLastmod());

      ps.executeUpdate();
    } finally {
      if (ps != null) {
        ps.close();
      }
    }
  }

  /** ===================================================================
   *                       Private methods
   *  =================================================================== */

  /** Iterate over the preferences multiple times to build each set of
   * inserts separately.
   */
  private void restoreAuthPrefs(AuthUserPrefsVO o) throws Exception {
    /* first the prefs */

    PreparedStatement ps = null;

    try {
      ps = conn.prepareStatement(
            "INSERT INTO authprefs VALUES (?,?,?,?)");
      ps.setInt(1, o.getAuthUser().getUser().getId());
      ps.setString(2, boolVal(o.getAutoAddKeywords()));
      ps.setString(3, boolVal(o.getAutoAddLocations()));
      ps.setString(4, boolVal(o.getAutoAddSponsors()));

      ps.executeUpdate();
      ps.close();

      /* preferred keywords */
      Set kpref = o.getPreferredKeywords();

      if (kpref != null) {
        Iterator it = kpref.iterator();

        while (it.hasNext()) {
          KeywordVO p = (KeywordVO)it.next();

          ps = conn.prepareStatement(
              "INSERT INTO authprefKeywords VALUES (?,?)");
          ps.setInt(1, o.getAuthUser().getUser().getId());
          ps.setInt(2, p.getId());

          ps.executeUpdate();
          ps.close();
        }
      }

      /* preferred locations */
      Set lpref = o.getPreferredLocations();

      if (lpref != null) {
        Iterator it = lpref.iterator();

        while (it.hasNext()) {
          LocationVO p = (LocationVO)it.next();

          ps = conn.prepareStatement(
              "INSERT INTO authprefLocations VALUES (?,?)");
          ps.setInt(1, o.getAuthUser().getUser().getId());
          ps.setInt(2, p.getId());

          ps.executeUpdate();
          ps.close();
        }
      }

      /* preferred sponsors */
      Set spref = o.getPreferredSponsors();

      if (spref != null) {
        Iterator it = spref.iterator();

        while (it.hasNext()) {
          SponsorVO p = (SponsorVO)it.next();

          ps = conn.prepareStatement(
              "INSERT INTO authprefSponsors VALUES (?,?)");
          ps.setInt(1, o.getAuthUser().getUser().getId());
          ps.setInt(2, p.getId());

          ps.executeUpdate();
        }
      }
    } finally {
      if (ps != null) {
        ps.close();
      }
    }
  }

  private void restoreEventKeys(EventVO ev) throws Exception {
    Collection evks = ev.getKeywords();
    int evid = ev.getId();

    Iterator it = evks.iterator();

    while (it.hasNext()) {
      KeywordVO k = (KeywordVO)it.next();

      PreparedStatement ps = null;

      try {
        ps = conn.prepareStatement(
              "INSERT INTO event_keywords VALUES (?,?)");
        ps.setInt(1, evid);
        ps.setInt(2, k.getId());

        ps.executeUpdate();
      } finally {
      if (ps != null) {
        ps.close();
      }
      }
    }
  }

  private String boolVal(boolean val) {
    if (val) {
      return "T";
    }

    return "F";
  }

  /** Convert external user type to internal db type.
   *
   * @return String   internal type - null for no privileges.
   */
  private String toDbtype(int usertype) {
    if (usertype == UserAuth.noPrivileges) {
      return null;
    }

    if ((usertype & UserAuth.superUser) != 0) {
      return "A";
    }

    if (usertype == UserAuth.contentAdminUser) {
      return "C";
    }

    if (usertype == UserAuth.publicEventUser) {
      return "P";
    }

    if (usertype == UserAuth.alertUser) {
      return "a";
    }

    if (usertype == (UserAuth.alertUser + UserAuth.publicEventUser)) {
      // semi-priv
      return "S";
    }

    return null;
  }
}

