package edu.rpi.cct.uwcal.tools.dump;

import edu.washington.cac.calfacade.shared.AdminGroupVO;
import edu.washington.cac.calfacade.shared.AuthUserVO;
import edu.washington.cac.calfacade.shared.AuthUserPrefsVO;
import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.EventRefVO;
import edu.washington.cac.calfacade.shared.filter.AndFilterVO;
import edu.washington.cac.calfacade.shared.filter.CreatorFilterVO;
import edu.washington.cac.calfacade.shared.filter.FilterVO;
import edu.washington.cac.calfacade.shared.filter.KeywordFilterVO;
import edu.washington.cac.calfacade.shared.filter.LocationFilterVO;
import edu.washington.cac.calfacade.shared.filter.NotFilterVO;
import edu.washington.cac.calfacade.shared.filter.OrFilterVO;
import edu.washington.cac.calfacade.shared.filter.SponsorFilterVO;
import edu.washington.cac.calfacade.shared.KeywordAttrVO;
import edu.washington.cac.calfacade.shared.KeywordVO;
import edu.washington.cac.calfacade.shared.LocationVO;
import edu.washington.cac.calfacade.shared.SponsorVO;
import edu.washington.cac.calfacade.shared.UserVO;

import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.HashMap;
import java.util.Set;
import java.util.Stack;
import java.util.TreeSet;
import java.util.Vector;

import org.xml.sax.Attributes;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;

/** A class to parse the xml dump and call abstract emthods with the
 * results..
 *
 * @author Mike Douglass   douglm@rpi.edu
 * @version 1.0
 */
public abstract class ParseDump extends DefaultHandler implements DumpDefs {
  protected boolean debug;

  protected int errors;

  private String parserName = "org.apache.xerces.parsers.SAXParser";

  private HashMap sectionNames;

  private HashMap objectNames;

  /** An index defining the class we're restoring
   */
  private static final int restoringNothing = 0;

  private static final int restoringUser = 1;
  private static final int restoringLocation = 2;
  private static final int restoringSponsor = 3;
  private static final int restoringKeyword = 4;
  private static final int restoringEvent = 5;
  private static final int restoringEventRef = 6;
  private static final int restoringAuthUser = 7;
  private static final int restoringAdminGroup = 8;

  private static final int restoringAndFilter = 9;
  private static final int restoringCreatorFilter = 10;
  private static final int restoringKeywordFilter = 11;
  private static final int restoringLocationFilter = 12;
  private static final int restoringNotFilter = 13;
  private static final int restoringOrFilter = 14;
  private static final int restoringSponsorFilter = 15;

  private static final int restoringDbLastmod = 16;

  private int restoreState = restoringNothing;

  protected static class UserMap extends HashMap {
    HashMap nameMap = new HashMap();

    protected void put(UserVO u) {
      put(new Integer(u.getId()), u);
      nameMap.put(u.getAccount(), u);
    }

    protected UserVO get(int id) {
      return (UserVO)get(new Integer(id));
    }

    protected UserVO get(String account) {
      return (UserVO)nameMap.get(account);
    }
  }

  protected static class KeywordMap extends HashMap {
    protected void put(KeywordVO k) {
      put(new Integer(k.getId()), k);
    }

    protected KeywordVO get(int id) {
      return (KeywordVO)get(new Integer(id));
    }
  }

  protected static class LocationMap extends HashMap {
    protected void put(LocationVO l) {
      put(new Integer(l.getId()), l);
    }

    protected LocationVO get(int id) {
      return (LocationVO)get(new Integer(id));
    }
  }

  protected static class SponsorMap extends HashMap {
    protected void put(SponsorVO s) {
      put(new Integer(s.getId()), s);
    }

    protected SponsorVO get(int id) {
      return (SponsorVO)get(new Integer(id));
    }
  }

  /** We need implant something like the real id in objects
   */
  private UserMap users = new UserMap();

  private KeywordMap keywords = new KeywordMap();

  private LocationMap locations = new LocationMap();

  private SponsorMap sponsors = new SponsorMap();

  /* --------------------------------------------------------------------
     Define the state
     -------------------------------------------------------------------- */

  /** When we're within the outermost tag
   */
  private static final int restoringData = 1;

  /** When we're within a goup tag e.g keywords
   */
  private static final int restoringObjects = 2;

  /** When we're within an object e.g. an event
   */
  private static final int restoringObject = 3;

  /** When we're within a field e.g. creator
   */
  private static final int restoringField = 4;

  private static class State {
    String expectedEndTag;
    int state;

    public State(String tag, int st) {
      expectedEndTag = tag;
      state = st;
    }
  }

  private static class StateStack extends Stack {
    State getTop() {
      if (empty()) {
        return null;
      }

      return (State)peek();
    }
  };

  private StateStack curState = new StateStack();

  /** Indexed by restoreState.
   */
  private HashMap[] fieldDefs;

  /** Indexed by restoreState.
   */
  private DumpDefs.DumpFieldDef[][] fieldDefsInit = {
    null,
    userFdefs,
    locationFdefs,
    sponsorFdefs,
    keywordFdefs,
    eventFdefs,
    eventRefFdefs,
    authUserFdefs,
    adminGroupFdefs,

    andFilterFdefs,
    creatorFilterFdefs,
    keyFilterFdefs,
    locationFilterFdefs,
    notFilterFdefs,
    orFilterFdefs,
    sponsorFilterFdefs,

    dbLastmodDefs};

  /** Current field defs
   */
  private HashMap curFieldDefs;

  /** Top of stack is current field def.
   * Should be empty at the end of an object.
   */
  private Stack curFdef = new Stack();

  /** Current object
   */
  private Object curObject;

  /** Current value and converted values.
   */
  private String curVal;
  private int curIntVal;
  private Timestamp curTimestampVal;
  private Date curDateVal;
  private Time curTimeVal;
  private boolean curBooleanVal;
  private char curCharVal;

  /* --------------------------------------------------------------------
     These are fields associated with restoring the current object.
     -------------------------------------------------------------------- */
  private KeywordAttrVO curKeyAttrs;
  private AuthUserPrefsVO curAuthPrefs;

  private XMLReader xrdr;

  /** ===================================================================
   *                       Constructor
   *  =================================================================== */

  public ParseDump() {
    sectionNames = new HashMap();

    sectionNames.put(sectionFilters, sectionFilters);
    sectionNames.put(sectionUsers, sectionUsers);
    sectionNames.put(sectionLocations, sectionLocations);
    sectionNames.put(sectionSponsors, sectionSponsors);
    sectionNames.put( sectionKeywords, sectionKeywords);
    sectionNames.put(sectionAuthUsers, sectionAuthUsers);
    sectionNames.put(sectionEvents, sectionEvents);
    sectionNames.put(sectionEventRefs, sectionEventRefs);
    sectionNames.put(sectionAdminGroups, sectionAdminGroups);
    sectionNames.put(sectionDbLastmods, sectionDbLastmods);

    objectNames = new HashMap();

    objectNames.put(objectAndFilter, new Integer(restoringAndFilter));
    objectNames.put(objectCreatorFilter,
                    new Integer(restoringCreatorFilter));
    objectNames.put(objectKeywordFilter,
                    new Integer(restoringKeywordFilter));
    objectNames.put(objectLocationFilter,
                    new Integer(restoringLocationFilter));
    objectNames.put(objectNotFilter, new Integer(restoringNotFilter));
    objectNames.put(objectOrFilter, new Integer(restoringOrFilter));
    objectNames.put(objectSponsorFilter, new Integer(restoringSponsorFilter));

    objectNames.put(objectUser, new Integer(restoringUser));
    objectNames.put(objectLocation, new Integer(restoringLocation));
    objectNames.put(objectSponsor, new Integer(restoringSponsor));
    objectNames.put(objectKeyword, new Integer(restoringKeyword));
    objectNames.put(objectAuthUser, new Integer(restoringAuthUser));
    objectNames.put(objectEvent, new Integer(restoringEvent));
    objectNames.put(objectEventRef, new Integer(restoringEventRef));
    objectNames.put(objectAdminGroup, new Integer(restoringAdminGroup));
    objectNames.put(objectDbLastmod, new Integer(restoringDbLastmod));

    fieldDefs = new HashMap[fieldDefsInit.length];

    for (int i = 0; i < fieldDefsInit.length; i++) {
      HashMap hm = new HashMap();
      DumpDefs.DumpFieldDef[] fdefs = fieldDefsInit[i];

      if (fdefs != null) {
        for (int fi = 0; fi < fdefs.length; fi++) {
          hm.put(fdefs[fi].tagName, fdefs[fi]);
        }
      }

      fieldDefs[i] = hm;
    }
  }

  /** ===================================================================
   *                       ParseDump methods
   *  =================================================================== */

  public void parse(String fileName) throws Exception {
    try {
      xrdr = XMLReaderFactory.createXMLReader(parserName);

      xrdr.setContentHandler(this);
      xrdr.setErrorHandler(this);

      xrdr.parse(fileName);
    } catch (Throwable t) {
      t.printStackTrace();
    }
  }

  /** ===================================================================
   *                       Abstract methods
   *  =================================================================== */

  /** Called when we reach the end of the dump
   */
  public abstract void endDump();

  /** Called when we start a new dump section. We can expect different
   * object classes
   */
  public abstract void startSection(String name);

  /** Called when we end a dump section.A start section should follow soon
   * after.
   */
  public abstract void endSection(String name) throws Throwable;

  /** Called when an object has been restored. Any references to other
   * objects will be filled in with dummy objects defining the key. They
   * need to be replaced with the real objects either immediately if they
   * exist, or later in the restore process.
   */
  public abstract void endObject(Object o) throws Throwable;

  /** ===================================================================
   *                    Content handler methods
   *  =================================================================== */

  public void startDocument() throws SAXException {
  }

  public void startElement(String uri, String localName,
                           String qName,
                           Attributes atts) throws SAXException {
    State st = curState.getTop();

    if (st == null) {
      /* Waiting for dumpTag */
      if (dumpTag.equals(localName)) {
        curState.push(new State(dumpTag, restoringData));
      } else {
        expected(dumpTag);
      }
    } else if (st.state == restoringData) {
      // We expect the start of a group
      if (sectionNames.get(localName) != null) {
        curState.push(new State(localName, restoringObjects));
        startSection(localName);
      } else {
        expected("object section start. Found " + localName);
      }
    } else if (st.state == restoringObjects) {
      // We're expecting an object name
      Integer i = (Integer)objectNames.get(localName);

      if (i != null) {
        /** Starting a new object
         */

        if (restoreState != restoringNothing) {
          throw new SAXException("Started new object " + localName +
                                 " while in restore state " + restoreState);
        }

        restoreState = i.intValue();

        startObject();
        curState.push(new State(localName, restoringObject));
      } else {
        expected("object start");
      }
    } else if ((st.state == restoringObject) ||
                (st.state == restoringField)) {
      if (objectFieldTag(localName, true)) {
        /** We know this tag
         */
        curState.push(new State(localName, restoringField));
      } else {
        if (debug) {
          dmsg("startElement: uri=" + uri + "  localName=" + localName +
               " qName=" + qName);
          expected("field start. Found " + localName);
        }
      }
    }
  }

  public void endElement(String uri,
                         String localName,
                         String qName) throws SAXException {
    State st = curState.getTop();

    if (st == null) {
      throw new SAXException("Ending tag " + localName +
                                 " but not started");
    }

    /* First - the end tag should match the start tag.
     */
    if (!st.expectedEndTag.equals(localName)) {
      throw new SAXException("Ending tag " + localName +
                                 " but expected " + st.expectedEndTag);
    }

    if (st.state == restoringField) {
      if (objectFieldTag(localName, false)) {
        /** We know this tag
         */
      } else if (debug) {
        dmsg("endElement: uri=" + uri + "  localName=" + localName +
             " qName=" + qName);
        throw new SAXException("Unexpected tag " + localName);
      }
    } else if (st.state == restoringObject) {
      Integer i = (Integer)objectNames.get(localName);

      if (i != null) {
        /** Ending an object
         */

        if (restoreState == restoringNothing) {
          throw new SAXException("Ending new object " + localName +
                                 " which wasnt' started");
        }

        endObject();

        restoreState = restoringNothing;
      } else {
        throw new SAXException("Unexpected ending tag " + localName);
      }
    } else if (st.state == restoringObjects) {
      try {
        endSection(localName);
      } catch (Throwable t) {
        error("Exception from endSection");
        t.printStackTrace();
        throw new SAXException("Exception from endSection");
      }
    } else if (st.state == restoringData) {
      endDump();
    } else {
      throw new SAXException("Unexpected ending tag " + localName +
                             " when doing nothing.");
    }

    if (!curState.empty()) {
      curState.pop();
    }
  }

  /** Here's where we see the character data.
   */
  public void characters(char[] ch,
                         int start,
                         int length) throws SAXException {
    if (curFdef.empty()) {
      return;
    }

    DumpDefs.DumpFieldDef fd = (DumpDefs.DumpFieldDef)curFdef.peek();

    if ((fd == null) || (ch == null)) {
      // Not in any known tag

      if ((debug) && (ch != null)) {
        dmsg("Unclaimed characters=" + new String(ch, start, length));
      }
      return;
    }

    boolean allWhitespace = true;

    for (int i = start; i < start + length; i++) {
      if (!Character.isWhitespace(ch[i])) {
        allWhitespace = false;
        break;
      }
    }

    if (allWhitespace) {
      return;
    }

    if (length != 0) {
      if (curVal == null) {
        curVal = new String(ch, start, length);
      } else {
        curVal += new String(ch, start, length);
      }
    }
  }

  public void warning(SAXParseException e) throws SAXException {
    handleParseError("Warning", e);
  }

  public void error(SAXParseException e) throws SAXException {
    handleParseError("Error", e);
  }

  public void fatalError(SAXParseException e) throws SAXException {
    handleParseError("Fatal error", e);
  }

  /** ===================================================================
   *                    Convenience methods
   *  =================================================================== */

  public void dmsg(String msg) {
    System.out.println(msg);
  }

  /** ===================================================================
   *                    Private methods
   *  =================================================================== */

  private void handleParseError(String level, SAXParseException e) {
    String sysid = e.getSystemId();

    error(level + ": " + sysid + " at line " +
          e.getLineNumber() + " col " +
          e.getColumnNumber());
  }

  private void expected(String msg) throws SAXException {
    throw new SAXException("Expected " + msg);
  }

  /** Start a new object VO on restoreState
   */
  private void startObject() throws SAXException {
    switch (restoreState) {
      case restoringUser:
        curObject = new UserVO();
        break;

      case restoringLocation:
        curObject = new LocationVO();
        break;

      case restoringSponsor:
        curObject = new SponsorVO();
        break;

      case restoringKeyword:
        curObject = new KeywordVO();
        break;

      case restoringEvent:
        curObject = new EventVO();
        break;

      case restoringEventRef:
        curObject = new EventRefVO();
        break;

      case restoringAuthUser:
        curObject = new AuthUserVO();
        break;

      case restoringAdminGroup:
        curObject = new AdminGroupVO();
        break;

      case restoringDbLastmod:
        curObject = new DumpDefs.DbLastmodVO();
        break;

      case restoringAndFilter:
        curObject = new AndFilterVO();
        break;

      case restoringCreatorFilter:
        curObject = new CreatorFilterVO();
        break;

      case restoringKeywordFilter:
        curObject = new KeywordFilterVO();
        break;

      case restoringLocationFilter:
        curObject = new LocationFilterVO();
        break;

      case restoringNotFilter:
        curObject = new NotFilterVO();
        break;

      case restoringOrFilter:
        curObject = new OrFilterVO();
        break;

      case restoringSponsorFilter:
        curObject = new SponsorFilterVO();
        break;

      default:
        throw new SAXException("Unhandled restoreState: " + restoreState);
    }

    curFieldDefs = fieldDefs[restoreState];
  }

  /** Is the given tag an object field tag?
   */
  private boolean objectFieldTag(String name,
                                 boolean start) throws SAXException {
    if (curFieldDefs == null) {
      return false;
    }

    DumpDefs.DumpFieldDef fd = (DumpDefs.DumpFieldDef)curFieldDefs.get(name);

    if (fd == null) {
      return false;
    }

    if (start) {
      if (debug) {
        dmsg("objectFieldTag, start: tagName=" + fd.tagName);
      }

      // Push it, give the object processor a chance to initialise if needed
      // and return true.
      curFdef.push(fd);
      field(fd, true);

      curVal = null;

      return true;
    }

    // Ending. fd should match the top entry.

    DumpDefs.DumpFieldDef tfd = (DumpDefs.DumpFieldDef)curFdef.peek();

    if (debug) {
      dmsg("objectFieldTag, end: tfd.tagName=" + tfd.tagName +
           " fd.tagName=" + fd.tagName);
    }

    if (!tfd.tagName.equals(fd.tagName)) {
      throw new SAXException("Mismatched tags " + tfd.tagName +
                             " and " + fd.tagName);
    }

    /** Convert the field
     */

    try {
      if (curVal == null) {
        curIntVal = 0;
        curTimestampVal = null;
        curDateVal = null;
        curTimeVal = null;
        curBooleanVal = false;
      } else if (fd.ftype == DumpDefs.ftypeInt) {
        curIntVal = Integer.parseInt(curVal);
      } else if (fd.ftype == DumpDefs.ftypeTimestamp) {
        curTimestampVal = Timestamp.valueOf(curVal);
      } else if (fd.ftype == DumpDefs.ftypeDate) {
        curDateVal = Date.valueOf(curVal);
      } else if (fd.ftype == DumpDefs.ftypeTime) {
        curTimeVal = Time.valueOf(curVal);
      } else if (fd.ftype == DumpDefs.ftypeBoolean) {
        curBooleanVal = curVal.equals("true");
      } else if (fd.ftype == DumpDefs.ftypeChar) {
        if (curVal.length() != 1) {
          throw new Exception("Bad value for char");
        }
        curCharVal = curVal.charAt(0);
      }
    } catch (Throwable t) {
      error("Conversion error for val=\"" + curVal +
            "\" Expected type is " + fd.ftype);
      t.printStackTrace();
      throw new SAXException("Conversion error for val=\"" + curVal +
            "\" Expected type is " + fd.ftype);
    }

    curFdef.pop();
    field(fd, false);

    return true;
  }

  /** End the current object.
   */
  private void endObject() throws SAXException {
    if (!curFdef.empty()) {
      error("Unclosed tags for object.");
      while (!curFdef.empty()) {
        curFdef.pop();
      }
    }

    if (curObject instanceof UserVO) {
      users.put((UserVO)curObject);
    }

    if (curObject instanceof KeywordVO) {
      keywords.put((KeywordVO)curObject);
    }

    if (curObject instanceof LocationVO) {
      locations.put((LocationVO)curObject);
    }

    if (curObject instanceof SponsorVO) {
      sponsors.put((SponsorVO)curObject);
    }

    try {
      endObject(curObject);
    } catch (Throwable t) {
      error("Exception from endObject");
      t.printStackTrace();
      throw new SAXException("Exception from endObject");
    }
    curFieldDefs = null;
  }

  /** We've just started or ended a field.
   * We've got a field def passed as a parameter.
   *
   * <p>If starting, we might want to set up some variables for special
   * fields - e.g. a creator
   *
   * If ending, We've got a value in curVal (maybe). We've got an object in
   * curObject. Now set a field in the object - maybe.
   */
  private void field(DumpDefs.DumpFieldDef fd,
                     boolean start) throws SAXException {
    switch (restoreState) {
      case restoringUser:
        field(fd, start, (UserVO)curObject);
        break;

      case restoringLocation:
        field(fd, start, (LocationVO)curObject);
        break;

      case restoringSponsor:
        field(fd, start, (SponsorVO)curObject);
        break;

      case restoringKeyword:
        field(fd, start, (KeywordVO)curObject);
        break;

      case restoringEvent:
        field(fd, start, (EventVO)curObject);
        break;

      case restoringEventRef:
        field(fd, start, (EventRefVO)curObject);
        break;

      case restoringAuthUser:
        field(fd, start, (AuthUserVO)curObject);
        break;

      case restoringAdminGroup:
        field(fd, start, (AdminGroupVO)curObject);
        break;

      case restoringDbLastmod:
        field(fd, start, (DumpDefs.DbLastmodVO)curObject);
        break;

      case restoringAndFilter:
        field(fd, start, (AndFilterVO)curObject);
        break;

      case restoringCreatorFilter:
        field(fd, start, (CreatorFilterVO)curObject);
        break;

      case restoringKeywordFilter:
        field(fd, start, (KeywordFilterVO)curObject);
        break;

      case restoringLocationFilter:
        field(fd, start, (LocationFilterVO)curObject);
        break;

      case restoringNotFilter:
        field(fd, start, (NotFilterVO)curObject);
        break;

      case restoringOrFilter:
        field(fd, start, (OrFilterVO)curObject);
        break;

      case restoringSponsorFilter:
        field(fd, start, (SponsorFilterVO)curObject);
        break;

      default:
        throw new SAXException("Unhandled restoreState: " + restoreState);
    }

    curFieldDefs = fieldDefs[restoreState];
  }

  /** ===================================================================
   *                   UserVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     UserVO u) throws SAXException {
    if (start) {
      if (fd.tagName.equals("subscriptions")) {
        Collection s = u.getSubscriptions();

        if (s != null) {
          throw new SAXException("Subscriptions not null for " + u);
        }

        s = new Vector();
        u.setSubscriptions(s);
      }

      return;
    }

    if (fd.tagName.equals("id")) {
      u.setId(curIntVal);
    } else if (fd.tagName.equals("account")) {
      u.setAccount(curVal);
    } else if (fd.tagName.equals("created")) {
      u.setCreated(curTimestampVal);
    } else if (fd.tagName.equals("logon")) {
      u.setLogon(curTimestampVal);
    } else if (fd.tagName.equals("lastAccess")) {
      u.setLastAccess(curTimestampVal);
    } else if (fd.tagName.equals("lastModify")) {
      u.setLastModify(curTimestampVal);
    } else if (fd.tagName.equals("subscriptions")) {
      // Nothing to do now
    } else if (fd.tagName.equals("calendarid")) {
      FilterVO f = new FilterVO();

      f.setId(curIntVal);

      Collection s = u.getSubscriptions();

      if (s == null) {
        throw new SAXException("Subscriptions null for " + u);
      }

      s.add(f);
    }
  }

  /** ===================================================================
   *                   LocationVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     LocationVO l) throws SAXException {
    if (!start) {
      if (fd.tagName.equals("id")) {
        l.setId(curIntVal);

        return;
      }

      if (fd.tagName.equals("public")) {
        l.setIsPublic(curBooleanVal);
        return;
      }

      if (fd.tagName.equals("creator")) {
        UserVO cre = users.get(curIntVal);

        if (cre == null) {
          error("Missing creator " + curIntVal + " for " + l);
          cre = new UserVO();
          cre.setId(curIntVal);
        }

        l.setCreator(cre);
        return;
      }
    }

    if (fd.tagName.equals("address")) {
      l.setAddress(curVal);
    } else if (fd.tagName.equals("subaddress")) {
      l.setSubaddress(curVal);
    } else if (fd.tagName.equals("link")) {
      l.setLink(curVal);
    }
  }

  /** ===================================================================
   *                   SponsorVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     SponsorVO s) throws SAXException {
    if (!start) {
      if (fd.tagName.equals("id")) {
        s.setId(curIntVal);

        return;
      }

      if (fd.tagName.equals("public")) {
        s.setIsPublic(curBooleanVal);
        return;
      }

      if (fd.tagName.equals("creator")) {
        UserVO cre = users.get(curIntVal);

        if (cre == null) {
          error("Missing creator " + curIntVal + " for " + s);
          cre = new UserVO();
          cre.setId(curIntVal);
        }

        s.setCreator(cre);
        return;
      }
    }

    if (fd.tagName.equals("name")) {
      s.setName(curVal);
    } else if (fd.tagName.equals("phone")) {
      s.setPhone(curVal);
    } else if (fd.tagName.equals("email")) {
      s.setEmail(curVal);
    } else if (fd.tagName.equals("link")) {
      s.setLink(curVal);
    }
  }

  /** ===================================================================
   *                   KeywordVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     KeywordVO k) throws SAXException {
    if (start) {
      if (fd.tagName.equals("attribute")) {
        if (curKeyAttrs != null) {
          throw new SAXException("curKeyAttrs not null");
        }

        curKeyAttrs = new KeywordAttrVO();
        curKeyAttrs.setKeyword(k);
      }

      return;
    }

    String topTag = null;
    if (!curFdef.empty()) {
      topTag = ((DumpDefs.DumpFieldDef)curFdef.peek()).tagName;
    }

    if ("attribute".equals(topTag)) {
      if (curKeyAttrs == null) {
        throw new SAXException("curKeyAttrs is null");
      }

      try {
        k.addKeyAttr(curKeyAttrs.getName(), curKeyAttrs.getVal());
      } catch (Throwable t) {
        error("Exception adding key attr " + curKeyAttrs + " to " + k);
      }
      curKeyAttrs = null;

      return;
    }

    if ((fd.tagName.equals("id")) && (curKeyAttrs != null)) {
      curKeyAttrs.setId(curIntVal);
      return;
    }

    if (fd.tagName.equals("id")) {
      k.setId(curIntVal);

      return;
    }

    if (fd.tagName.equals("public")) {
      k.setIsPublic(curBooleanVal);
      return;
    }

    if (fd.tagName.equals("creator")) {
      UserVO cre = users.get(curIntVal);

      if (cre == null) {
        error("Missing creator " + curIntVal + " for " + k);
        cre = new UserVO();
        cre.setId(curIntVal);
      }

      k.setCreator(cre);
      return;
    }

    if (fd.tagName.equals("word")) {
      k.setWord(curVal);
    } else if (fd.tagName.equals("title")) {
      k.setTitle(curVal);
    } else if (fd.tagName.equals("description")) {
      k.setDescription(curVal);
    } else if (fd.tagName.equals("name")) {
      if (curKeyAttrs == null) {
        throw new SAXException("curKeyAttrs is null for " + fd.tagName);
      }

      curKeyAttrs.setName(curVal);
    } else if (fd.tagName.equals("val")) {
      if (curKeyAttrs == null) {
        throw new SAXException("curKeyAttrs is null for " + fd.tagName);
      }

      curKeyAttrs.setVal(curVal);
    }
  }

  /** ===================================================================
   *                   EventVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     EventVO e) throws SAXException {
    if (start) {
      return;
    }

    /** If we're adding location, sponsor or eventKeywords then look for the
     * id.
     */

    String topTag = null;
    if (!curFdef.empty()) {
      topTag = ((DumpDefs.DumpFieldDef)curFdef.peek()).tagName;
    }

    if ("eventKeywords".equals(topTag)) {
      expect(fd, "id", "eventKeywords");

      KeywordVO k = new KeywordVO();
      k.setId(curIntVal);
      e.addKeyword(k);

      return;
    }

    if (fd.tagName.equals("id")) {
      e.setId(curIntVal);
      return;
    }

    if (fd.tagName.equals("public")) {
      e.setPublic(curBooleanVal);
      return;
    }

    if (fd.tagName.equals("creator")) {
      UserVO cre = users.get(curIntVal);

      if (cre == null) {
        error("Missing creator " + curIntVal + " for " + e);
        cre = new UserVO();
        cre.setId(curIntVal);
      }

      e.setCreator(cre);
      return;
    }

    if (fd.tagName.equals("lastmod")) {
      e.setLastmod(curTimestampVal);
    } else if (fd.tagName.equals("seq")) {
      e.setSeq(curIntVal);
    } else if (fd.tagName.equals("created")) {
      e.setCreated(curTimestampVal);
    } else if (fd.tagName.equals("startdate")) {
      e.setStartdate(curDateVal);
    } else if (fd.tagName.equals("starttime")) {
      e.setStarttime(curTimeVal);
    } else if (fd.tagName.equals("enddate")) {
      e.setEnddate(curDateVal);
    } else if (fd.tagName.equals("endtime")) {
      e.setEndtime(curTimeVal);
    } else if (fd.tagName.equals("shortdesc")) {
      e.setShortdesc(curVal);
    } else if (fd.tagName.equals("longdesc")) {
      e.setLongdesc(curVal);
    } else if (fd.tagName.equals("link")) {
      e.setLink(curVal);
    } else if (fd.tagName.equals("location")) {
      LocationVO l = new LocationVO();
      l.setId(curIntVal);
      e.setLocation(l);
    } else if (fd.tagName.equals("sponsor")) {
      SponsorVO s = new SponsorVO();
      s.setId(curIntVal);
      e.setSponsor(s);
//    } else if (fd.tagName.equals("recurringStatus")) {
//      e.setRecurringStatus(curCharVal);
    } else if (fd.tagName.equals("cost")) {
      e.setCost(curVal);
    }
  }

  /** ===================================================================
   *                   EventRefVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     EventRefVO er) throws SAXException {
    if (start) {
      return;
    }

    if (fd.tagName.equals("eventid")) {
      er.setEventid(curIntVal);
    } else if (fd.tagName.equals("userid")) {
      UserVO u = users.get(curIntVal);

      if (u == null) {
        error("Missing owner " + curIntVal + " for " + er);
        u = new UserVO();
        u.setId(curIntVal);
      }

      er.setUser(u);
    } else if (fd.tagName.equals("purpose")) {
      er.setPurpose(curCharVal);
    }
  }

  /** ===================================================================
   *                   AuthUserVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     AuthUserVO au) throws SAXException {
    if (start) {
      if (fd.tagName.equals("preferences")) {
        if (curAuthPrefs != null) {
          throw new SAXException("curAuthPrefs not null");
        }

        curAuthPrefs = new AuthUserPrefsVO();
        curAuthPrefs.setId(au.getUser().getId());
        curAuthPrefs.setAuthUser(au);
        au.setPrefs(curAuthPrefs);
      }

      return;
    }

    /** If we have one of the preferred items on the top of the stack
     * then we should be ending the id of an item.
     */

    String topTag = null;
    if (!curFdef.empty()) {
      topTag = ((DumpDefs.DumpFieldDef)curFdef.peek()).tagName;
    }

    if ("preferredKeywords".equals(topTag)) {
      expect(fd, "id", "preferredKeywords");

      KeywordVO k = keywords.get(curIntVal);
      if (k == null) {
        error("Missing preferred keyword " + curIntVal + " for " + au.getUser());
      } else if (!curAuthPrefs.addKeyword(k)) {
        error("Preferred keyword " + k + " not added for " + au.getUser());
      }

      return;
    }

    if ("preferredLocations".equals(topTag)) {
      expect(fd, "id", "preferredLocations");

      LocationVO l = locations.get(curIntVal);
      if (l == null) {
        error("Missing preferred location " + curIntVal + " for " + au.getUser());
      } else if (!curAuthPrefs.addLocation(l)) {
        error("Preferred location " + l + " not added for " + au.getUser());
      }

      return;
    }

    if ("preferredSponsors".equals(topTag)) {
      expect(fd, "id", "preferredSponsors");

      SponsorVO s = sponsors.get(curIntVal);
      if (s == null) {
        error("Missing preferred sponsor " + curIntVal + " for " + au.getUser());
      } else if (!curAuthPrefs.addSponsor(s)) {
        error("Preferred sponsor " + s + " not added for " + au.getUser());
      }

      return;
    }

    if (fd.tagName.equals("id")) {
      UserVO u = users.get(curIntVal);

      if (u == null) {
        error("Missing userr " + curIntVal + " for " + au);
        u = new UserVO();
        u.setId(curIntVal);
      }
      au.setUser(u);
    } else if (fd.tagName.equals("userType")) {
      au.setUsertype(curIntVal);
    } else if (fd.tagName.equals("lastname")) {
      au.setLastname(curVal);
    } else if (fd.tagName.equals("firstname")) {
      au.setFirstname(curVal);
    } else if (fd.tagName.equals("phone")) {
      au.setPhone(curVal);
    } else if (fd.tagName.equals("email")) {
      au.setEmail(curVal);
    } else if (fd.tagName.equals("department")) {
      au.setDept(curVal);
    } else if (fd.tagName.equals("preferences")) {
      curAuthPrefs = null;
    } else if (fd.tagName.equals("autoAddKeywords")) {
      if (curAuthPrefs == null) {
        throw new SAXException("curAuthPrefs is null for autoAddKeywords");
      }

      curAuthPrefs.setAutoAddKeywords(curBooleanVal);
    } else if (fd.tagName.equals("autoAddLocations")) {
      if (curAuthPrefs == null) {
        throw new SAXException("curAuthPrefs is null for autoAddLocations");
      }

      curAuthPrefs.setAutoAddLocations(curBooleanVal);
    } else if (fd.tagName.equals("autoAddSponsors")) {
      if (curAuthPrefs == null) {
        throw new SAXException("curAuthPrefs is null for autoAddSponsors");
      }

      curAuthPrefs.setAutoAddSponsors(curBooleanVal);
    }
  }

  /** ===================================================================
   *                   AdminGroupVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     AdminGroupVO ag) throws SAXException {
    if (start) {
      return;
    }

    String topTag = null;
    if (!curFdef.empty()) {
      topTag = ((DumpDefs.DumpFieldDef)curFdef.peek()).tagName;
    }

    if ("groupMembers".equals(topTag)) {
      expect(fd, "id", "groupMembers");

      UserVO u = users.get(curIntVal);

      if (u == null) {
        error("Missing member " + curIntVal + " for " + ag);
        u = new UserVO();
        u.setId(curIntVal);
      }
      ag.addGroupMember(u);

      return;
    }

    try {
      if (fd.tagName.equals("groupname")) {
        ag.setName(curVal);
      } else if (fd.tagName.equals("description")) {
        ag.setDescription(curVal);
      } else if (fd.tagName.equals("groupOwner")) {
        UserVO u = users.get(curIntVal);

        if (u == null) {
          error("Missing group owner " + curIntVal + " for " + ag);
          u = new UserVO();
          u.setId(curIntVal);
        }

        ag.setGroupOwner(u);
      } else if (fd.tagName.equals("ownerid")) {
        UserVO u = users.get(curIntVal);

        if (u == null) {
          error("Missing owner " + curIntVal + " for " + ag);
          u = new UserVO();
          u.setId(curIntVal);
        }

        ag.setOwner(u);
      }
    } catch (Throwable t) {
      throw new SAXException(t.getMessage());
    }
  }

  /** ===================================================================
   *                   DbLastmodVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     DumpDefs.DbLastmodVO lm) throws SAXException {
    if (start) {
      return;
    }

    if (fd.tagName.equals("dbname")) {
      lm.setDbname(curVal);
    } else if (fd.tagName.equals("lastmod")) {
      lm.setLastmod(curTimestampVal);
    }
  }

  /** ===================================================================
   *                   AndFilterVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     AndFilterVO f) throws SAXException {
    if (start) {
      if (fieldFilterVO(fd, start, f)) {
        return;
      }

      return;
    }

    if (fieldFilterVO(fd, start, f)) {
      return;
    }
  }

  /** ===================================================================
   *                   CreatorFilterVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     CreatorFilterVO f) throws SAXException {
    if (start) {
      if (fieldFilterVO(fd, start, f)) {
        return;
      }

      return;
    }

    if (fieldFilterVO(fd, start, f)) {
      return;
    }

    if (fd.tagName.equals("ref")) {
      UserVO u = users.get(curIntVal);

      if (u == null) {
        error("Missing creator " + curIntVal + " for " + f);
        u = new UserVO();
        u.setId(curIntVal);
      }

      f.setCreator(u);
    }
  }

  /** ===================================================================
   *                   KeywordFilterVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     KeywordFilterVO f) throws SAXException {
    if (start) {
      if (fieldFilterVO(fd, start, f)) {
        return;
      }

      return;
    }

    if (fieldFilterVO(fd, start, f)) {
      return;
    }

    if (fd.tagName.equals("ref")) {
      KeywordVO k = new KeywordVO();
      k.setId(curIntVal);

      f.setKeyword(k);
    }
  }

  /** ===================================================================
   *                   LocationFilterVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     LocationFilterVO f) throws SAXException {
    if (start) {
      if (fieldFilterVO(fd, start, f)) {
        return;
      }

      return;
    }

    if (fieldFilterVO(fd, start, f)) {
      return;
    }

    if (fd.tagName.equals("ref")) {
      LocationVO l = new LocationVO();
      l.setId(curIntVal);

      f.setLocation(l);
    }
  }

  /** ===================================================================
   *                   NotFilterVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     NotFilterVO f) throws SAXException {
    if (start) {
      if (fieldFilterVO(fd, start, f)) {
        return;
      }

      return;
    }

    if (fieldFilterVO(fd, start, f)) {
      return;
    }

    if (fd.tagName.equals("ref")) {
      FilterVO ff = new FilterVO();
      ff.setId(curIntVal);

      f.setFilter(ff);
    }
  }

  /** ===================================================================
   *                   OrFilterVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     OrFilterVO f) throws SAXException {
    if (start) {
      if (fieldFilterVO(fd, start, f)) {
        return;
      }

      return;
    }

    if (fieldFilterVO(fd, start, f)) {
      return;
    }
  }

  /** ===================================================================
   *                   SponsorFilterVO
   *  =================================================================== */
  private void field(DumpDefs.DumpFieldDef fd, boolean start,
                     SponsorFilterVO f) throws SAXException {
    if (start) {
      if (fieldFilterVO(fd, start, f)) {
        return;
      }

      return;
    }

    if (fieldFilterVO(fd, start, f)) {
      return;
    }

    if (fd.tagName.equals("ref")) {
      SponsorVO s = new SponsorVO();
      s.setId(curIntVal);

      f.setSponsor(s);
    }
  }

  private void expect(DumpDefs.DumpFieldDef fd, String expectedTag,
                      String forTag) throws SAXException {
    if (!fd.tagName.equals(expectedTag)) {
      throw new SAXException("expected " + expectedTag +
                             " for " + forTag +
                             ", found " +
                             fd.tagName);
    }
  }

  /** ===================================================================
   *                   FilterVO
   *  =================================================================== */
  private boolean fieldFilterVO(DumpDefs.DumpFieldDef fd, boolean start,
                                  FilterVO f) throws SAXException {
    if (start) {
      return false;
    }

    if (fd.tagName.equals("id")) {
      f.setId(curIntVal);
      return true;
    }

    if (fd.tagName.equals("name")) {
      f.setName(curVal);
      return true;
    }

    if (fd.tagName.equals("title")) {
      f.setTitle(curVal);
      return true;
    }

    if (fd.tagName.equals("description")) {
      f.setDescription(curVal);
      return true;
    }

    if (fd.tagName.equals("parent")) {
      FilterVO pf = new FilterVO();
      pf.setId(curIntVal);
      f.setParent(pf);
      return true;
    }

    return false;
  }


  protected void error(String msg) {
    System.out.println(msg);
    errors++;
  }
}

